# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long, wildcard-import, unused-wildcard-import
"""
.. _`German Federal Institute of Hydrology (BfG)`: https://www.bafg.de/EN
.. _`LARSIM`: http://www.larsim.de/en/the-model/

One can understand version 3 of the L-Land model as an extended, more
complicated version of the application model |lland_v1|, using precipitation,
temperature, air pressure, wind speed, relative humidity and sunshine duration
as meteorological input parameters.  It relies on the equations after Penman
and Penman-Monteith for calculating evaporation and evapotranspiration, which
require much more input data than the simple Turc-Wendling equation but bring
flexibility regarding the simulation time step (Turc-Wendling is limited to
daily time steps, Penman and Penman-Monteith work for any time step size).
The second main difference is that |lland_v3| models the energy
balance of snow processes in more detail (following Knauf, 2006), taking not
only the air temperature but also forcings like the soil heat flux or global
radiation into account explicitly.  The third difference is that |lland_v3|
models the soil temperature to simulate the effect of soil water freezing on
runoff generation.  We created |lland_v3| on behalf of the `German Federal
Institute of Hydrology (BfG)`_ in the context of optimising the control of
the Kiel Canal (Nord-Ostsee-Kanal).

The following list summarises the main components of |lland_v3|:

 * Multiple routines for adjusting the meteorological input data
 * Transformation of sunshine duration into global radiation
   :cite:`ref-Thompson1981`
 * Evaporation from water surfaces after Penman :cite:`ref-penman1948natural`
 * Evapotranspiration (interception evaporation) after Penman-Monteith
   :cite:`ref-Thompson1981`
 * Actual soil evapotranspiration after Wigmosta
   :cite:`ref-wigmosta1994distributed`
 * Surface roughnesses after Quast and Böhm :cite:`ref-bohm1998ermittlung`
 * Mixed precipitation within a definable temperature-range
 * Energy balance of the snow-layer after Knauf :cite:`ref-LUBW2006a`
 * Dynamical adjustment of the snow albedo
 * Optional refinement of the snow-surface temperature
 * Optional (re)freezing of liquid water within the snow-layer
 * A simple snow retention routine
 * A soil routine based on the Xinanjiang model :cite:`ref-zhao1977flood`
 * Frost-sealing of the upper soil-layer
 * One base flow, two interflow and two direct flow components
 * A freely configurable capillary rise routine
 * Separate linear storages for modelling runoff concentration
 * Additional evaporation from water areas within the subcatchment
 * Optional evaporation from inflowing runoff

Some notes for `LARSIM`_ users: |lland_v3| is similar to many `LARSIM`_
models used for forecasting (relying on a combination of options equivalent
to the features listed above), but not identical.  Often, we strive for more
flexibility. One example is that you are allowed to modify parameter
|KTSchnee| to control if and how |lland_v3| adjusts the snow surface
temperature.  However, there are also differences due to technical reasons.
One relevant difference is that *HydPy* does not allow to include "future
values" into the current simulation time step. For example, `LARSIM`_ fills
nightly gaps regarding cloudiness with the average value of the particular
day, while |lland_v3| uses the average over the 24 hours preceding the
current simulation step.  Hence, do not expect the results of |lland_v3|
and `LARSIM`_ to be identical. When switching from one model system to the
other, the need to fine-tune some parameters via calibration might arise.

.. note::

    Some details of |lland_v3| are still under discussion and might change
    in the future.

Integration tests
=================

.. how_to_understand_integration_tests::

Many of the following integration tests are similar to the integration tests
performed for the application model |lland_v1|.  When there are no substantial
differences, we do not repeat our explanations, so please note the respective
sections of the documentation on |lland_v1|.  Some additional tests are
covering structural differences between land-use types that only exist for
|lland_v3| and others are showing the different possible snow-related
configurations of |lland_v3|.  Furthermore, we perform some tests both on a
daily and an hourly basis to point out some differences and to make sure
|lland_v3| works well for different simulation step sizes.

Note that, while our daily simulations always cover an entire month, our
hourly simulations cover only a period three days, focussing on the "most
interesting" part of the daily simulation.  To ensure the best possible
comparability, we use the same input data (we aggregate the hourly values
to daily values), the same parameter values (possibly adapted to the change
in the time step size), and the same initial conditions (we extract the
required initial conditions of the respective hourly simulation from the
longer daily simulation).

daily simulation
________________

The following general setup is identical with the one of |lland_v1|,
except that we now start with a different period and with a daily
simulation time step:

>>> from hydpy import pub, Timegrid
>>> pub.timegrids = "1997-08-01", "1997-09-01", "1d"
>>> from hydpy.models.lland_v3 import *
>>> parameterstep("1h")
>>> from hydpy import Node, Element
>>> outlet = Node("outlet")
>>> inlet = Node("inlet")
>>> land = Element("land", inlets=inlet, outlets=outlet)
>>> land.model = model
>>> nhru(1)
>>> ft(1.0)
>>> fhru(1.0)
>>> from hydpy import IntegrationTest
>>> IntegrationTest.plotting_options.axis1 = inputs.nied
>>> IntegrationTest.plotting_options.axis2 = fluxes.qah
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

.. _lland_v3_acker_summer_daily:

acre (summer)
-------------

First, we set the values of those parameters also required by the application
model |lland_v1| (see integration test :ref:`lland_v1_acker_summer`):

>>> lnk(ACKER)
>>> kg(0.94)
>>> kt(0.0)
>>> hinz(0.2)
>>> lai.acker_aug = 3.5
>>> tgr(0.0)
>>> trefn(0.0)
>>> tsp(4.0)
>>> pwmax(1.43)
>>> wmax(309.0)
>>> fk(199.0)
>>> pwp(119.4)
>>> kapgrenz(option="0_WMax/10")
>>> kapmax(0.08)
>>> beta(0.001/24)
>>> fbeta(1.0)
>>> rbeta(False)
>>> dmax(r_dmax=0.55)
>>> dmin(r_dmin=10.0)
>>> bsf(0.3)
>>> a1(0.0)
>>> a2(0.5)
>>> tind(0.191056)
>>> eqb(35000.0)
>>> eqi1(400.0)
>>> eqi2(400.0)
>>> eqd1(200.0)
>>> eqd2(50.0)
>>> negq(False)

Next, we set the values of the parameters specific to |lland_v3|, beginning
with those required for calculating the evaporation via Penman and the
evapotranspiration via Penman-Monteith:

>>> latitude(54.1)
>>> longitude(9.7)
>>> measuringheightwindspeed(10.0)
>>> angstromconstant(0.25)
>>> angstromfactor(0.5)
>>> angstromalternative(0.15)
>>> emissivity(0.95)
>>> py(119.4)
>>> cropheight.acker_aug = 0.4
>>> albedo.acker_aug = 0.24
>>> wg2z.aug = -0.25
>>> surfaceresistance.acker = 40.0

The following parameters are related to the energy accounting snow
modelling approach after Knauf :cite:`ref-LUBW2006a`:

>>> p1strahl(0.5)
>>> p2strahl(1.0/35.0)
>>> p1wind(0.6)
>>> p2wind(1.0/70.0)
>>> turb0(0.0072)
>>> turb1(0.00576)
>>> albedo0snow(0.9)
>>> snowagingfactor(0.35)
>>> refreezeflag(True)
>>> ktschnee(inf)

Now we set the values of the parameters controlling the frost sealing
of the upper soil layer:

>>> fvf(0.3)
>>> bsff(2.0)

Due to preparing a simulation with a daily simulation step size, we only need
to define the initial values of the state sequences:

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 30.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

Finally, we prepare the input time series:

>>> inputs.nied.series = (
...     13.3, 0.1, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     2.7, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.3, 0.0, 0.1, 10.5, 6.5,
...     15.0, 9.2, 0.0, 0.0)
>>> inputs.teml.series = (
...     16.1, 17.3, 18.9, 17.8, 18.5, 19.9, 20.8, 20.9, 22.3, 23.0, 23.1,
...     23.2, 23.2, 21.7, 20.8, 19.5, 21.0, 21.6, 21.1, 22.4, 22.5, 19.6,
...     20.3, 24.1, 23.7, 21.1, 20.9, 20.2, 16.1, 17.0, 18.2)
>>> inputs.sunshineduration.series = (
...     6.3, 1.7, 4.5, 12.4, 13.9, 13.0, 13.8, 12.3, 13.1, 12.8, 13.1, 13.3,
...     12.7, 10.2, 9.4, 10.3, 11.1, 11.0, 8.5, 11.3, 12.4, 0.1, 6.7, 10.4,
...     6.5, 4.9, 6.6, 0.3, 0.1, 5.0, 3.8)
>>> inputs.windspeed.series = (
...     2.4, 1.6, 1.5, 2.8, 3.0, 3.5, 3.8, 3.3, 1.8, 1.9, 3.0, 2.5, 1.6, 2.7,
...     3.2, 1.3, 1.9, 2.0, 2.6, 3.6, 2.5, 2.1, 1.5, 2.2, 1.9, 2.3, 2.1, 2.6,
...     2.6, 2.1, 2.1)
>>> inputs.relativehumidity.series = (
...     86.2, 85.8, 82.4, 77.6, 74.9, 77.9, 73.5, 80.1, 72.9, 68.5, 68.6,
...     66.0, 69.0, 75.6, 81.4, 79.0, 75.9, 70.0, 70.3, 69.0, 62.1, 84.5,
...     83.6, 76.5, 76.6, 84.2, 85.8, 86.5, 89.6, 78.3, 78.5)
>>> inputs.atmosphericpressure.series = (
...     100.7, 101.3, 101.6, 101.7, 101.8, 101.8, 101.9, 101.9, 101.9, 102.0,
...     102.1, 102.2, 101.9, 101.5, 101.6, 101.9, 101.9, 101.8, 101.9, 101.9,
...     101.7, 101.5, 101.5, 101.1, 100.8, 100.6, 100.5, 100.0, 100.4, 101.1,
...     101.4)
>>> inlet.sequences.sim.series = 0.0

The following results illustrate the behaviour of |lland_v3| for relatively
dry and hot summer condions. Compared to the result table of integration
test :ref:`lland_v1_acker_summer` of the application model |lland_v1|,
there are much more columns, because of to the higher number of input,
flux, and state sequences.  These are mainly due to the increased data
requirements and the more complex calculations of the Penman-Monteith
equation and the Knauf :cite:`ref-LUBW2006a` approach.  Additionally,
|lland_v3| calculates many daily values (containing the averages or sums
over the last 24 hours), required for the Penman equation, which we generally
apply based on the input data of the previous 24-hours. (Note that we could
save some computation time by not calculating daily values that are never
required.  First, land-use type |ACKER| does only rely on the Penman-Monteith
equation.  Second, the values of the "normal" sequences are already daily
values, due to the selected simulation step size of one day.  Maybe we
perform some related optimisations later.):

.. integration-test::

    >>> conditions_acker_summer = test(
    ...     "lland_v3_acker_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 0.196005 | 0.002269 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.263667 | 0.003052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.003052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.140696 | 0.001628 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.075077 | 0.000869 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.040062 | 0.000464 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.021377 | 0.000247 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.011407 | 0.000132 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.006087 |  0.00007 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003248 | 0.000038 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001733 |  0.00002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000925 | 0.000011 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000494 | 0.000006 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000263 | 0.000003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000141 | 0.000002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 0.012919 |  0.00015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.017318 |   0.0002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.0002 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 0.009173 | 0.000106 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |   0.0 | 0.000106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 0.004366 | 0.000051 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |   0.0 | 0.000051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 0.000839 |  0.00001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |   0.0 |  0.00001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 0.002764 | 0.000032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |   0.0 | 0.000032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 0.031583 | 0.000366 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |   0.0 | 0.000366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 0.101111 |  0.00117 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |   0.0 |  0.00117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 0.168586 | 0.001951 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |   0.0 | 0.001951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.127614 | 0.001477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |   0.0 | 0.001477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.053743 | 0.000622 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |   0.0 | 0.000622 |

.. _lland_v3_acker_routing_daily:

:ref:`acre (routing) <lland_v1_acker_routing>`
----------------------------------------------

The following calculation shows the possibility to route inflowing runoff,
discussed in the documentation of |lland_v1|:

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test("lland_v3_acker_routing_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa)
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 1.924005 | 0.022269 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.022269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.991667 | 0.023052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.023052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.868696 | 0.021628 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.021628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.803077 | 0.020869 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.768062 | 0.020464 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.749377 | 0.020247 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.739407 | 0.020132 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.734087 |  0.02007 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.731248 | 0.020038 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.729733 |  0.02002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728925 | 0.020011 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728494 | 0.020006 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728263 | 0.020003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728141 | 0.020002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 1.740919 |  0.02015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.745318 |   0.0202 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |   0.0202 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 1.737173 | 0.020106 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |  0.02 | 0.020106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 1.732366 | 0.020051 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |  0.02 | 0.020051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 1.728839 |  0.02001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |  0.02 |  0.02001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 | 1.725699 | 0.019973 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |  0.02 | 0.019973 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 | 1.722458 | 0.019936 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |  0.02 | 0.019936 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 | 1.722219 | 0.019933 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |  0.02 | 0.019933 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 | 1.719664 | 0.019904 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |  0.02 | 0.019904 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 | 1.713886 | 0.019837 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |  0.02 | 0.019837 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 | 1.708793 | 0.019778 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |  0.02 | 0.019778 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 1.730764 | 0.020032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |  0.02 | 0.020032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 1.759583 | 0.020366 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |  0.02 | 0.020366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 1.829111 |  0.02117 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |  0.02 |  0.02117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 1.896586 | 0.021951 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |  0.02 | 0.021951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.855614 | 0.021477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |  0.02 | 0.021477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.781743 | 0.020622 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |  0.02 | 0.020622 |

.. _lland_v3_acker_heavy_rain_daily:

acre (heavy rain)
-----------------

Integration test :ref:`lland_v3_acker_summer_daily` deals with dry summer
conditions.  To show how |lland_v3| works for warm but wet conditions,
we set all values of the precipitation input time series to 20 mm.
Now, the soil water content (|BoWa|) rises from its initial value of
72 mm and nearly reaches its maximum value of 309 mm (|WMax|), resulting
in a relevant response of all runoff components:

.. integration-test::

    >>> lnk(ACKER)
    >>> inlet.sequences.sim.series = 0.0
    >>> nied = inputs.nied.series.copy()
    >>> inputs.nied.series = 20.0
    >>> conditions_acker_heavy_rain = test(
    ...     "lland_v3_acker_heavy_rain_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo | evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada |       qdb |     qib1 |     qib2 |      qbb | qkap |      qdgz |       qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |       bowa |     qdgz1 |    qdgz2 |    qigz1 |    qigz2 |     qbgz |     qdga1 |    qdga2 |    qiga1 |    qiga2 |     qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 20.0 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 18.8 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 18.1 |  0.0 | 3.623966 | 0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.217805 |      0.0 |      0.0 |      0.0 |  0.0 |  1.217805 |   0.31352 | 0.003629 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  87.161988 |  1.217805 |      0.0 |      0.0 |      0.0 |      0.0 |   0.31352 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.003629 |
    | 1997-02-08 | 20.0 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 | 18.8 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                90.407167 |               91.066405 | 18.1 |  0.0 | 2.555713 | 0.7 | 1.294549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.480868 |      0.0 |      0.0 |      0.0 |  0.0 |  1.480868 |  0.802993 | 0.009294 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  102.48657 |  1.480868 |      0.0 |      0.0 |      0.0 |      0.0 |  0.802993 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.009294 |
    | 1997-03-08 | 20.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 | 18.8 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |                70.327061 |                80.36502 | 18.1 |  0.0 | 3.286521 | 0.7 | 1.922457 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.761522 |      0.0 |      0.0 |      0.0 |  0.0 |  1.761522 |    1.1914 | 0.013789 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 116.902592 |  1.761522 |      0.0 |      0.0 |      0.0 |      0.0 |    1.1914 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.013789 |
    | 1997-04-08 | 20.0 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 | 18.8 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |                51.993585 |               67.804653 | 18.1 |  0.0 | 5.847436 | 0.7 | 3.397259 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.040913 |      0.0 |      0.0 |      0.0 |  0.0 |  2.040913 |  1.529227 | 0.017699 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  129.56442 |  2.040913 |      0.0 |      0.0 |      0.0 |      0.0 |  1.529227 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.017699 |
    | 1997-05-08 | 20.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 | 18.8 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |                 48.85611 |               65.355461 | 18.1 |  0.0 | 6.638616 | 0.7 | 3.917199 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.300279 | 0.101438 |      0.0 | 0.010164 |  0.0 |  2.300279 |  1.849046 | 0.021401 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  141.33534 |  2.300279 |      0.0 | 0.101438 |      0.0 | 0.010164 |  1.834644 |      0.0 | 0.014384 |      0.0 | 0.000018 |   0.0 | 0.021401 |
    | 1997-06-08 | 20.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 | 18.8 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |                 48.85611 |               65.425338 | 18.1 |  0.0 | 6.829116 | 0.7 | 3.913663 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.554684 | 0.110653 |      0.0 | 0.021935 |  0.0 |  2.554684 |  2.156535 |  0.02496 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 152.834404 |  2.554684 |      0.0 | 0.110653 |      0.0 | 0.021935 |  2.117307 |      0.0 | 0.039153 |      0.0 | 0.000076 |   0.0 |  0.02496 |
    | 1997-07-08 | 20.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 | 18.8 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |                 48.85611 |               65.496024 | 18.1 |  0.0 | 7.945284 | 0.7 | 4.551539 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.817223 | 0.119656 |      0.0 | 0.033434 |  0.0 |  2.817223 |  2.448759 | 0.028342 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 163.412552 |  2.817223 |      0.0 | 0.119656 |      0.0 | 0.033434 |  2.388884 |      0.0 |   0.0597 |      0.0 | 0.000175 |   0.0 | 0.028342 |
    | 1997-08-08 | 20.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 | 18.8 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |                 48.85611 |               65.567492 | 18.1 |  0.0 | 6.428125 | 0.7 | 3.767989 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.072608 | 0.127938 |      0.0 | 0.044013 |  0.0 |  3.072608 |   2.73175 | 0.031617 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 174.500005 |  3.072608 |      0.0 | 0.127938 |      0.0 | 0.044013 |  2.654405 |      0.0 | 0.077033 |      0.0 | 0.000313 |   0.0 | 0.031617 |
    | 1997-09-08 | 20.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 | 18.8 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |                 48.85611 |               65.639719 | 18.1 |  0.0 | 6.036736 | 0.7 | 4.121081 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.356591 | 0.136618 |      0.0 |   0.0551 |  0.0 |  3.356591 |  3.015033 | 0.034896 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 184.930615 |  3.356591 |      0.0 | 0.136618 |      0.0 |   0.0551 |   2.92256 |      0.0 | 0.091983 |      0.0 | 0.000489 |   0.0 | 0.034896 |
    | 1997-10-08 | 20.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 | 18.8 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |                 48.85611 |               65.712681 | 18.1 |  0.0 | 6.494046 | 0.7 | 4.450439 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.641259 | 0.144785 |      0.0 | 0.065531 |  0.0 |  3.641259 |  3.304149 | 0.038242 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 194.728603 |  3.641259 |      0.0 | 0.144785 |      0.0 | 0.065531 |  3.198274 |      0.0 | 0.105171 |      0.0 | 0.000703 |   0.0 | 0.038242 |
    | 1997-11-08 | 20.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 | 18.8 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |                 48.85611 |               65.786357 | 18.1 |  0.0 | 7.914762 | 0.7 | 4.993103 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.926571 | 0.152455 |      0.0 | 0.075329 |  0.0 |  3.926571 |  3.596218 | 0.041623 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 203.681145 |  3.926571 |      0.0 | 0.152455 |      0.0 | 0.075329 |   3.47833 |      0.0 | 0.116935 |      0.0 | 0.000953 |   0.0 | 0.041623 |
    | 1997-12-08 | 20.0 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 | 18.8 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |                 48.85611 |               65.860724 | 18.1 |  0.0 | 7.566718 | 0.7 | 4.979876 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.204866 | 0.159465 | 0.009557 | 0.084281 |  0.0 |  4.204866 |  3.889124 | 0.045013 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |   212.3431 |  4.204866 |      0.0 | 0.159465 | 0.009557 | 0.084281 |  3.759031 |      0.0 | 0.127502 | 0.001355 | 0.001236 |   0.0 | 0.045013 |
    | 1997-13-08 | 20.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 | 18.8 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |                 48.85611 |               65.935761 | 18.1 |  0.0 | 5.996364 | 0.7 |  4.19851 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.492755 | 0.166246 | 0.045992 | 0.092943 |  0.0 |  4.492755 |  4.188438 | 0.048477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 221.446655 |  4.492755 |      0.0 | 0.166246 | 0.045992 | 0.092943 |  4.041079 |      0.0 | 0.137078 | 0.008732 | 0.001549 |   0.0 | 0.048477 |
    | 1997-14-08 | 20.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 | 18.8 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |                 48.85611 |               66.011448 | 18.1 |  0.0 | 5.983925 | 0.7 | 3.702096 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.818552 | 0.173373 | 0.100351 | 0.102047 |  0.0 |  4.818552 |  4.509936 | 0.052198 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 230.650235 |  4.818552 |      0.0 | 0.173373 | 0.100351 | 0.102047 |  4.335611 |      0.0 |  0.14595 | 0.026482 | 0.001893 |   0.0 | 0.052198 |
    | 1997-15-08 | 20.0 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 | 18.8 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |                 48.85611 |               66.087764 | 18.1 |  0.0 | 5.470984 | 0.7 | 3.155298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.176929 | 0.180579 |  0.16802 |  0.11125 |  0.0 |  5.176929 |  4.865729 | 0.056316 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 239.958158 |  5.176929 |      0.0 | 0.180579 |  0.16802 |  0.11125 |  4.653113 |      0.0 | 0.154363 | 0.055986 | 0.002268 |   0.0 | 0.056316 |
    | 1997-16-08 | 20.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 | 18.8 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |                 48.85611 |                66.16469 | 18.1 |  0.0 |  4.08134 | 0.7 | 2.708461 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.575829 | 0.187866 | 0.247347 | 0.120558 |  0.0 |  5.575829 |  5.262676 | 0.060911 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 249.218097 |  5.575829 |      0.0 | 0.187866 | 0.247347 | 0.120558 |   5.00011 |      0.0 | 0.162462 | 0.097429 | 0.002675 |   0.0 | 0.060911 |
    | 1997-17-08 | 20.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 | 18.8 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |                 48.85611 |               66.242207 | 18.1 |  0.0 | 5.175424 | 0.7 | 3.378706 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.018699 | 0.195116 | 0.335804 | 0.129818 |  0.0 |  6.018699 |  5.706462 | 0.066047 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 257.259955 |  6.018699 |      0.0 | 0.195116 | 0.335804 | 0.129818 |  5.382634 |      0.0 | 0.170336 | 0.150377 | 0.003114 |   0.0 | 0.066047 |
    | 1997-18-08 | 20.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 | 18.8 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |                 48.85611 |               66.320298 | 18.1 |  0.0 | 5.738005 | 0.7 | 3.779831 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.451503 | 0.201412 | 0.419614 |  0.13786 |  0.0 |  6.451503 |  6.184438 | 0.071579 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 264.369735 |  6.451503 |      0.0 | 0.201412 | 0.419614 |  0.13786 |  5.790712 |      0.0 | 0.177908 | 0.212236 | 0.003582 |   0.0 | 0.071579 |
    | 1997-19-08 | 20.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 | 18.8 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                 48.85611 |               66.398945 | 18.1 |  0.0 | 5.867083 | 0.7 |  3.62847 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.883173 | 0.206978 | 0.498724 |  0.14497 |  0.0 |  6.883173 |  6.678482 | 0.077297 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 271.107419 |  6.883173 |      0.0 | 0.206978 | 0.498724 |  0.14497 |  6.210029 |      0.0 | 0.185032 | 0.279345 | 0.004076 |   0.0 | 0.077297 |
    | 1997-20-08 | 20.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 | 18.8 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |                 48.85611 |                66.47813 | 18.1 |  0.0 | 7.829385 | 0.7 | 4.625035 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.348865 | 0.212253 | 0.577784 | 0.151707 |  0.0 |  7.348865 |  7.189836 | 0.083216 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 276.291774 |  7.348865 |      0.0 | 0.212253 | 0.577784 | 0.151707 |  6.643867 |      0.0 | 0.191695 | 0.349681 | 0.004593 |   0.0 | 0.083216 |
    | 1997-21-08 | 20.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 | 18.8 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |                 48.85611 |               66.557838 | 18.1 |  0.0 |  7.32284 | 0.7 | 4.757012 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.757458 | 0.216312 | 0.641202 | 0.156892 |  0.0 |  7.757458 |  7.700952 | 0.089131 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 280.862897 |  7.757458 |      0.0 | 0.216312 | 0.641202 | 0.156892 |  7.077861 |      0.0 | 0.197811 |  0.42015 | 0.005129 |   0.0 | 0.089131 |
    | 1997-22-08 | 20.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 | 18.8 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |                 48.85611 |               66.638052 | 18.1 |  0.0 |  2.56747 | 0.7 | 1.361228 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.166601 | 0.219891 | 0.698918 | 0.161463 |  0.0 |  8.166601 |  8.197047 | 0.094873 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 288.354796 |  8.166601 |      0.0 | 0.219891 | 0.698918 | 0.161463 |   7.50015 |      0.0 | 0.203305 | 0.487911 | 0.005681 |   0.0 | 0.094873 |
    | 1997-23-08 | 20.0 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 | 18.8 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |                 48.85611 |               66.718758 | 18.1 |  0.0 | 3.374286 | 0.7 | 2.097563 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.984192 | 0.225757 | 0.797025 | 0.168955 |  0.0 |  8.984192 |  8.795013 | 0.101794 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 294.181304 |  8.984192 |      0.0 | 0.225757 | 0.797025 | 0.168955 |  8.021462 |      0.0 | 0.208607 | 0.558691 | 0.006253 |   0.0 | 0.101794 |
    | 1997-24-08 | 20.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 | 18.8 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |                 48.85611 |                66.79994 | 18.1 |  0.0 | 5.573178 | 0.7 | 3.649927 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  9.834859 | 0.230318 | 0.876239 | 0.174781 |  0.0 |  9.834859 |   9.54435 | 0.110467 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 297.515179 |  9.834859 |      0.0 | 0.230318 | 0.876239 | 0.174781 |   8.68947 |      0.0 | 0.213876 | 0.634158 | 0.006846 |   0.0 | 0.110467 |
    | 1997-25-08 | 20.0 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 | 18.8 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |                 48.85611 |               66.881585 | 18.1 |  0.0 | 4.467819 | 0.7 | 2.895981 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 10.482817 | 0.232928 | 0.922678 | 0.178115 |  0.0 | 10.482817 | 10.322599 | 0.119475 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  300.90266 | 10.482817 |      0.0 | 0.232928 | 0.922678 | 0.178115 |  9.390481 |      0.0 | 0.218677 | 0.705987 | 0.007454 |   0.0 | 0.119475 |
    | 1997-26-08 | 20.0 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 | 18.8 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |                 48.85611 |               66.963678 | 18.1 |  0.0 | 3.569973 | 0.7 | 2.071555 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  11.39404 |  0.23558 | 0.970674 | 0.181503 |  0.0 |  11.39404 | 11.136685 | 0.128897 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 304.149308 |  11.39404 |      0.0 |  0.23558 | 0.970674 | 0.181503 | 10.134526 |      0.0 | 0.222894 | 0.771194 | 0.008071 |   0.0 | 0.128897 |
    | 1997-27-08 | 20.0 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 | 18.8 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |                 48.85611 |               67.046207 | 18.1 |  0.0 | 3.567075 | 0.7 | 2.108966 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 13.249308 | 0.238122 | 1.017431 | 0.184749 |  0.0 | 13.249308 | 12.737269 | 0.147422 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 305.450732 |      12.0 | 1.249308 | 0.238122 | 1.017431 | 0.184749 | 10.877952 | 0.792359 | 0.226674 | 0.831586 | 0.008698 |   0.0 | 0.147422 |
    | 1997-28-08 | 20.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 | 18.8 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                 48.85611 |               67.129158 | 18.1 |  0.0 | 2.621374 | 0.7 | 1.331712 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.550732 | 0.239141 | 1.036378 | 0.186051 |  0.0 | 14.550732 | 14.562531 | 0.168548 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 306.206718 |      12.0 | 2.550732 | 0.239141 | 1.036378 | 0.186051 | 11.401263 | 2.037672 | 0.229904 | 0.884361 | 0.009331 |   0.0 | 0.168548 |
    | 1997-29-08 | 20.0 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 | 18.8 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 48.85611 |               67.212519 | 18.1 |  0.0 | 2.027722 | 0.7 | 0.881862 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 15.306718 | 0.239733 | 1.047438 | 0.186807 |  0.0 | 15.306718 |  15.83846 | 0.183316 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  306.64416 |      12.0 | 3.306718 | 0.239733 | 1.047438 | 0.186807 | 11.680507 |  2.98861 | 0.232477 |   0.9269 | 0.009966 |   0.0 | 0.183316 |
    | 1997-30-08 | 20.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 | 18.8 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |                 48.85611 |               67.296277 | 18.1 |  0.0 | 3.393609 | 0.7 | 1.913447 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  15.74416 | 0.240076 | 1.053855 | 0.187244 |  0.0 |  15.74416 | 16.593262 | 0.192052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 305.605378 |      12.0 |  3.74416 | 0.240076 | 1.053855 | 0.187244 | 11.829515 | 3.558369 | 0.234481 | 0.960296 |   0.0106 |   0.0 | 0.192052 |
    | 1997-31-08 | 20.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 | 18.8 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |                 48.85611 |               67.380423 | 18.1 |  0.0 | 3.322101 | 0.7 | 1.882222 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.705378 | 0.239262 | 1.038637 | 0.186205 |  0.0 | 14.705378 | 16.209748 | 0.187613 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 305.653673 |      12.0 | 2.705378 | 0.239262 | 1.038637 | 0.186205 | 11.909027 | 3.070262 | 0.235874 | 0.983354 | 0.011231 |   0.0 | 0.187613 |

>>> inputs.nied.series = nied

.. _lland_v3_wasser_daily:

water
-----

L-Land defines three types of water areas, |WASSER|, |FLUSS| and |SEE|.
|lland_v3| calculates their evaporation via the Penman equation based on
daily input data.  Besides that, there are no structural differences to
|lland_v1|,  so please read the detailed discussions in integration test
:ref:`lland_v1_wasser` for better understanding the following test
results for water type |WASSER|:

.. integration-test::

    >>> lnk(WASSER)
    >>> surfaceresistance.wasser_aug = 0.0
    >>> cropheight.wasser_aug = 0.05
    >>> albedo.wasser_aug = 0.7
    >>> conditions_wasser = test(
    ...     "lland_v3_wasser_daily",
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 | 0.132232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 | 0.011672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 | 0.012428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |    0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |  0.10177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 | 0.058014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 | 0.153995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 | 0.093058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

.. _lland_v3_wasser_routing_daily:

water (routing)
---------------

The following calculation shows the possibility to subtract evaporation
from inflowing runoff, discussed in the integration test
:ref:`lland_v1_wasser_routing` of |lland_v1|:

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test("lland_v3_wasser_routing_daily",
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.152803 | 0.152232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.152232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.95655 | 0.011071 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.011071 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.635635 | 0.007357 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.007357 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.256106 | 0.002964 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002964 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |    1.822 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055264 |  0.00064 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00064 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.736458 | 0.031672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.031672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.444099 |  0.00514 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00514 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.217257 | 0.002515 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002515 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.133428 | 0.001544 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.001544 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.186433 | 0.002158 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002158 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.801812 | 0.032428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.032428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.92579 | 0.010715 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.010715 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055768 | 0.000645 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.000645 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.447898 | 0.005184 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.005184 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 10.520904 |  0.12177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.12177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   6.74044 | 0.078014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.078014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 15.033141 | 0.173995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.173995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  9.768224 | 0.113058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.113058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.82735 | 0.009576 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009576 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.815922 | 0.009444 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009444 |

.. _lland_v3_wasser_negq_daily:

water (negative runoff)
-----------------------

The following calculation shows the possibility to calculate negative
discharge values, discussed in the integration test :ref:`lland_v1_wasser_negq`
of |lland_v1|:

.. integration-test::

    >>> negq(True)
    >>> inlet.sequences.sim.series = 0.0
    >>> test("lland_v3_wasser_negq_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |        qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |    outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 |  0.132232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.77145 | -0.008929 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.008929 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.092365 | -0.012643 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.012643 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.471894 | -0.017036 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017036 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.729733 |  -0.02002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.02002 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826087 | -0.021135 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021135 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.041956 | -0.023634 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.023634 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.800053 | -0.020834 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020834 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |  1.82186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -1.82186 | -0.021086 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021086 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 | 1.891469 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.891469 | -0.021892 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021892 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 | 2.106575 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.106575 | -0.024382 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.024382 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 | 2.036929 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.942929 | -0.022488 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022488 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | 1.808559 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.808559 | -0.020932 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020932 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672736 |  -0.01936 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01936 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 |  0.011672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.283901 |  -0.01486 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01486 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.510743 | -0.017485 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017485 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.594572 | -0.018456 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.018456 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.541567 | -0.017842 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017842 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 | 1.939383 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.939383 | -0.022447 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022447 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 | 1.826811 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826811 | -0.021144 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021144 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 |  0.012428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.80221 | -0.009285 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.009285 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672232 | -0.019355 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.019355 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.280102 | -0.014816 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.014816 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |   0.10177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 |  0.058014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 |  0.153995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 |  0.093058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.90065 | -0.010424 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010424 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -0.912078 | -0.010556 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010556 |

.. _lland_v3_see_daily:

lakes
-----

The following example focusses on water-type |SEE| (for further information,
see integration test :ref:`lland_v1_see` of |lland_v1|):

.. integration-test::

    >>> lnk(SEE)
    >>> surfaceresistance.see_aug = 0.0
    >>> cropheight.see_aug = 0.05
    >>> albedo.see_aug = 0.7
    >>> negq(False)
    >>> test("lland_v3_see_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz | qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.020478 | 0.000237 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 11.424803 |   0.0 |   0.0 |   0.0 |   0.0 |  0.020478 |   0.0 | 0.000237 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039475 | 0.000457 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.77145 |   0.0 |   0.0 |   0.0 |   0.0 |  0.039475 |   0.0 | 0.000457 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035994 | 0.000417 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.092365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035994 |   0.0 | 0.000417 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.031272 | 0.000362 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.471894 |   0.0 |   0.0 |   0.0 |   0.0 |  0.031272 |   0.0 | 0.000362 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.025424 | 0.000294 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.729733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.025424 |   0.0 | 0.000294 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.018963 | 0.000219 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826087 |   0.0 |   0.0 |   0.0 |   0.0 |  0.018963 |   0.0 | 0.000219 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011966 | 0.000138 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.041956 |   0.0 |   0.0 |   0.0 |   0.0 |  0.011966 |   0.0 | 0.000138 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005041 | 0.000058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.800053 |   0.0 |   0.0 |   0.0 |   0.0 |  0.005041 |   0.0 | 0.000058 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 | 1.820396 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -1.82186 |   0.0 |   0.0 |   0.0 |   0.0 | -0.001465 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 | 1.883357 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.891469 |   0.0 |   0.0 |   0.0 |   0.0 | -0.008111 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 | 2.091331 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.106575 |   0.0 |   0.0 |   0.0 |   0.0 | -0.015244 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 | 2.014486 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.942929 |   0.0 |   0.0 |   0.0 |   0.0 | -0.022443 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | 1.779476 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.808559 |   0.0 |   0.0 |   0.0 |   0.0 | -0.029083 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.637521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672736 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035215 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.493266 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.008458 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036276 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.247259 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.283901 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036642 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.469226 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.510743 |   0.0 |   0.0 |   0.0 |   0.0 | -0.041517 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.547641 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.594572 |   0.0 |   0.0 |   0.0 |   0.0 | -0.046931 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.489186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.541567 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.05238 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 | 1.880954 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.939383 |   0.0 |   0.0 |   0.0 |   0.0 | -0.058428 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 | 1.761846 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826811 |   0.0 |   0.0 |   0.0 |   0.0 | -0.064965 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.740109 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.073812 |   0.0 |   0.0 |   0.0 |   0.0 | -0.066078 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 | 1.018853 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.80221 |   0.0 |   0.0 |   0.0 |   0.0 | -0.065357 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.602675 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672232 |   0.0 |   0.0 |   0.0 |   0.0 | -0.069557 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.299507 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.280102 |   0.0 |   0.0 |   0.0 |   0.0 | -0.074595 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.016237 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.792904 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.06086 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 | 1.061645 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |   5.01244 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035916 |   0.0 |      0.0 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.791893 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 13.305141 |   0.0 |   0.0 |   0.0 |   0.0 | -0.002966 |   0.0 |      0.0 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035276 | 0.000408 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.040224 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035276 |   0.0 | 0.000408 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047929 | 0.000555 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.90065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.047929 |   0.0 | 0.000555 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.04451 | 0.000515 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -0.912078 |   0.0 |   0.0 |   0.0 |   0.0 |   0.04451 |   0.0 | 0.000515 |

.. _lland_v3_fluss_daily:

streams
-------

The following example focusses on water-type |FLUSS| (for further information,
see integration test :ref:`lland_v1_fluss` of |lland_v1|):

.. integration-test::

    >>> lnk(FLUSS)
    >>> surfaceresistance.fluss_aug = 0.0
    >>> cropheight.fluss_aug = 0.05
    >>> albedo.fluss_aug = 0.7
    >>> test("lland_v3_fluss_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |       evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa |     qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |     qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 |  1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 11.424803 | 2.941276 | 0.034043 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | 11.424803 |      0.0 |   0.0 |   0.0 |  0.0 |  2.941276 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034043 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |   0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.77145 | 3.758016 | 0.043496 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.77145 |      0.0 |   0.0 |   0.0 |  0.0 |  3.758016 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043496 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 |  1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.092365 | 1.562905 | 0.018089 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.092365 |      0.0 |   0.0 |   0.0 |  0.0 |  1.562905 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018089 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 |  1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.471894 | 0.226808 | 0.002625 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.471894 |      0.0 |   0.0 |   0.0 |  0.0 |  0.226808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002625 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |  1.097905 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.729733 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.729733 |      0.0 |   0.0 |   0.0 |  0.0 | -0.631827 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |  0.657403 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826087 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826087 |      0.0 |   0.0 |   0.0 |  0.0 | -1.168684 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |  0.511091 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.041956 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.041956 |      0.0 |   0.0 |   0.0 |  0.0 | -1.530864 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |  0.093098 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.800053 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.800053 |      0.0 |   0.0 |   0.0 |  0.0 | -1.706955 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |  0.065871 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -1.82186 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -1.82186 |      0.0 |   0.0 |   0.0 |  0.0 | -1.755989 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |  0.086838 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.891469 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.891469 |      0.0 |   0.0 |   0.0 |  0.0 | -1.804631 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |  0.206065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.106575 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.106575 |      0.0 |   0.0 |   0.0 |  0.0 | -1.900509 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |  0.082443 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.942929 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.942929 |      0.0 |   0.0 |   0.0 |  0.0 | -1.954486 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | -0.105944 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.808559 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.808559 |      0.0 |   0.0 |   0.0 |  0.0 | -1.914503 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | -0.157388 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672736 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672736 |      0.0 |   0.0 |   0.0 |  0.0 | -1.830125 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 |  0.463086 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.008458 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.008458 |      0.0 |   0.0 |   0.0 |  0.0 | -1.066456 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 |  0.595002 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.283901 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.283901 |      0.0 |   0.0 |   0.0 |  0.0 | -0.688899 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 |  0.485942 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.510743 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.510743 |      0.0 |   0.0 |   0.0 |  0.0 | -1.024801 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 |  0.321551 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.594572 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.594572 |      0.0 |   0.0 |   0.0 |  0.0 |  -1.27302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 |  0.132224 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.541567 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.541567 |      0.0 |   0.0 |   0.0 |  0.0 | -1.409342 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |  0.365956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.939383 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.939383 |      0.0 |   0.0 |   0.0 |  0.0 | -1.573427 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |  0.111688 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826811 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826811 |      0.0 |   0.0 |   0.0 |  0.0 | -1.715123 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 |  -0.21427 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.073812 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.073812 |      0.0 |   0.0 |   0.0 |  0.0 | -1.020458 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  0.557521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.80221 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.80221 |      0.0 |   0.0 |   0.0 |  0.0 | -0.526689 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 |  0.793059 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672232 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672232 |      0.0 |   0.0 |   0.0 |  0.0 | -0.879173 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 |  0.226008 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.280102 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.280102 |      0.0 |   0.0 |   0.0 |  0.0 | -1.148095 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 |  1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.792904 |   1.3836 | 0.016014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.792904 |      0.0 |   0.0 |   0.0 |  0.0 |    1.3836 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016014 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |   1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |   5.01244 |  3.86595 | 0.044745 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   5.01244 |      0.0 |   0.0 |   0.0 |  0.0 |   3.86595 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044745 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 |  0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 13.305141 | 7.027354 | 0.081335 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |      12.0 | 1.305141 |   0.0 |   0.0 |  0.0 |  6.199583 |  0.82777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081335 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 |  0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.040224 |  8.32407 | 0.096343 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.040224 |      0.0 |   0.0 |   0.0 |  0.0 |  7.885403 | 0.438667 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096343 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |   0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.90065 | 5.691378 | 0.065872 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.90065 |      0.0 |   0.0 |   0.0 |  0.0 |  5.655812 | 0.035566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065872 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 |  0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -0.912078 | 2.597892 | 0.030068 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -0.912078 |      0.0 |   0.0 |   0.0 |  0.0 |  2.595008 | 0.002884 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030068 |

.. _lland_v3_vers_daily:

sealed surfaces
---------------

As also shown in the integration test :ref:`lland_v1_vers` of |lland_v1|,
sealed surfaces route the not intercepted water to the linear storages for
direct discharge immediately:

.. integration-test::

    >>> lnk(VERS)
    >>> surfaceresistance.vers_aug = 500
    >>> cropheight.vers_aug = 5.0
    >>> albedo.vers_aug = 0.10
    >>> lai.vers_aug = 10.0
    >>> test("lland_v3_vers_daily",
    ...      axis1=(inputs.nied, fluxes.qah, states.bowa))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |      evpo |   evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |    qdb | qib1 | qib2 | qbb | qkap |   qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn | bowa |  qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.793956 |                 14.793956 |                  14.793956 |                  1.877708 |                      0.0 |    12.916248 |              0.0 |         12.916248 |             16.038785 |                   nan |                    500.0 |                   500.0 | 10.502 |  0.0 |  7.049585 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 10.502 | 10.502 |  0.0 |  0.0 | 0.0 |  0.0 | 10.502 | 2.703704 | 0.031293 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  0.0 | 10.502 |   0.0 |   0.0 |   0.0 |  0.0 | 2.703704 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031293 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.926902 |                  9.926902 |                   9.926902 |                   0.96939 |                      0.0 |     8.957512 |              0.0 |          8.957512 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  5.020895 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 3.637039 | 0.042095 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 3.637039 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042095 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.783845 |                 12.783845 |                  12.783845 |                  1.386217 |                      0.0 |    11.397628 |              0.0 |         11.397628 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  6.335547 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.940764 | 0.022463 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.940764 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022463 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.919034 |                 20.919034 |                  20.919034 |                  3.108707 |                      0.0 |    17.810327 |              0.0 |         17.810327 |              13.74753 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 12.230738 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.035613 | 0.011986 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.035613 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011986 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             22.383112 |                 22.383112 |                  22.383112 |                  3.414868 |                      0.0 |    18.968244 |              0.0 |         18.968244 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 14.271962 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.552615 | 0.006396 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.552615 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006396 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             21.358671 |                 21.358671 |                  21.358671 |                  2.845741 |                      0.0 |     18.51293 |              0.0 |          18.51293 |             10.998024 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 14.791696 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.294881 | 0.003413 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.294881 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003413 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             22.087564 |                 22.087564 |                  22.087564 |                  3.051087 |                      0.0 |    19.036477 |              0.0 |         19.036477 |             10.129759 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 18.203529 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.157352 | 0.001821 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157352 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001821 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             20.448645 |                 20.448645 |                  20.448645 |                   2.45312 |                      0.0 |    17.995525 |              0.0 |         17.995525 |             11.664571 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.456812 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.083965 | 0.000972 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083965 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000972 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             21.170659 |                 21.170659 |                  21.170659 |                  2.688326 |                      0.0 |    18.482332 |              0.0 |         18.482332 |             21.385046 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  11.84053 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.044804 | 0.000519 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044804 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000519 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.763414 |                 20.763414 |                  20.763414 |                   2.76572 |                      0.0 |    17.997695 |              0.0 |         17.997695 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.388795 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.023908 | 0.000277 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.023908 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000277 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |              20.96616 |                  20.96616 |                   20.96616 |                  2.799361 |                      0.0 |    18.166799 |              0.0 |         18.166799 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 18.009869 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.012758 | 0.000148 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.012758 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000148 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             21.063496 |                 21.063496 |                  21.063496 |                  2.986484 |                      0.0 |    18.077012 |              0.0 |         18.077012 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 16.851837 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.006808 | 0.000079 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000079 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             20.349571 |                 20.349571 |                  20.349571 |                  2.708378 |                      0.0 |    17.641193 |              0.0 |         17.641193 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 11.954049 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.003633 | 0.000042 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003633 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000042 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.726334 |                 17.726334 |                  17.726334 |                  2.239353 |                      0.0 |    15.486981 |              0.0 |         15.486981 |             14.256697 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.005903 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001938 | 0.000022 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001938 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000022 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.823868 |                 16.823868 |                  16.823868 |                  2.021579 |                      0.0 |    14.802289 |              0.0 |         14.802289 |             12.029088 |                   nan |                    500.0 |                   500.0 |  0.538 |  0.0 | 11.728768 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.538 |  0.538 |  0.0 |  0.0 | 0.0 |  0.0 |  0.538 | 0.139541 | 0.001615 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 |  0.0 |  0.538 |   0.0 |   0.0 |   0.0 |  0.0 | 0.139541 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001615 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.623363 |                 17.623363 |                  17.623363 |                  2.487815 |                      0.0 |    15.135549 |              0.0 |         15.135549 |             29.610064 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |   7.56081 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.186871 | 0.002163 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.186871 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002163 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             18.314696 |                 18.314696 |                  18.314696 |                  2.520953 |                      0.0 |    15.793743 |              0.0 |         15.793743 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  10.39177 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.099717 | 0.001154 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.099717 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001154 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             18.106372 |                 18.106372 |                  18.106372 |                  2.702703 |                      0.0 |    15.403668 |              0.0 |         15.403668 |             19.246541 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 12.298568 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 |  0.05321 | 0.000616 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05321 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000616 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.529721 |                 15.529721 |                  15.529721 |                  2.307856 |                      0.0 |    13.221865 |              0.0 |         13.221865 |             14.805032 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.759102 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.028393 | 0.000329 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.028393 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000329 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             18.178103 |                 18.178103 |                  18.178103 |                  2.688476 |                      0.0 |    15.489627 |              0.0 |         15.489627 |             10.692523 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 19.233472 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.015151 | 0.000175 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.015151 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000175 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             19.138391 |                 19.138391 |                  19.138391 |                  3.308913 |                      0.0 |    15.829478 |              0.0 |         15.829478 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  17.26069 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.008085 | 0.000094 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.008085 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000094 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              7.043929 |                  7.043929 |                   7.043929 |                   0.60663 |                      0.0 |     6.437299 |              0.0 |          6.437299 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  5.830469 |  1.88 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.004314 |  0.00005 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004314 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00005 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.370443 |                 13.370443 |                  13.370443 |                  1.641158 |                      0.0 |    11.729284 |              0.0 |         11.729284 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  6.371493 | 0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.002302 | 0.000027 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000027 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.838913 |                 16.838913 |                  16.838913 |                  1.868809 |                      0.0 |    14.970104 |              0.0 |         14.970104 |             17.496856 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 11.504016 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001228 | 0.000014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001228 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000014 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             12.979453 |                 12.979453 |                  12.979453 |                  1.401603 |                      0.0 |     11.57785 |              0.0 |          11.57785 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  9.461825 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.000655 | 0.000008 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000655 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000008 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.351841 |                 11.351841 |                  11.351841 |                  1.261689 |                      0.0 |    10.090152 |              0.0 |         10.090152 |             16.736123 |                   nan |                    500.0 |                   500.0 |   7.87 |  0.0 |  7.472563 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   7.87 |   7.87 |  0.0 |  0.0 | 0.0 |  0.0 |   7.87 | 2.026454 | 0.023454 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 |  0.0 |   7.87 |   0.0 |   0.0 |   0.0 |  0.0 | 2.026454 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023454 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.872693 |                 12.872693 |                  12.872693 |                  1.498964 |                      0.0 |    11.373729 |              0.0 |         11.373729 |             18.330039 |                   nan |                    500.0 |                   500.0 |   4.11 |  0.0 |  6.960765 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   4.11 |   4.11 |  0.0 |  0.0 | 0.0 |  0.0 |   4.11 |  3.78382 | 0.043794 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 |  0.0 |   4.11 |   0.0 |   0.0 |   0.0 |  0.0 |  3.78382 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043794 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              6.819035 |                  6.819035 |                   6.819035 |                  0.589077 |                      0.0 |     6.229957 |              0.0 |          6.229957 |             14.805032 |                   nan |                    500.0 |                   500.0 |   12.1 |  0.0 |  6.075158 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   12.1 |   12.1 |  0.0 |  0.0 | 0.0 |  0.0 |   12.1 | 6.030624 | 0.069799 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 |  0.0 |   12.0 |   0.1 |   0.0 |   0.0 |  0.0 | 5.967201 | 0.063424 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069799 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.559591 |                  6.559591 |                   6.559591 |                  0.708412 |                      0.0 |     5.851179 |              0.0 |          5.851179 |             14.805032 |                   nan |                    500.0 |                   500.0 |  6.648 |  0.0 |  4.534236 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  6.648 |  6.648 |  0.0 |  0.0 | 0.0 |  0.0 |  6.648 | 7.436588 | 0.086072 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  0.0 |  6.648 |   0.0 |   0.0 |   0.0 |  0.0 | 7.402977 | 0.033611 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086072 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             11.065964 |                 11.065964 |                  11.065964 |                  1.881434 |                      0.0 |      9.18453 |              0.0 |           9.18453 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  7.752367 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 5.342084 |  0.06183 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 5.339359 | 0.002725 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06183 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.853909 |                  9.853909 |                   9.853909 |                  1.511299 |                      0.0 |      8.34261 |              0.0 |           8.34261 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  7.711096 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 2.849362 | 0.032979 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849141 | 0.000221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032979 |

.. _lland_v3_nadelw_summer_daily:

conifers (summer)
-----------------

The following tests focus on functionalities of |lland_v3| not implemented
by application model |lland_v1|.  We start with the land-use type |NADELW|.
For |lland_v1|, there is no structural difference between |NADELW| and,
for example, |ACKER|.  Different simulation results only are based solely
on different parameterisations.  In contrast, |lland_v3| handles |NADELW|
unlike all other land use types (even unlike |LAUBW| and |MISCHW|) regarding
the calculation of the surface resistance, relevant for the Penman-Monteith
equation (see method |Calc_LanduseSurfaceResistance_V1| for more details).
The calculated soil evapotranspiration (|EvB|) is similar to the one
calculated with the same input data for land-use type |ACKER| in the
integration test :ref:`lland_v3_acker_summer_daily`:

.. integration-test::

    >>> lnk(NADELW)
    >>> surfaceresistance.nadelw = 56.0
    >>> cropheight.nadelw = 10.0
    >>> albedo.nadelw = 0.12
    >>> lai.nadelw = 11.0
    >>> conditions_nadelw_summer = test(
    ...     "lland_v3_nadelw_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah, states.bowa),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |      evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.465201 |                  2.686395 |                  14.465201 |                  1.877708 |                      0.0 |    12.587493 |              0.0 |         12.587493 |             39.166667 |                 100.0 |               211.364559 |              122.747678 | 10.302 |  0.0 |  4.686428 |   2.2 | 1.110404 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 10.302 | 0.657815 |  0.0 |  0.0 | 0.0 |      0.0 | 0.657815 | 0.169352 |  0.00196 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 | 80.533782 | 0.657815 |   0.0 |   0.0 |   0.0 |       0.0 | 0.169352 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00196 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.706305 |                  1.802599 |                   9.706305 |                   0.96939 |                      0.0 |     8.736914 |              0.0 |          8.736914 |                 58.75 |                 100.0 |               181.457925 |               115.35055 |    0.0 |  0.0 |  3.312002 | 0.094 | 1.809101 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.227814 | 0.002637 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 78.724681 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.227814 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002637 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.499759 |                  2.321384 |                  12.499759 |                  1.386217 |                      0.0 |    11.113542 |              0.0 |         11.113542 |             62.666667 |                 100.0 |               184.734569 |              115.873944 |    0.0 |  0.0 |   4.24805 |   0.0 | 2.495048 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.121564 | 0.001407 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 76.229633 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001407 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.454167 |                  3.798631 |                  20.454167 |                  3.108707 |                      0.0 |     17.34546 |              0.0 |          17.34546 |             33.571429 |                 100.0 |               209.622626 |              121.261768 |    0.0 |  0.0 |  7.655528 | 0.094 |  3.24309 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.064868 | 0.000751 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 72.986543 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000751 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |              21.88571 |                  4.064489 |                   21.88571 |                  3.414868 |                      0.0 |    18.470842 |              0.0 |         18.470842 |             31.333333 |                 100.0 |               223.524712 |              123.686516 |    0.0 |  0.0 |  8.731502 |   0.0 | 3.618105 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.034614 | 0.000401 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 69.368437 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000401 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           1.242466 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             20.884034 |                  3.878463 |                  20.884034 |                  2.845741 |                      0.0 |    18.038293 |              0.0 |         18.038293 |             26.857143 |                 100.0 |               217.806222 |              122.172235 |    0.0 |  0.0 |  8.967691 |   0.0 | 3.538589 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.018471 | 0.000214 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 65.829848 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.018471 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000214 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             21.596729 |                  4.010821 |                  21.596729 |                  3.051087 |                      0.0 |    18.545642 |              0.0 |         18.545642 |             24.736842 |                 100.0 |                249.89699 |              127.503607 |    0.0 |  0.0 | 10.546527 |   0.0 | 3.943139 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.009856 | 0.000114 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 61.886709 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.009856 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000114 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           1.171468 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             19.994231 |                  3.713214 |                  19.994231 |                   2.45312 |                      0.0 |    17.541111 |              0.0 |         17.541111 |             28.484848 |                 100.0 |               234.560208 |              124.470702 |    0.0 |  0.0 |  8.375029 |   0.0 | 3.447428 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.005259 | 0.000061 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.439281 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.005259 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000061 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |               20.7002 |                  3.844323 |                    20.7002 |                  2.688326 |                      0.0 |    18.011873 |              0.0 |         18.011873 |             52.222222 |                 100.0 |                290.49321 |              132.398807 |    0.0 |  0.0 |  7.801371 |   0.0 | 4.238315 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.002806 | 0.000032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.200966 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002806 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000032 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.302005 |                  3.770372 |                  20.302005 |                   2.76572 |                      0.0 |    17.536286 |              0.0 |         17.536286 |             49.473684 |                 100.0 |               348.466656 |              138.318012 |    0.0 |  0.0 |  8.457407 |   0.0 | 4.453782 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001498 | 0.000017 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 49.747184 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001498 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000017 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             20.500246 |                  3.807188 |                  20.500246 |                  2.799361 |                      0.0 |    17.700884 |              0.0 |         17.700884 |             31.333333 |                 100.0 |               368.725681 |              139.570111 |    0.0 |  0.0 | 10.478761 |   0.0 | 4.442334 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000799 | 0.000009 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  45.30485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000799 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000009 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             20.595418 |                  3.824863 |                  20.595418 |                  2.986484 |                      0.0 |    17.608935 |              0.0 |         17.608935 |                  37.6 |                 100.0 |               419.711234 |              142.937942 |    0.0 |  0.0 |  9.985603 | 0.094 | 4.546047 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000426 | 0.000005 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 40.758803 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000426 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000005 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             19.897359 |                  3.695224 |                  19.897359 |                  2.708378 |                      0.0 |    17.188981 |              0.0 |         17.188981 |                 58.75 |                 100.0 |               411.253273 |              141.649314 |    0.0 |  0.0 |  7.775599 |   0.0 | 4.347844 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000228 | 0.000003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 36.410959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000228 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.332416 |                  3.218877 |                  17.332416 |                  2.239353 |                      0.0 |    15.093063 |              0.0 |         15.093063 |             34.814815 |                 100.0 |               357.548831 |              136.707247 |    0.0 |  0.0 |  7.857065 |   0.0 | 3.459696 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000121 | 0.000001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 32.951263 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000121 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000001 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           1.135969 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.450004 |                  3.055001 |                  16.450004 |                  2.021579 |                      0.0 |    14.428425 |              0.0 |         14.428425 |                29.375 |                 100.0 |               333.908335 |              133.915514 |  0.338 |  0.0 |  7.162052 |   2.2 | 1.984501 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.338 | 0.008727 |  0.0 |  0.0 | 0.0 |      0.0 | 0.008727 | 0.002312 | 0.000027 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 31.296034 | 0.008727 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002312 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000027 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.231733 |                  3.200179 |                  17.231733 |                  2.487815 |                      0.0 |    14.743918 |              0.0 |         14.743918 |             72.307692 |                 100.0 |               354.695699 |              135.231698 |    0.0 |  0.0 |  5.274274 |   0.0 | 3.081998 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003057 | 0.000035 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 28.214037 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003057 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000035 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             17.907703 |                  3.325716 |                  17.907703 |                  2.520953 |                      0.0 |     15.38675 |              0.0 |          15.38675 |             49.473684 |                 100.0 |               400.426126 |               138.20342 |    0.0 |  0.0 |  6.725557 |   0.0 | 3.429544 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.166895 |      0.0 | 0.001332 | 0.000015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 24.951387 |      0.0 |   0.0 |   0.0 |   0.0 | -0.166895 | 0.001631 |   0.0 |   0.0 |   0.0 | -0.000299 |   0.0 | 0.000015 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             17.704008 |                  3.287887 |                  17.704008 |                  2.702703 |                      0.0 |    15.001304 |              0.0 |         15.001304 |                  47.0 |                 100.0 |               491.324867 |              142.949964 |    0.0 |  0.0 |   7.53971 |   0.0 | 3.733858 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.369623 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 21.587152 |      0.0 |   0.0 |   0.0 |   0.0 | -0.369623 |  0.00087 |   0.0 |   0.0 |   0.0 | -0.001259 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.184616 |                      2.82 |                  15.184616 |                  2.307856 |                      0.0 |     12.87676 |              0.0 |          12.87676 |             36.153846 |                 100.0 |               525.389413 |              143.816091 |    0.0 |  0.0 |  7.832014 |   0.0 |  3.37772 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.578662 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 18.788095 |      0.0 |   0.0 |   0.0 |   0.0 | -0.578662 | 0.000464 |   0.0 |   0.0 |   0.0 | -0.002954 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           1.277965 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             17.774145 |                  3.300913 |                  17.774145 |                  2.688476 |                      0.0 |    15.085669 |              0.0 |         15.085669 |             26.111111 |                 100.0 |               629.540329 |              146.963401 |    0.0 |  0.0 | 10.542903 |   0.0 | 3.855581 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.752584 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 15.685099 |      0.0 |   0.0 |   0.0 |   0.0 | -0.752584 | 0.000248 |   0.0 |   0.0 |   0.0 | -0.005328 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             18.713094 |                  3.475289 |                  18.713094 |                  3.308913 |                      0.0 |    15.404181 |              0.0 |         15.404181 |                  37.6 |                 100.0 |               884.502325 |              152.120949 |    0.0 |  0.0 |  9.796451 |   0.0 |  4.29127 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.945392 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 12.339221 |      0.0 |   0.0 |   0.0 |   0.0 | -0.945392 | 0.000132 |   0.0 |   0.0 |   0.0 | -0.008351 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              6.887397 |                  1.279088 |                   6.887397 |                   0.60663 |                      0.0 |     6.280768 |              0.0 |          6.280768 |             44.761905 |                 100.0 |               696.001395 |              147.247892 |    0.0 |  0.0 |  3.397063 |  1.88 | 0.698098 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.153291 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 12.794413 |      0.0 |   0.0 |   0.0 |   0.0 | -1.153291 | 0.000071 |   0.0 |   0.0 |   0.0 | -0.012081 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.073322 |                  2.427903 |                  13.073322 |                  1.641158 |                      0.0 |    11.432163 |              0.0 |         11.432163 |             62.666667 |                 100.0 |               667.674311 |              145.668066 |    0.0 |  0.0 |  4.351914 | 0.282 | 2.206487 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.125007 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 11.712934 |      0.0 |   0.0 |   0.0 |   0.0 | -1.125007 | 0.000038 |   0.0 |   0.0 |   0.0 | -0.016118 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.464715 |                  3.057733 |                  16.464715 |                  1.868809 |                      0.0 |    14.595906 |              0.0 |         14.595906 |             42.727273 |                 100.0 |               938.453811 |              150.290353 |    0.0 |  0.0 |  7.236094 |   0.0 | 3.494179 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.192206 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  9.410961 |      0.0 |   0.0 |   0.0 |   0.0 | -1.192206 |  0.00002 |   0.0 |   0.0 |   0.0 | -0.020212 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |              12.69102 |                  2.356904 |                   12.69102 |                  1.401603 |                      0.0 |    11.289418 |              0.0 |         11.289418 |             49.473684 |                 100.0 |              1353.717528 |               153.70968 |    0.0 |  0.0 |  5.830325 | 0.094 | 2.901447 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.335241 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  7.844755 |      0.0 |   0.0 |   0.0 |   0.0 | -1.335241 | 0.000011 |   0.0 |   0.0 |   0.0 | -0.024667 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.099578 |                   2.06135 |                  11.099578 |                  1.261689 |                      0.0 |     9.837889 |              0.0 |          9.837889 |             40.869565 |                 100.0 |              1694.004243 |              154.756666 |   7.67 |  0.0 |  4.656548 |   2.2 | 1.080941 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   7.67 | 0.067757 |  0.0 |  0.0 | 0.0 | 1.432559 | 0.067757 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 15.798616 | 0.067757 |   0.0 |   0.0 |   0.0 | -1.432559 |  0.01745 |   0.0 |   0.0 |   0.0 | -0.029537 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.586633 |                  2.337518 |                  12.586633 |                  1.498964 |                      0.0 |    11.087669 |              0.0 |         11.087669 |             44.761905 |                 100.0 |               514.362016 |              137.508257 |   3.91 |  0.0 |  4.603863 |   2.2 | 1.172642 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   3.91 | 0.052955 |  0.0 |  0.0 | 0.0 | 0.938338 | 0.052955 | 0.003425 |  0.00004 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 19.421358 | 0.052955 |   0.0 |   0.0 |   0.0 | -0.938338 | 0.037102 |   0.0 |   0.0 |   0.0 | -0.033677 |   0.0 |  0.00004 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |                6.6675 |                   1.23825 |                     6.6675 |                  0.589077 |                      0.0 |     6.078423 |              0.0 |          6.078423 |             36.153846 |                 100.0 |               423.391318 |               132.52466 |   11.9 |  0.0 |  3.480236 |   2.2 | 0.567552 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   11.9 |   0.2321 |  0.0 |  0.0 | 0.0 | 0.713236 |   0.2321 | 0.054101 | 0.000626 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 31.234942 |   0.2321 |   0.0 |   0.0 |   0.0 | -0.713236 | 0.090616 |   0.0 |   0.0 |   0.0 | -0.036515 |   0.0 | 0.000626 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.413823 |                  1.191138 |                   6.413823 |                  0.708412 |                      0.0 |     5.705411 |              0.0 |          5.705411 |             36.153846 |                 100.0 |               347.072847 |              127.000616 |  6.448 |  0.0 |  2.685024 |   2.2 | 0.203792 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  6.448 | 0.173145 |  0.0 |  0.0 | 0.0 |      0.0 | 0.173145 | 0.103764 | 0.001201 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 37.306005 | 0.173145 |   0.0 |   0.0 |   0.0 |       0.0 | 0.141425 |   0.0 |   0.0 |   0.0 | -0.037661 |   0.0 | 0.001201 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             10.820053 |                  2.009438 |                  10.820053 |                  1.881434 |                      0.0 |      8.93862 |              0.0 |           8.93862 |             44.761905 |                 100.0 |               346.875642 |              126.428035 |    0.0 |  0.0 |  4.533079 |   0.0 | 2.160946 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.074117 | 0.000858 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 35.145059 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.111643 |   0.0 |   0.0 |   0.0 | -0.037526 |   0.0 | 0.000858 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.634933 |                  1.789345 |                   9.634933 |                  1.511299 |                      0.0 |     8.123635 |              0.0 |          8.123635 |             44.761905 |                 100.0 |               345.936719 |              125.805103 |    0.0 |  0.0 |   4.43981 |   0.0 | 2.165797 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.022183 | 0.000257 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 32.979262 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.059574 |   0.0 |   0.0 |   0.0 | -0.037392 |   0.0 | 0.000257 |

.. _lland_v3_acker_winter_daily:

acre (winter)
-------------

None of the examples discussed so far dealt with snow processes.  Thus,
we now switch from a warm August to a snowy December and adapt the initial
conditions and the input time series accordingly:

>>> pub.timegrids = "2010-12-01", "2011-01-01", "1d"
>>> lnk(ACKER)
>>> lai.acker_dec = 0.3
>>> cropheight.acker_dec = 0.05
>>> albedo.acker_dec = 0.1
>>> wg2z.dec = 0.7416
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 0.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     0.0, 11.1, 0.4, 1.2, 6.4, 0.0, 0.0, 0.0, 0.7, 3.3, 9.4, 0.2, 2.9, 0.0,
...     0.0, 6.7, 0.0, 0.1, 0.0, 0.2, 0.2, 2.4, 7.0, 0.2, 0.0, 1.2, 0.9, 0.0,
...     0.0, 0.2, 0.9)
>>> inputs.teml.series = (
...     -5.8, -3.0, -5.6, -3.8, 0.2, -1.1, -2.9, -2.9, -2.1, -3.1, 5.1, 0.2,
...     -3.5, -3.9, -6.3, -3.5, -7.3, -5.8, -7.9, -5.7, -5.6, -4.8, -1.1,
...     -2.9, -9.1, -8.6, -5.3, -7.7, -9.2, -7.4, 1.7)
>>> inputs.sunshineduration.series = (
...     3.5, 0.0, 0.1, 3.7, 0.0, 0.1, 0.0, 0.0, 4.4, 0.1, 0.2, 0.7, 0.0, 1.2,
...     3.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.8, 0.0, 0.0, 0.0, 3.2, 4.3, 0.0, 3.4,
...     0.0, 0.0, 4.0)
>>> inputs.windspeed.series = (
...     6.9, 5.2, 2.0, 4.3, 4.0, 2.4, 1.9, 1.8, 2.7, 3.0, 7.4, 5.7, 2.2, 3.1,
...     2.8, 3.7, 0.9, 1.3, 1.9, 2.4, 2.3, 4.6, 8.6, 7.6, 3.8, 1.5, 1.3, 3.0,
...     2.7, 1.4, 4.2)
>>> inputs.relativehumidity.series = (
...     70.3, 85.0, 95.9, 90.6, 96.5, 99.0, 99.2, 99.1, 96.9, 95.0, 90.6,
...     78.7, 90.5, 88.1, 92.8, 98.1, 96.6, 97.7, 95.8, 97.5, 95.0, 93.8,
...     94.9, 85.4, 87.2, 92.7, 97.8, 95.4, 94.9, 96.0, 96.5)
>>> inputs.atmosphericpressure.series = (
...     101.5, 100.7, 100.5, 100.8, 99.6, 99.8, 100.4, 100.6, 100.9, 101.8,
...     100.4, 101.2, 102.0, 102.5, 102.3, 100.0, 99.1, 99.7, 99.7, 99.7,
...     100.4, 100.7, 100.4, 100.8, 101.8, 101.9, 101.7, 102.2, 102.4, 102.1,
...     101.6)

When comparing the following results with those of integration test
:ref:`lland_v1_acker_winter` on |lland_v1|, it is not immediately clear what
differences are due to the much higher complexities of the snow module of
|lland_v3|.  At least, one can see that |lland_v3| allows for the (re)freezing
of liquid water within the snow-layer.  Refreezing  occurs around December 7th,
where the frozen water equivalent (|WATS|) grows until it reaches to total
water equivalent (|WAeS|) (a note for `LARSIM`_ users: currently, `LARSIM`_
does not implement such a freezing feature; you can disable it in |lland_v3|
through setting |RefreezeFlag| to |False|):

.. integration-test::

    >>> conditions_acker_winter = test(
    ...     "lland_v3_acker_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |         tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |     evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |        0.0 |   3.00672 |              2.483872 |                  2.483872 |                   2.483872 |                   4.01525 |                      0.0 |    -1.531378 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |      0.0 |      0.0 |  1.808179 |       0.0 | 0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                     0.381881 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |    -7.5504 | -0.665713 |              0.106229 |                  0.106229 |                   0.106229 |                  1.277835 |                 2.406897 |    -1.171606 |        -2.300668 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 |   10.374 |   10.374 |  0.779906 |      0.06 |     0.0 | -0.168357 | -0.065045 |    -6.266232 |     0.875479 |       0.0 |       0.0 | 0.406773 |      0.0 | -0.449083 |  -2.44648 | -0.594896 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 10.542357 | 10.542357 | -0.135862 | 0.084415 | -0.857807 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                     0.357079 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 |  -2.859357 |  2.216462 |              0.200706 |                  0.200706 |                   0.200706 |                  1.329454 |                  3.24186 |    -1.128748 |        -3.041154 |         -1.128748 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |    0.316 |    0.316 |  0.039575 |  0.039575 |     0.0 | -0.073822 | -0.003698 |    -7.134939 |     0.860761 |       0.0 |       0.0 | 0.484791 |      0.0 | -0.196916 | -0.605417 |  2.238821 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.020425 | 10.932178 | 10.932178 |  -0.16192 | 0.142292 | -2.332669 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                     0.323612 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 |  -7.775563 |  0.321699 |              0.409591 |                  0.409591 |                   0.409591 |                  3.727317 |                 4.828749 |    -3.317726 |        -4.419159 |         -3.317726 |             49.536202 |                 100.0 |               112.102145 |               99.524551 | 1.088425 | 1.088425 |  0.092184 |      0.06 |     0.0 | -0.402205 | -0.008644 |    -8.396125 |     0.850958 |       0.0 |       0.0 | 0.631546 |      0.0 | -1.072863 | -2.984225 |  0.362071 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 12.422809 | 12.422809 | -0.210936 | 0.184396 | -1.912768 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                     0.491269 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 |  -6.375893 | -1.777198 |              0.126436 |                  0.126436 |                   0.126436 |                  1.140917 |                 1.372579 |    -1.014481 |        -1.246143 |         -1.014481 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |    5.956 |   2.6802 |  0.134809 |      0.06 |     0.0 |  -0.43406 |  0.003863 |    -2.947657 |     0.846485 |       0.0 |       0.0 | 0.399614 | 0.399614 | -1.157835 | -1.939108 |   -1.8508 | 0.977319 | 0.286546 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 15.868755 | 18.812869 |       0.0 | 0.204657 |   0.60603 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                     0.515142 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |        0.0 |  1.198359 |              0.260442 |                  0.260442 |                   0.260442 |                  1.217715 |                 2.425289 |    -0.957272 |        -2.164847 |         -0.957272 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |      0.0 |      0.0 |  0.000623 |       0.0 |     0.0 | -0.123607 |       0.0 |    -2.311649 |     0.813253 |       0.0 |       0.0 | 0.314856 | 0.314856 | -0.329716 |  -0.53161 |  1.303521 | 0.994414 | 0.296657 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 16.289943 | 18.936477 |       0.0 | 0.380049 |  0.149271 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                     0.462335 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |        0.0 |  1.947227 |              0.171497 |                  0.171497 |                   0.171497 |                  1.193664 |                 2.796896 |    -1.022168 |          -2.6254 |         -1.022168 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |      0.0 |      0.0 | -0.015896 | -0.015896 |     0.0 | -0.067645 |       0.0 |    -3.756225 |     0.787515 |       0.0 |       0.0 | 0.507501 | 0.507501 |  -0.18044 | -0.328227 |  2.116732 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.015896 | 16.857448 | 19.004122 |       0.0 | 0.555441 | -1.056356 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                     0.422308 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 |  -3.521188 |  0.741715 |              0.186355 |                  0.186355 |                   0.186355 |                  1.194208 |                 2.345156 |    -1.007852 |        -2.158801 |         -1.007852 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |      0.0 |      0.0 |  -0.01263 |  -0.01263 |     0.0 | -0.162793 |       0.0 |    -4.951965 |     0.766993 |       0.0 |       0.0 | 0.655613 | 0.655613 | -0.434243 | -0.763868 |   0.96069 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.028526 | 17.663081 | 19.166915 |       0.0 | 0.730833 | -1.056472 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                     0.389674 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 |  -3.521572 |  1.286683 |              0.708095 |                  0.708095 |                   0.708095 |                  4.005891 |                 5.077157 |    -3.297796 |        -4.369062 |         -3.297796 |             78.890988 |                 100.0 |               112.102145 |               99.513532 | 0.626526 | 0.626526 | -0.234885 | -0.234885 |     0.0 | -0.365657 |  -0.00275 |    -6.003598 |     0.754667 |       0.0 |       0.0 | 0.800268 | 0.800268 | -0.975372 | -1.842468 |  1.551222 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.294885 | 19.442535 | 20.159098 |       0.0 | 0.856575 | -1.601554 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                     0.402618 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 |  -5.338514 |  0.123879 |              0.312643 |                  0.312643 |                   0.312643 |                  1.286731 |                 2.471706 |    -0.974088 |        -2.159063 |         -0.974088 |             71.001889 |                 100.0 |               112.102145 |               99.511814 | 3.336885 | 3.336885 |  0.136846 |      0.06 |     0.0 | -0.196001 |  -0.02162 |    -5.577477 |      0.76791 |       0.0 |       0.0 | 0.845123 | 0.716563 | -0.522823 |  -1.25171 |   0.38453 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 23.691984 | 23.691984 | -0.042939 | 0.722158 | -0.983833 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 |  -3.279442 | -1.700063 |              0.343097 |                  0.343097 |                   0.343097 |                   1.16859 |                -0.065197 |    -0.825492 |         0.408294 |         -0.825492 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |    8.776 |      0.0 |  1.032578 |      0.06 |     0.0 | -1.215856 |  0.187391 |          0.0 |     0.750116 | 21.434861 | 21.434861 |      0.0 |      0.0 | -3.243236 | -5.063324 | -8.714854 |  0.94544 | 0.268157 | 30.335284 | 9.737948 |  0.0 |  0.0 | 0.0 |  0.0 | 9.737948 | nan | nan |      0.0 |  2.341648 |  3.348556 |       0.0 | 0.907745 |   1.45783 | 92.043596 | 9.737948 |   0.0 |   0.0 |   0.0 |  0.0 | 2.507002 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                      0.51762 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |        0.0 |  1.164919 |              0.405635 |                  0.405635 |                   0.405635 |                  1.733999 |                 3.137107 |    -1.328364 |        -2.731472 |         -1.328364 |             37.369415 |                 100.0 |                83.924464 |               98.591233 |    0.128 |   0.0576 |  1.429059 |      0.06 |     0.0 |  0.158585 |  0.000045 |    -2.247143 |     0.736593 |       0.0 |       0.0 | 0.062798 | 0.062798 |  0.423016 | -1.968551 |  1.185938 | 0.961283 |  0.27722 |  0.045632 | 0.015236 |  0.0 |  0.0 | 0.0 |  0.0 | 0.015236 | nan | nan |      0.0 |  2.348245 |   3.27234 |       0.0 | 1.077936 |  1.034511 | 92.073992 | 0.015236 |   0.0 |   0.0 |   0.0 |  0.0 | 3.376356 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                     0.432015 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |        0.0 |  2.412148 |              0.192965 |                  0.192965 |                   0.192965 |                  1.254261 |                 3.092729 |    -1.061296 |        -2.899764 |         -1.061296 |             96.820757 |                 100.0 |                83.884383 |               98.589894 |    2.666 |    2.666 |  0.193851 |      0.06 |     0.0 |  0.014995 | -0.019502 |    -4.653064 |     0.750217 |       0.0 |       0.0 | 0.199897 | 0.199897 |  0.039999 | -0.480351 |  2.459412 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  5.200976 |  5.923345 |       0.0 | 0.906585 | -0.636037 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.804844 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                     0.383836 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |  -2.120123 |  2.114956 |              0.448113 |                  0.448113 |                   0.448113 |                  2.107297 |                  4.01017 |    -1.659183 |        -3.562057 |         -1.659183 |             68.711505 |                 100.0 |                83.884383 |               98.590228 |      0.0 |      0.0 |  0.281575 |       0.0 |     0.0 | -0.065035 |       0.0 |    -6.199899 |     0.736299 |       0.0 |       0.0 | 0.257917 | 0.257917 | -0.173479 | -1.187477 |  2.201101 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  5.518828 |   5.98838 |       0.0 | 1.081977 | -2.009393 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.963085 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                     0.282602 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 |  -6.697977 |  1.768774 |              0.638578 |                  0.638578 |                   0.638578 |                  3.397964 |                 4.986818 |    -2.759386 |        -4.348241 |         -2.759386 |             76.073452 |                 100.0 |                83.884383 |               98.590507 |      0.0 |      0.0 |   -0.0286 |   -0.0286 |     0.0 | -0.225866 |       0.0 |   -10.109965 |     0.724543 |       0.0 |       0.0 | 0.408644 | 0.408644 | -0.602486 | -1.840493 |  1.905262 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |   0.0286 |  6.151042 |  6.214246 |       0.0 | 1.257369 | -3.036568 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.513913 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                     0.357863 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -10.121892 | -1.563101 |              0.164206 |                  0.164206 |                   0.164206 |                  1.211777 |                  1.91563 |    -1.047571 |        -1.751423 |         -1.047571 |             57.569099 |                 100.0 |                83.884383 |               98.590731 |   6.2666 |   6.2666 |  0.035323 |  0.035323 |     0.0 |  -0.40191 |  -0.04584 |    -7.106651 |     0.784952 |       0.0 |       0.0 |  0.55784 | 0.063204 | -1.072074 | -2.101972 | -1.422622 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024677 | 12.882755 | 12.882755 | -0.165208 | 0.575375 | -0.731867 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.274229 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                     0.326227 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 |  -2.439556 |  3.034693 |              0.169472 |                  0.169472 |                   0.169472 |                  1.287913 |                   3.6017 |     -1.11844 |        -3.432228 |          -1.11844 |            236.672963 |                 100.0 |                83.884383 |               98.590899 |      0.0 |      0.0 | -0.001233 | -0.001233 |     0.0 | -0.025746 |       0.0 |    -8.293516 |     0.777544 |       0.0 |       0.0 | 0.668571 |      0.0 | -0.068677 | -0.270764 |  3.092787 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.02591 | 12.908502 | 12.908502 | -0.223303 | 0.635987 | -3.024959 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146332 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                     0.283779 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -10.083198 | -0.013178 |              0.173382 |                  0.173382 |                   0.173382 |                  1.257034 |                 2.131881 |    -1.083652 |          -1.9585 |         -1.083652 |            163.850513 |                 100.0 |                83.884383 |               98.591011 |  0.05991 |  0.05991 | -0.002775 | -0.002775 |     0.0 | -0.215187 | -0.000726 |   -10.057778 |      0.77207 |       0.0 |       0.0 | 0.816185 |      0.0 |    -0.574 | -1.349101 |  0.035399 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062775 | 13.183599 | 13.183599 | -0.272606 | 0.683858 | -2.270182 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.078084 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                     0.272284 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 |  -7.567272 |  1.559622 |              0.177036 |                  0.177036 |                   0.177036 |                  1.301361 |                 3.118663 |    -1.124325 |        -2.941627 |         -1.124325 |            112.108246 |                 100.0 |                83.884383 |               98.591067 | 0.002775 | 0.002775 |   0.03257 |   0.03257 |     0.0 | -0.126504 | -0.000046 |   -10.575801 |     0.767104 |       0.0 |       0.0 | 0.872647 |      0.0 | -0.337444 | -1.025749 |  1.578434 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.02743 | 13.312878 | 13.312878 | -0.291464 | 0.729776 | -3.088203 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041667 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                     0.293059 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -10.294011 | -0.332318 |              0.315755 |                  0.315755 |                   0.315755 |                  1.393316 |                 2.436172 |    -1.077561 |        -2.120417 |         -1.077561 |             88.752361 |                 100.0 |                83.884383 |               98.591067 |  0.15543 |  0.15543 |  0.021054 |  0.021054 |     0.0 | -0.277409 | -0.001852 |    -9.652965 |     0.763684 |       0.0 |       0.0 |  0.81353 |      0.0 | -0.739974 | -1.734358 | -0.353915 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.038946 | 13.745716 | 13.745716 | -0.271719 |  0.76289 | -2.014285 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.022234 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                     0.312873 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 |  -6.714283 |  1.094405 |              0.370424 |                  0.370424 |                   0.370424 |                  1.820398 |                 3.346196 |    -1.449974 |        -2.975772 |         -1.449974 |             92.611159 |                 100.0 |                83.884383 |               98.591011 | 0.166946 | 0.166946 |  0.035079 |  0.035079 |     0.0 | -0.194575 | -0.001954 |    -8.825403 |     0.760211 |       0.0 |       0.0 | 0.768324 |      0.0 | -0.519019 |   -1.3794 |  1.077352 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024921 | 14.107237 | 14.107237 |  -0.25662 | 0.797831 |  -2.36709 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.011864 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                     0.344247 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 |  -7.890299 | -0.147589 |              0.172781 |                  0.172781 |                   0.172781 |                  1.259735 |                 2.599284 |    -1.086953 |        -2.426503 |         -1.086953 |              46.30558 |                 100.0 |                83.884383 |               98.590899 | 2.220921 | 2.220921 |  0.238308 |      0.06 |     0.0 | -0.254372 |  -0.02228 |    -7.605598 |     0.773322 |       0.0 |       0.0 | 0.777088 |      0.0 | -0.678525 | -1.914919 | -0.166942 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  16.58253 |  16.58253 | -0.259548 | 0.672663 | -1.477901 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006331 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                     0.487686 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 |  -4.926336 | -0.975038 |              0.144842 |                  0.144842 |                   0.144842 |                  1.179808 |                 2.170767 |    -1.034966 |        -2.025925 |         -1.034966 |             24.768101 |                 100.0 |                83.884383 |               98.590731 |     6.52 |    5.053 |  0.461781 |      0.06 |     0.0 | -0.349504 | -0.018373 |    -3.045475 |     0.810455 |       0.0 |       0.0 | 0.468313 | 0.468313 | -0.932284 | -2.190183 | -1.096542 | 0.991065 | 0.294663 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 22.438239 | 23.452034 |       0.0 | 0.397177 |  0.238737 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003378 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                      0.45101 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |        0.0 |  2.117638 |              0.164009 |                  0.164009 |                   0.164009 |                  1.273508 |                 3.158235 |    -1.109498 |        -2.994226 |         -1.109498 |             28.027061 |                 100.0 |                83.884383 |               98.590507 |    0.128 |    0.128 |  1.132583 |      0.06 |     0.0 |  0.200688 | -0.000776 |    -4.084949 |     0.786082 |       0.0 |       0.0 |  0.62874 |  0.62874 |  0.535324 | -1.202689 |  2.326861 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 22.997568 | 23.379346 |       0.0 | 0.566529 |   -1.1373 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001803 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                     0.260779 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 |  -3.791001 |  3.796323 |              0.561629 |                  0.561629 |                   0.561629 |                  3.730222 |                 6.170919 |    -3.168592 |         -5.60929 |         -3.168592 |             56.054123 |                 100.0 |                83.884383 |               98.590228 |      0.0 |      0.0 |  0.146367 |       0.0 |     0.0 | -0.024302 |       0.0 |   -11.114155 |     0.765835 |       0.0 |       0.0 | 1.652593 | 0.381779 | -0.064824 | -1.196177 |  4.348289 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 23.403648 | 23.403648 | -0.424452 | 0.741921 | -4.192023 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000962 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                     0.181886 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -13.973411 |  0.795365 |               0.64551 |                   0.64551 |                    0.64551 |                  4.423241 |                 4.808458 |    -3.777731 |        -4.162948 |         -3.777731 |            142.003778 |                 100.0 |                83.884383 |               98.589894 |    1.068 |    1.068 |  -0.16832 |  -0.16832 |     0.0 |  -0.25352 | -0.019196 |   -15.507681 |       0.7689 |       0.0 |       0.0 | 2.374706 |      0.0 | -0.676251 | -2.341828 |  1.144869 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.22832 | 24.725168 | 24.725168 | -0.793152 | 0.712889 | -4.245789 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000513 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                     0.253812 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -14.152629 | -1.400795 |              0.176393 |                  0.176393 |                   0.176393 |                  1.247575 |                 1.363527 |    -1.071182 |        -1.187134 |         -1.071182 |            163.850513 |                 100.0 |                83.884383 |               98.589505 |  1.01432 |  1.01432 | -0.002431 | -0.002431 |     0.0 | -0.310229 | -0.011236 |   -11.450477 |     0.773273 |       0.0 |       0.0 | 1.844265 |      0.0 | -0.827519 | -1.948813 | -1.589198 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062431 | 26.049716 | 26.049716 | -0.615984 | 0.673105 | -2.103393 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000274 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                      0.25535 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 |  -7.011311 |  2.262411 |              0.579189 |                  0.579189 |                   0.579189 |                    3.6896 |                 5.330232 |    -3.110411 |        -4.751043 |         -3.110411 |             71.001889 |                 100.0 |                83.884383 |               98.589062 | 0.002431 | 0.002431 | -0.122138 | -0.122138 |     0.0 | -0.235528 | -0.000039 |   -11.375531 |     0.768701 |       0.0 |       0.0 | 1.854449 |      0.0 | -0.628259 |  -1.85701 |  2.265774 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.182138 | 26.287676 | 26.287676 | -0.619386 |  0.71474 | -3.624205 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000146 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                     0.225287 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -12.080682 |  0.434734 |              0.185161 |                  0.185161 |                   0.185161 |                  1.325064 |                  2.98994 |    -1.139902 |        -2.804778 |         -1.139902 |             78.890988 |                 100.0 |                83.884383 |               98.588564 | 0.122138 | 0.122138 |  0.072845 |      0.06 |     0.0 | -0.196144 | -0.002348 |   -12.919289 |     0.765211 |       0.0 |       0.0 | 2.135029 |      0.0 | -0.523203 | -1.755476 |  0.526099 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 26.605957 | 26.605957 |   -0.7131 | 0.747953 | -3.317338 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000078 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                     0.247537 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -11.057795 |  0.364297 |                0.1889 |                    0.1889 |                     0.1889 |                  1.292505 |                 2.434681 |    -1.103604 |        -2.245781 |         -1.103604 |            152.146905 |                 100.0 |                83.884383 |               98.588014 |    0.128 |    0.128 |  0.018911 |  0.018911 |     0.0 | -0.190998 |  -0.00198 |   -11.760528 |     0.762353 |       0.0 |       0.0 | 1.967387 |      0.0 | -0.509478 | -1.429978 |  0.306325 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.041089 | 26.924955 | 26.924955 | -0.657107 | 0.776116 | -2.940036 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000042 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                     0.446978 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 |  -9.800119 | -2.901162 |              0.668643 |                  0.668643 |                   0.668643 |                  3.475519 |                 3.601527 |    -2.806876 |        -2.932884 |         -2.806876 |             50.715635 |                 100.0 |                83.884383 |                98.58741 | 0.827089 | 0.062032 | -0.128392 | -0.128392 |     0.0 | -0.924887 |  0.005666 |    -4.203742 |     0.758628 |       0.0 |       0.0 | 0.750204 | 0.750204 | -2.467091 | -3.767829 | -3.302036 |   0.9737 | 0.284428 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.188392 | 28.661583 | 28.676932 |       0.0 | 0.814224 |  0.702726 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000022 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_nadelw_winter_daily:

conifers (winter)
-----------------

Next, we repeat the integration test :ref:`lland_v3_acker_winter_daily` but
select the land use type |NADELW| instead of |ACKER|.  Like for |LAUBW|
and |MISCHW|, |lland_v3| now modifies the global radiation, the net longwave
radiation, and the wind speed to take the shadowing effects of the
tree-canopies into account.  Here, this net-effect of these modifications is
that the peak amounts of the total water equivalent are two to three times
smaller than for land use type |ACKER|:

.. integration-test::

    >>> lnk(NADELW)
    >>> conditions_nadelw_winter = test(
    ...     "lland_v3_nadelw_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |           2.449434 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.428675 |                   0.45104 |                   2.428675 |                   4.01525 |                      0.0 |    -1.586575 |              0.0 |         -1.586575 |             13.623188 |                 100.0 |             27827.355778 |              101.116829 |      0.0 |      0.0 |  4.142976 |       0.0 | 0.674992 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |            1.84595 |          5.2 |                 0.489349 |                      0.489349 |                     0.395271 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -0.898177 |              0.106229 |                  0.019728 |                   0.106229 |                  1.277835 |                 0.531756 |    -1.171606 |        -0.512028 |         -1.171606 |             18.076923 |                 100.0 |              2039.690308 |               99.509579 |    8.234 |    8.234 |  1.794481 |  1.794481 |      0.0 | -0.058384 | -0.051627 |    -5.817805 |     0.875479 |      0.0 |      0.0 | 0.299757 |      0.0 | -0.155737 | -1.205976 | -0.849686 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.405519 |  8.292384 |  8.292384 | -0.100119 | 0.084415 | -0.625343 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 0.401924 |                      0.401924 |                     0.442417 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -2.084476 |  1.168887 |              0.202282 |                  0.037567 |                   0.202282 |                  1.329454 |                 0.568267 |    -1.127172 |        -0.530701 |         -1.127172 |                  47.0 |                 100.0 |             28069.018521 |              100.749738 |      0.0 |      0.0 |   0.12953 |   0.12953 |      0.0 |   0.10185 |       0.0 |    -4.339273 |     0.859668 |      0.0 |      0.0 | 0.225163 |      0.0 |   0.27168 |  0.341591 |  1.143972 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.651989 |  8.190534 |  8.190534 | -0.075204 |  0.14684 | -1.052629 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           1.526459 |          4.3 |                 0.460812 |                      0.460812 |                     0.427292 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 | -3.508765 |  0.668199 |              0.429653 |                  0.079793 |                   0.429653 |                  3.727317 |                 1.071685 |    -3.297664 |        -0.991893 |         -3.297664 |             21.860465 |                 100.0 |              3347.213388 |                99.76651 |      0.0 |      0.0 |  0.584703 |  0.584703 |      0.0 |  0.024809 |       0.0 |    -4.797729 |     0.843658 |      0.0 |      0.0 | 0.245894 |      0.0 |  0.066176 | -0.382946 |  0.675123 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.195286 |  8.165725 |  8.165725 | -0.082129 | 0.217823 | -0.979228 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           1.419962 |          4.0 |                 0.619748 |                      0.619748 |                     0.537181 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 | -3.264094 |  -0.78634 |               0.13628 |                  0.025309 |                    0.13628 |                  1.140917 |                 0.332476 |    -1.004637 |        -0.307167 |         -1.004637 |                  23.5 |                 100.0 |               763.990266 |               96.624998 | 5.011286 | 2.255079 |  0.342821 |  0.342821 |      0.0 | -0.148252 |  0.003251 |    -1.747234 |     0.834534 |      0.0 |      0.0 | 0.174301 | 0.174301 | -0.395454 | -0.718715 | -0.807002 | 0.979464 | 0.287805 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.857179 | 10.714308 | 13.325263 |       0.0 | 0.262355 |  0.548712 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.563475 |                      0.563475 |                     0.572963 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |       0.0 |  0.452563 |              0.273217 |                   0.05074 |                   0.273217 |                  1.217715 |                 0.389096 |    -0.944497 |        -0.338355 |         -0.944497 |             39.166667 |                 100.0 |              1010.452619 |               97.419093 |      0.0 |      0.0 |   0.03379 |   0.03379 |      0.0 |  0.028994 |       0.0 |       -0.873 |     0.804093 |      0.0 |      0.0 | 0.087103 | 0.087103 |  0.077339 |  0.065961 |  0.481656 | 0.968647 | 0.281483 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.823389 | 10.777908 | 13.296269 |       0.0 | 0.437747 |  0.837749 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 0.493024 |                      0.493024 |                     0.538352 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  0.890518 |              0.177343 |                  0.032935 |                   0.177343 |                  1.193664 |                 0.435045 |    -1.016321 |         -0.40211 |         -1.016321 |             49.473684 |                 100.0 |               1874.96551 |               98.634658 |      0.0 |      0.0 |  0.000946 |  0.000946 |      0.0 |   0.08649 |       0.0 |    -1.717821 |     0.780271 |      0.0 |      0.0 | 0.170083 | 0.170083 |  0.230709 |  0.314507 |  0.947326 |  0.97422 | 0.284732 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.822443 | 10.876776 | 13.209779 |       0.0 | 0.613139 |  0.688831 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 0.493024 |                      0.493024 |                     0.538702 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 |       0.0 |  0.885964 |              0.191051 |                  0.035481 |                   0.191051 |                  1.194208 |                 0.435661 |    -1.003157 |         -0.40018 |         -1.003157 |             52.222222 |                 100.0 |              1875.428739 |                98.51545 |      0.0 |      0.0 |  0.005497 |  0.005497 |      0.0 |  0.086348 |       0.0 |    -1.709036 |     0.761122 |      0.0 |      0.0 |   0.1663 |   0.1663 |  0.230328 |     0.311 |  0.941508 | 0.979623 | 0.287899 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.816946 | 10.970891 | 13.123431 |       0.0 | 0.788531 |  0.544466 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.523311 |                      0.523311 |                     0.528127 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 |       0.0 |  1.024664 |               0.72913 |                   0.13541 |                    0.72913 |                  4.005891 |                 1.072417 |     -3.27676 |        -0.937007 |          -3.27676 |             34.814815 |                 100.0 |              1366.137242 |               97.824519 | 0.274946 | 0.274946 | -0.145708 | -0.145708 |      0.0 |  0.042379 | -0.001207 |     -1.97659 |     0.747379 |      0.0 |      0.0 | 0.192428 | 0.192428 |  0.113045 |  0.037677 |  1.087729 | 0.990217 | 0.294159 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.345708 | 11.402086 | 13.355998 |       0.0 | 0.939906 |  0.261402 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.485698 |                      0.485698 |                     0.525121 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 |       0.0 |  1.064535 |              0.320728 |                  0.059564 |                   0.320728 |                  1.286731 |                 0.489004 |    -0.966004 |        -0.429441 |         -0.966004 |             31.333333 |                 100.0 |              2093.665722 |               98.468592 | 3.247708 | 3.247708 |  0.333009 |  0.333009 |      0.0 |  0.134519 | -0.021042 |    -2.053502 |     0.761908 |      0.0 |      0.0 | 0.238543 | 0.238543 |  0.358824 |  0.334902 |  1.123167 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.866991 | 14.767714 | 16.469186 |       0.0 | 0.780583 | -0.061533 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           2.626929 |          7.4 |                 0.879017 |                      0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 | -0.205111 |  -0.10633 |              0.349872 |                  0.064976 |                   0.349872 |                   1.16859 |                 0.213257 |    -0.818717 |        -0.148281 |         -0.818717 |             12.702703 |                 100.0 |               405.842399 |               92.890929 | 8.502991 |      0.0 |  2.300327 |       2.2 |      0.0 | -0.656354 |  0.099261 |          0.0 |     0.745182 | 12.96039 | 12.96039 |      0.0 |      0.0 | -1.750792 | -2.733328 | -4.335839 | 0.970569 | 0.282601 | 22.963731 | 7.623127 |  0.0 |  0.0 | 0.0 |  0.0 | 7.623127 | nan | nan |      0.0 |  1.863497 |  2.664801 |       0.0 |  0.96651 |  0.786396 | 86.665611 | 7.623127 |   0.0 |   0.0 |   0.0 |  0.0 | 1.962548 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |           2.023445 |          5.7 |                 0.619748 |                      0.619748 |                     0.564435 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |  0.558256 |              0.412606 |                  0.076627 |                   0.412606 |                  1.733999 |                 0.609687 |    -1.321393 |        -0.533061 |         -1.321393 |             16.491228 |                 100.0 |               656.301186 |               95.339607 |      0.0 |      0.0 |  3.264511 |     0.188 |      0.0 |  0.228989 |       0.0 |    -1.076882 |     0.732066 |      0.0 |      0.0 | 0.023374 | 0.023374 |  0.610818 | -0.577816 |  0.566063 | 0.963707 |  0.27862 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |   1.72473 |  2.435811 |       0.0 | 1.141902 |  0.969741 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.640032 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 0.471336 |                      0.471336 |                     0.513926 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |  1.214827 |              0.211461 |                  0.039271 |                   0.211461 |                  1.254261 |                 0.511811 |      -1.0428 |         -0.47254 |           -1.0428 |             42.727273 |                 100.0 |              2173.673838 |               98.277258 |    0.526 |    0.526 |  0.463476 |  0.463476 |      0.0 |  0.161846 | -0.003848 |    -2.343417 |     0.726275 |      0.0 |      0.0 | 0.053893 | 0.053893 |  0.431715 |  0.324725 |   1.22898 | 0.981418 | 0.288954 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.736524 |  2.178689 |  2.799966 |       0.0 | 1.229697 |  0.496513 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  1.40875 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |            1.10047 |          3.1 |                 0.457351 |                      0.457351 |                     0.493405 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |  1.497998 |              0.482654 |                  0.089636 |                   0.482654 |                  2.107297 |                 0.762591 |    -1.624642 |        -0.672956 |         -1.624642 |             30.322581 |                 100.0 |              2978.799239 |               98.578374 |      0.0 |      0.0 |  0.747786 |  0.747786 |      0.0 |  0.193978 |       0.0 |    -2.889656 |     0.715972 |      0.0 |      0.0 | 0.061899 | 0.061899 |  0.517426 |   0.32829 |  1.518672 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.988738 |  2.085363 |  2.605988 |       0.0 | 1.405089 | -0.259884 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.751725 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 0.380889 |                      0.380889 |                     0.436373 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 | -0.866281 |  1.894499 |              0.679084 |                  0.126116 |                   0.679084 |                  3.397964 |                 1.045602 |     -2.71888 |        -0.919486 |          -2.71888 |             33.571429 |                 100.0 |             22767.949382 |               99.384825 |      0.0 |      0.0 |  0.185118 |  0.185118 |      0.0 |  0.169693 |       0.0 |    -4.520792 |      0.70707 |      0.0 |      0.0 | 0.088496 | 0.088496 |  0.452648 |  0.551922 |  1.924056 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.80362 |  2.032305 |  2.436295 |       0.0 | 1.580481 | -1.412783 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.401129 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           1.313465 |          3.7 |                 0.471336 |                      0.471336 |                     0.437438 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -4.709277 | -0.114381 |              0.190433 |                  0.035366 |                   0.190433 |                  1.211777 |                 0.463683 |    -1.021345 |        -0.428317 |         -1.021345 |             25.405405 |                 100.0 |              2134.566544 |               98.100564 |  4.90162 |  4.90162 |  0.125363 |  0.125363 |      0.0 |  -0.05832 | -0.035855 |    -4.488635 |     0.750606 |      0.0 |      0.0 | 0.217488 | 0.217488 | -0.155566 | -0.350346 | -0.077595 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.074637 |  7.208251 |  7.396235 |       0.0 | 0.902108 | -0.556802 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.214047 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.319491 |          0.9 |                 0.352531 |                      0.352531 |                     0.433413 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -1.856007 |   1.42792 |              0.200646 |                  0.037263 |                   0.200646 |                  1.287913 |                 0.600105 |    -1.087267 |        -0.562843 |         -1.087267 |            104.444444 |                 100.0 |             22767.949382 |               99.309706 |      0.0 |      0.0 |  0.027204 |  0.027204 |      0.0 |  0.132947 |       0.0 |    -4.610482 |     0.736625 |      0.0 |      0.0 | 0.218823 | 0.187984 |  0.354629 |  0.583535 |  1.501006 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.047433 |  7.263288 |  7.263288 |   -0.0103 |   1.0775 | -1.243122 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114218 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 0.395811 |                      0.395811 |                     0.409234 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 |  -4.14374 |  0.632759 |              0.207622 |                  0.038558 |                   0.207622 |                  1.257034 |                 0.545575 |    -1.049412 |        -0.507016 |         -1.049412 |             72.307692 |                 100.0 |             22767.949382 |               99.288228 |      0.0 |      0.0 |  0.029597 |  0.029597 |      0.0 |  0.035165 |       0.0 |     -5.36434 |     0.727057 |      0.0 |      0.0 | 0.243809 |      0.0 |    0.0938 |  0.103075 |  0.703891 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.111836 |  7.228124 |  7.228124 | -0.081432 | 1.217412 | -1.134281 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060948 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  0.33643 |                       0.33643 |                     0.382007 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -3.780937 |  1.286157 |              0.210531 |                  0.039099 |                   0.210531 |                  1.301361 |                 0.623068 |     -1.09083 |        -0.583969 |          -1.09083 |             49.473684 |                 100.0 |             22767.949382 |               99.277487 |      0.0 |      0.0 |  0.112465 |  0.112465 |      0.0 |  0.104807 |       0.0 |    -6.261951 |      0.72304 |      0.0 |      0.0 | 0.283227 |      0.0 |  0.279567 |  0.435787 |  1.299323 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.999371 |  7.123317 |  7.123317 | -0.094598 | 1.281941 | -1.678839 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.032522 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.398857 |                      0.398857 |                     0.382911 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -5.596129 |  0.329248 |              0.374753 |                  0.069597 |                   0.374753 |                  1.393316 |                 0.581946 |    -1.018563 |        -0.512349 |         -1.018563 |             39.166667 |                 100.0 |             22767.949382 |               99.277487 |      0.0 |      0.0 |  0.092592 |  0.092592 |      0.0 | -0.011456 |       0.0 |    -6.231253 |     0.719529 |      0.0 |      0.0 | 0.277752 |      0.0 | -0.030559 |  -0.15437 |   0.32742 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.094779 |  7.134773 |  7.134773 | -0.092769 | 1.341508 | -1.266487 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.017354 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 0.401924 |                      0.401924 |                     0.401392 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 | -4.221623 |  0.723518 |              0.437541 |                  0.081258 |                   0.437541 |                  1.820398 |                  0.70235 |    -1.382858 |        -0.621092 |         -1.382858 |             40.869565 |                 100.0 |             22767.949382 |               99.288228 | 0.082779 | 0.082779 |  0.167454 |  0.167454 |      0.0 |  0.036876 | -0.000969 |    -5.617299 |     0.716764 |      0.0 |      0.0 | 0.253698 |      0.0 |  0.098365 | -0.004942 |  0.714515 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.032546 |  7.180676 |  7.180676 | -0.084735 | 1.390636 | -1.248405 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.00926 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |           1.632956 |          4.6 |                 0.427219 |                      0.427219 |                      0.41713 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 |  -4.16135 |  0.493809 |              0.205392 |                  0.038144 |                   0.205392 |                  1.259735 |                 0.535417 |    -1.054343 |        -0.497273 |         -1.054343 |             20.434783 |                 100.0 |             16151.471046 |               99.256536 | 2.088546 | 2.088546 |   0.58112 |   0.58112 |      0.0 |  0.043122 | -0.020952 |    -5.113914 |      0.73054 |      0.0 |      0.0 | 0.296617 |      0.0 |  0.115027 | -0.125107 |  0.487192 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  1.61888 |    9.2261 |    9.2261 |  -0.09907 | 1.164357 | -1.000614 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004941 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           3.052917 |          8.6 |                 0.563475 |                      0.563475 |                     0.515075 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -3.335381 | -0.529796 |              0.175813 |                  0.032651 |                   0.175813 |                  1.179808 |                 0.394667 |    -1.003995 |        -0.362016 |         -1.003995 |             10.930233 |                 100.0 |               829.225836 |               96.016503 |  5.99888 | 4.649132 |  1.086922 |  1.086922 |      0.0 | -0.077173 | -0.016905 |    -2.313399 |     0.769925 |      0.0 |      0.0 |     0.24 |     0.24 | -0.205854 | -0.721772 |  -0.56561 | 0.989866 |  0.29395 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.113078 |  14.18678 | 15.302153 |       0.0 | 0.703392 |  0.270781 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002637 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           2.697927 |          7.6 |                 0.493024 |                      0.493024 |                     0.503792 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |       0.0 |  1.353398 |              0.189632 |                  0.035217 |                   0.189632 |                  1.273508 |                 0.525497 |    -1.083876 |        -0.490279 |         -1.083876 |             12.368421 |                 100.0 |               1574.53183 |               97.661543 |      0.0 |      0.0 |  2.588786 |  1.301078 |      0.0 |  0.297978 |       0.0 |    -2.610721 |     0.752662 |      0.0 |      0.0 | 0.268265 | 0.268265 |  0.794841 |  0.157877 |  1.442998 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 14.173563 | 15.004175 |       0.0 | 0.878784 | -0.341016 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001407 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                  0.30617 |                       0.30617 |                     0.384345 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -1.136721 |  2.615827 |              0.627559 |                  0.116547 |                   0.627559 |                  3.730222 |                 1.146755 |    -3.102663 |        -1.030208 |         -3.102663 |             24.736842 |                 100.0 |             22767.949382 |               99.438432 |      0.0 |      0.0 |  0.606006 |       0.0 |      0.0 |  0.278218 |       0.0 |    -6.182684 |     0.738346 |      0.0 |      0.0 | 0.612722 | 0.612722 |  0.742132 |  1.048136 |  2.820476 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 14.512107 | 14.725957 |       0.0 | 1.054176 | -2.215243 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000751 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 0.318472 |                      0.318472 |                     0.311154 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -7.384144 |  0.783397 |              0.764548 |                  0.141987 |                   0.764548 |                  4.423241 |                  1.20683 |    -3.658694 |        -1.064843 |         -3.658694 |             62.666667 |                 100.0 |             22767.949382 |               99.502706 |      0.0 |      0.0 |  -0.08636 |  -0.08636 |      0.0 |  0.025901 |       0.0 |    -8.895328 |     0.726283 |      0.0 |      0.0 | 0.831624 |  0.21385 |  0.069089 | -0.072772 |   1.06116 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  1.21436 | 14.700056 | 14.700056 | -0.206336 | 1.229568 | -2.257041 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000401 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                  0.41125 |                       0.41125 |                     0.355399 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -7.523469 | -0.169932 |              0.215734 |                  0.040065 |                   0.215734 |                  1.247575 |                 0.528268 |    -1.031841 |        -0.488204 |         -1.031841 |             72.307692 |                 100.0 |             22767.949382 |               99.577619 |      0.0 |      0.0 |  0.030808 |  0.030808 |      0.0 | -0.073062 |       0.0 |    -7.195669 |     0.722706 |      0.0 |      0.0 | 0.661896 |      0.0 |  -0.19489 | -0.448508 | -0.155195 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.029552 | 14.773119 | 14.773119 | -0.221073 | 1.287482 | -1.345509 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000214 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.341723 |                      0.341723 |                     0.358381 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 |  -4.48503 |  1.349382 |              0.702605 |                  0.130484 |                   0.702605 |                    3.6896 |                 1.099425 |    -2.986995 |        -0.968941 |         -2.986995 |             31.333333 |                 100.0 |             22767.949382 |               99.663136 |      0.0 |      0.0 |  0.016489 |  0.016489 |      0.0 |  0.068366 |       0.0 |    -7.088005 |     0.719415 |      0.0 |      0.0 | 0.655233 |      0.0 |  0.182364 |  0.195852 |  1.347156 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.013063 | 14.704753 | 14.704753 | -0.218848 | 1.343486 | -1.953291 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000114 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.303761 |                      0.303761 |                     0.322491 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 |  -6.51097 |  1.000185 |              0.223758 |                  0.041555 |                   0.223758 |                  1.325064 |                 0.666484 |    -1.101305 |        -0.624929 |         -1.101305 |             34.814815 |                 100.0 |             22767.949382 |                99.75922 |      0.0 |      0.0 |  0.206218 |  0.206218 |      0.0 |  0.068935 |       0.0 |     -8.44034 |     0.716269 |      0.0 |      0.0 | 0.776636 |      0.0 |  0.183881 |  0.231924 |  1.040734 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.806845 | 14.635818 | 14.635818 | -0.259396 | 1.399653 | -2.211876 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000061 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           0.496987 |          1.4 |                 0.349801 |                      0.349801 |                      0.33091 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -7.372921 |  0.382915 |              0.227699 |                  0.042287 |                   0.227699 |                  1.292505 |                 0.606599 |    -1.064805 |        -0.564312 |         -1.064805 |             67.142857 |                 100.0 |             22767.949382 |               99.865828 |      0.0 |      0.0 |   0.07775 |   0.07775 |      0.0 | -0.007807 |       0.0 |    -8.111568 |     0.713542 |      0.0 |      0.0 | 0.742885 |      0.0 | -0.020824 | -0.171846 |  0.371642 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.917095 | 14.643624 | 14.643624 | -0.248124 | 1.450552 | -1.853191 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000032 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |            1.49096 |          4.2 |                 0.690794 |                      0.690794 |                     0.501108 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 | -6.177303 | -1.811807 |              0.800486 |                  0.148662 |                   0.800486 |                  3.475519 |                 0.945014 |    -2.675033 |        -0.796353 |         -2.675033 |             22.380952 |                 100.0 |               481.569845 |                94.15647 | 0.563095 | 0.042232 |  0.114221 |  0.114221 |      0.0 | -0.413784 |  0.003857 |    -2.682304 |     0.711034 |      0.0 |      0.0 | 0.264008 | 0.264008 | -1.103747 |   -1.6605 | -1.967895 | 0.973794 | 0.284483 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.085779 | 15.356659 | 15.620503 |       0.0 | 1.499321 |  0.700216 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000017 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_acker_winter_ktschnee_daily:

acre (snow surface temperature)
-------------------------------

In integration test :ref:`lland_v3_acker_winter_daily`, |lland_v3| assumes
that the snow-layer has a constant temperature over its complete depth.
For less simplified modelling of the processes within the snow layer, one
can enable an iterative adjustment of the snow surface temperature (see
method |Return_TempSSurface_V1|).  Therefore, one must set the thermal
conductivity of the top snow layer (|KTSchnee|) to a value smaller than
|numpy.inf|.  The lower the conductivity, the higher the possible differences
between the bulk and the surface temperature of the snow layer. Setting
zero values (which is not advisable) would prevent any thermal exchange,
setting |numpy.inf| values results in an infinitely fast exchange, avoiding
any temperature differences within the snow layer.

We show what happens when we use the default |KTSchnee| value used by
`LARSIM`_.  In integration test :ref:`lland_v3_acker_winter_daily`, with
:math:`KTSchnee = inf`, nearly the complete frozen water equivalent melts
on December 11th, where the average air temperature is 5 °C.  In our next
test, with :math:`KTSchnee = 0.018 MJ/m³/K/h`, the limited thermal
conductivity of the top snow layer weakens the flux of energy into the
snow layer and fewer snow melts.  The snow surface temperature rises
faster than the bulk temperature, and the higher surface temperature results
in smaller sensible (|WSensSnow|) and latent (|WLatSnow|) heat fluxes
from the atmosphere into the snow layer and to larger longwave radiation
losses (|NetLongwaveRadiationSnow|):

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> conditions_acker_winter_ktschnee = test(
    ...     "lland_v3_acker_winter_ktschnee_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |      wada |       qdb | qib1 | qib2 | qbb | qkap |      qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |     qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.483872 |                  2.483872 |                   2.483872 |                   4.01525 |                      0.0 |    -1.531378 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |      0.0 |      0.0 |  1.808179 |       0.0 |  0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                     0.387464 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -0.391921 |              0.106229 |                  0.106229 |                   0.106229 |                  1.277835 |                  2.47724 |    -1.171606 |        -2.371011 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 |   10.374 |   10.374 |  0.779906 |      0.06 |      0.0 | -0.140764 | -0.065045 |     -6.07759 |     0.875479 |       0.0 |       0.0 | 0.441058 |      0.0 |  -0.37548 | -2.305184 | -0.309653 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 10.514764 | 10.514764 | -0.147313 | 0.084415 | -1.131599 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                     0.324339 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -3.771996 |  1.092644 |              0.199858 |                  0.199858 |                   0.199858 |                  1.329454 |                 2.790348 |    -1.129596 |        -2.590489 |         -1.129596 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |    0.316 |    0.316 |  0.039485 |  0.039485 |      0.0 | -0.159024 | -0.003698 |    -8.367514 |     0.861349 |       0.0 |       0.0 | 0.398488 |      0.0 | -0.424188 | -1.091574 |  1.074727 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.020515 | 10.989788 | 10.989788 | -0.133095 | 0.139861 | -1.482643 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                     0.331306 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 | -4.942143 |  0.733022 |               0.41176 |                   0.41176 |                    0.41176 |                  3.727317 |                 4.937974 |    -3.315556 |        -4.526213 |         -3.315556 |             49.536202 |                 100.0 |               112.102145 |               99.524551 | 1.088515 | 1.088515 |  0.092461 |      0.06 |      0.0 | -0.369245 | -0.008645 |    -8.096278 |     0.850169 |       0.0 |       0.0 | 0.480397 |      0.0 | -0.984942 | -2.789537 |  0.751735 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 12.447547 | 12.447547 | -0.160452 | 0.187922 | -1.474065 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                     0.523188 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 |  -4.91355 | -0.666518 |              0.128581 |                  0.128581 |                   0.128581 |                  1.140917 |                 1.700561 |    -1.012336 |         -1.57198 |         -1.012336 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |    5.956 |   2.6802 |  0.135129 |      0.06 |      0.0 | -0.304322 |  0.003863 |    -2.103176 |     0.843881 |       0.0 |       0.0 | 0.492387 | 0.492387 | -0.811763 | -1.418866 |  -0.65865 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 15.878429 | 18.707869 |       0.0 | 0.216776 | -0.065947 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                     0.493463 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 | -0.219823 |  0.595141 |              0.263213 |                  0.263213 |                   0.263213 |                  1.217715 |                 2.201596 |    -0.954501 |        -1.938383 |         -0.954501 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |      0.0 |      0.0 |  0.000991 |       0.0 |      0.0 | -0.186366 |       0.0 |    -2.888087 |     0.811266 |       0.0 |       0.0 | 0.184424 | 0.184424 | -0.497123 | -0.784521 |  0.656739 | 0.996987 | 0.298195 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |  16.22287 | 18.894235 |       0.0 | 0.392168 |  0.080512 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                     0.429685 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  1.062786 |               0.17276 |                   0.17276 |                    0.17276 |                  1.193664 |                 2.428029 |    -1.020904 |        -2.255269 |         -1.020904 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |      0.0 |      0.0 | -0.015747 | -0.015747 |      0.0 | -0.150228 |       0.0 |    -4.724262 |     0.785949 |       0.0 |       0.0 | 0.276769 | 0.276769 | -0.400725 | -0.699317 |  1.155226 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.015747 | 16.630827 | 19.044463 |       0.0 |  0.56756 | -0.240674 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                     0.424191 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 | -0.802246 |  0.901846 |              0.187367 |                  0.187367 |                   0.187367 |                  1.194208 |                 2.367263 |     -1.00684 |        -2.179896 |          -1.00684 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |      0.0 |      0.0 | -0.012511 | -0.012511 |      0.0 | -0.158168 |       0.0 |    -4.893505 |     0.765728 |       0.0 |       0.0 | 0.341438 | 0.341438 | -0.421905 | -0.742105 |  1.015886 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.028258 | 17.113222 | 19.202631 |       0.0 | 0.742952 | -0.400919 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                     0.381219 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 | -1.336398 |  1.082838 |              0.711106 |                  0.711106 |                   0.711106 |                  4.005891 |                 4.970801 |    -3.294784 |        -4.259695 |         -3.294784 |             78.890988 |                 100.0 |               112.102145 |               99.513532 | 0.626258 | 0.626258 | -0.234493 | -0.234493 |      0.0 | -0.391985 | -0.002749 |    -6.288734 |     0.753624 |       0.0 |       0.0 | 0.469924 | 0.469924 |   -1.0456 |  -1.97705 |  1.237044 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.294493 | 18.569375 | 20.220873 |       0.0 | 0.868064 | -0.742157 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                     0.412084 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 | -2.473858 |  0.568831 |              0.313857 |                  0.313857 |                   0.313857 |                  1.286731 |                 2.586023 |    -0.972874 |        -2.272166 |         -0.972874 |             71.001889 |                 100.0 |               112.102145 |               99.511814 | 3.336493 | 3.336493 |     0.137 |      0.06 |      0.0 | -0.164445 | -0.021617 |    -5.273465 |     0.767009 |       0.0 |       0.0 | 0.563446 | 0.563446 | -0.438649 | -1.098112 |  0.735405 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 22.626164 | 23.721812 |       0.0 | 0.730687 | -0.569388 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 |  -1.89796 | -0.983902 |              0.344426 |                  0.344426 |                   0.344426 |                   1.16859 |                -0.065197 |    -0.824163 |         0.409623 |         -0.824163 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |    8.776 |      0.0 |  1.032826 |      0.06 |  0.33884 | -1.215856 |  0.184351 |          0.0 |     0.749149 | 23.702487 | 22.626164 |      0.0 |      0.0 | -3.243236 | -5.063324 | -8.716183 | 0.956732 | 0.274601 | 33.713668 | 11.061268 |  0.0 |  0.0 | 0.0 |  0.0 | 11.061268 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 | 0.918992 |  1.156114 | 93.759821 | 11.061268 |   0.0 |   0.0 |   0.0 |  0.0 | 2.847686 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                      0.50495 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |  0.607758 |               0.40701 |                   0.40701 |                    0.40701 |                  1.733999 |                 3.007697 |    -1.326988 |        -2.600687 |         -1.326988 |             37.369415 |                 100.0 |                81.666498 |               98.491327 |    0.128 |   0.0576 |  1.429271 |      0.06 | 0.519503 |  0.082368 |  0.000045 |    -2.579932 |     0.735699 |       0.0 |       0.0 | 0.000787 | 0.000787 |  0.243635 | -2.236256 |  0.608066 | 0.951723 | 0.271733 |  0.045632 |  0.015061 |  0.0 |  0.0 | 0.0 |  0.0 |  0.015061 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 | 1.090255 |  1.289956 | 93.270889 |  0.015061 |   0.0 |   0.0 |   0.0 |  0.0 | 3.834601 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                     0.397724 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |  1.354569 |              0.191046 |                  0.191046 |                   0.191046 |                  1.254261 |                 2.684039 |    -1.063215 |        -2.492993 |         -1.063215 |             96.820757 |                 100.0 |                82.308727 |               98.520788 |    2.666 |    2.666 |  0.193623 |      0.06 |      0.0 | -0.079257 | -0.019502 |    -5.737115 |     0.752701 |       0.0 |       0.0 | 0.043591 |      0.0 | -0.211415 | -0.931952 |  1.349626 | 0.974664 | 0.284991 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |  2.745257 |  2.745257 | -0.014559 | 0.878342 |  0.676988 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.049333 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                     0.368082 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |  1.587287 |              0.433394 |                  0.433394 |                   0.433394 |                  2.107297 |                 3.808282 |    -1.673903 |        -3.374889 |         -1.673903 |             68.711505 |                 100.0 |                82.308727 |               98.521225 |      0.0 |      0.0 |   0.27977 |       0.0 |      0.0 | -0.118703 |       0.0 |    -6.743136 |     0.744961 |       0.0 |       0.0 | 0.052598 |      0.0 | -0.316633 |  -1.46796 |  1.590296 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |   2.86396 |   2.86396 | -0.017568 | 0.969233 | -0.168699 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.093546 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                     0.288286 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 | -0.562332 |  2.184786 |              0.606631 |                  0.606631 |                   0.606631 |                  3.397964 |                 5.076196 |    -2.791333 |        -4.469565 |         -2.791333 |             76.073452 |                 100.0 |                82.308727 |                98.52159 |      0.0 |      0.0 | -0.032058 | -0.032058 |      0.0 | -0.207747 |       0.0 |    -9.859707 |     0.738324 |       0.0 |       0.0 | 0.085606 |      0.0 | -0.554154 |   -1.7196 |  2.195811 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.032058 |  3.071707 |  3.071707 | -0.028592 | 1.054479 | -1.611886 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.583528 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                     0.393621 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -5.372953 |  0.099031 |              0.153156 |                  0.153156 |                   0.153156 |                  1.211777 |                  2.37426 |    -1.058622 |        -2.221104 |         -1.058622 |             57.569099 |                 100.0 |                82.308727 |               98.521883 | 6.270058 | 6.270058 |  0.033918 |  0.033918 |      0.0 | -0.264407 | -0.045865 |    -5.872332 |     0.799424 |       0.0 |       0.0 | 0.325248 |      0.0 | -0.705293 | -1.382605 |  0.133206 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.026082 |  9.606172 |  9.606172 | -0.108633 | 0.469049 | -0.969317 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.311377 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                     0.279974 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -3.231056 |  1.637802 |              0.159665 |                  0.159665 |                   0.159665 |                  1.287913 |                 2.905395 |    -1.128248 |         -2.74573 |         -1.128248 |            236.672963 |                 100.0 |                82.308727 |               98.522102 |      0.0 |      0.0 | -0.002006 | -0.002006 |      0.0 | -0.108916 |       0.0 |   -10.227129 |     0.790417 |       0.0 |       0.0 |  0.38413 |      0.0 | -0.290529 | -0.797733 |  1.657468 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.028088 |  9.715089 |  9.715089 | -0.128299 | 0.533462 | -1.865519 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166154 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                     0.303291 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -6.218396 |  0.693372 |              0.164972 |                  0.164972 |                   0.164972 |                  1.257034 |                 2.432406 |    -1.092062 |        -2.267434 |         -1.092062 |            163.850513 |                 100.0 |                82.308727 |               98.522249 | 0.062088 | 0.062088 | -0.003574 | -0.003574 |      0.0 | -0.174394 | -0.000753 |    -9.219622 |     0.783126 |       0.0 |       0.0 | 0.462275 |      0.0 | -0.465188 | -1.083526 |   0.71872 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.063574 |  9.951571 |  9.951571 |   -0.1544 | 0.589899 | -1.817291 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088662 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                     0.262451 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -6.057636 |  1.162802 |              0.169867 |                  0.169867 |                   0.169867 |                  1.301361 |                 2.956332 |    -1.131493 |        -2.786464 |         -1.131493 |            112.108246 |                 100.0 |                82.308727 |               98.522322 | 0.003574 | 0.003574 |    0.0319 |    0.0319 |      0.0 | -0.151375 | -0.000059 |   -11.034606 |     0.776535 |       0.0 |       0.0 | 0.517085 |      0.0 | -0.403786 | -1.201629 |   1.18105 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |   0.0281 |  10.10652 |  10.10652 | -0.172706 |  0.64461 | -2.238493 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047311 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                     0.309875 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -7.461643 |  0.350156 |               0.30497 |                   0.30497 |                    0.30497 |                  1.393316 |                 2.690073 |    -1.088346 |        -2.385102 |         -1.088346 |             88.752361 |                 100.0 |                82.308727 |               98.522322 |   0.1561 |   0.1561 |  0.019885 |  0.019885 |      0.0 | -0.228727 |  -0.00186 |    -8.947562 |     0.771755 |       0.0 |       0.0 |  0.52255 |      0.0 | -0.610118 | -1.424864 |  0.350121 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.040115 | 10.491346 | 10.491346 | -0.174532 | 0.686699 | -1.847048 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.025246 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                     0.303545 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 | -6.156828 |  0.723701 |              0.359506 |                  0.359506 |                   0.359506 |                  1.820398 |                 3.207669 |    -1.460893 |        -2.848164 |         -1.460893 |             92.611159 |                 100.0 |                82.308727 |               98.522249 | 0.168115 | 0.168115 |  0.033896 |  0.033896 |      0.0 | -0.220897 | -0.001968 |     -9.20903 |     0.767279 |       0.0 |       0.0 | 0.503786 |      0.0 | -0.589232 | -1.543465 |  0.715467 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.026104 | 10.880359 | 10.880359 | -0.168265 |  0.72812 |  -1.82915 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.013471 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                     0.352899 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 | -6.097166 |  0.284926 |              0.168405 |                  0.168405 |                   0.168405 |                  1.259735 |                 2.716552 |    -1.091329 |        -2.548146 |         -1.091329 |              46.30558 |                 100.0 |                82.308727 |               98.522102 | 2.222104 | 2.222104 |  0.237772 |      0.06 |      0.0 | -0.215407 | -0.022292 |    -7.286572 |     0.779063 |       0.0 |       0.0 |  0.54496 |      0.0 | -0.574588 | -1.697173 |  0.276386 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |  13.31787 |  13.31787 | -0.182017 |  0.62318 | -1.372475 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.007188 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                     0.498385 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -4.574918 |  -0.41133 |              0.142507 |                  0.142507 |                   0.142507 |                  1.179808 |                 2.283002 |    -1.037301 |        -2.140495 |         -1.037301 |             24.768101 |                 100.0 |                82.308727 |               98.521883 |     6.52 |    5.053 |  0.461431 |      0.06 |      0.0 | -0.270034 | -0.018373 |    -2.755289 |      0.81351 |       0.0 |       0.0 | 0.504238 | 0.504238 | -0.720303 | -1.863497 | -0.443304 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 19.132036 | 20.107904 |       0.0 | 0.378495 | -0.219545 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003836 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                     0.428712 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 | -0.731818 |  0.889043 |              0.162157 |                  0.162157 |                   0.162157 |                  1.273508 |                 2.903767 |     -1.11135 |         -2.74161 |          -1.11135 |             28.027061 |                 100.0 |                82.308727 |                98.52159 |    0.128 |    0.128 |  1.132327 |      0.06 |      0.0 |  0.051362 | -0.000776 |    -4.754101 |     0.788497 |       0.0 |       0.0 | 0.324817 | 0.324817 |  0.137005 | -1.881858 |  0.996757 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 19.535144 | 20.184542 |       0.0 | 0.547933 | -0.366989 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002047 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                     0.233415 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -1.223296 |   2.50954 |              0.556949 |                  0.556949 |                   0.556949 |                  3.730222 |                 5.691508 |    -3.173272 |        -5.134559 |         -3.173272 |             56.054123 |                 100.0 |                82.308727 |               98.521225 |      0.0 |      0.0 |   0.14591 |       0.0 |      0.0 | -0.131527 |       0.0 |    -12.48465 |     0.767786 |       0.0 |       0.0 | 0.790662 | 0.649398 | -0.350843 | -2.010095 |  2.773621 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 |  20.31607 |  20.31607 | -0.047182 | 0.723325 | -2.134929 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001092 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                     0.197909 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 |  -7.11643 |  1.529174 |              0.658862 |                  0.658862 |                   0.658862 |                  4.423241 |                  5.15027 |     -3.76438 |        -4.491409 |          -3.76438 |            142.003778 |                 100.0 |                82.308727 |               98.520788 |    1.068 |    1.068 |  -0.16717 |  -0.16717 |      0.0 | -0.217678 | -0.019196 |   -14.493732 |      0.76412 |       0.0 |       0.0 | 1.346967 |      0.0 | -0.580645 | -1.998081 |  1.912682 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |  0.22717 | 21.601748 | 21.601748 | -0.449887 | 0.758599 | -2.922503 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000583 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                     0.296296 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -9.741677 | -0.050668 |               0.18107 |                   0.18107 |                    0.18107 |                  1.247575 |                 2.051073 |    -1.066505 |        -1.870003 |         -1.066505 |            163.850513 |                 100.0 |                82.308727 |                98.52028 |  1.01317 |  1.01317 | -0.001974 | -0.001974 |      0.0 |  -0.22141 | -0.011223 |    -9.514411 |     0.767261 |       0.0 |       0.0 | 1.364738 |      0.0 | -0.590599 |  -1.33536 | -0.055955 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.061974 | 22.836327 | 22.836327 | -0.455822 |  0.72829 | -2.130235 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000311 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                     0.235845 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 | -7.100784 |  1.240127 |              0.594897 |                  0.594897 |                   0.594897 |                    3.6896 |                 4.987053 |    -3.094703 |        -4.392156 |         -3.094703 |             71.001889 |                 100.0 |                82.308727 |                 98.5197 | 0.001974 | 0.001974 | -0.120535 | -0.120535 |      0.0 | -0.300552 | -0.000032 |    -12.35734 |     0.762428 |       0.0 |       0.0 | 1.356648 |      0.0 | -0.801707 | -2.353055 |  1.237393 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.180535 | 23.138853 | 23.138853 | -0.453121 | 0.775366 | -2.628762 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000166 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                     0.232729 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 |  -8.76254 |  0.857788 |              0.190324 |                  0.190324 |                   0.190324 |                  1.325064 |                  3.12786 |     -1.13474 |        -2.937536 |          -1.13474 |             78.890988 |                 100.0 |                82.308727 |                98.51905 | 0.120535 | 0.120535 |  0.073326 |      0.06 |      0.0 | -0.172968 | -0.002318 |   -12.520818 |     0.758665 |       0.0 |       0.0 | 1.516178 |      0.0 | -0.461382 |   -1.5674 |  0.908753 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan |      0.0 | 23.432356 | 23.432356 | -0.506403 |  0.81384 |  -2.74495 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000089 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                     0.253654 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -9.149832 |  0.547134 |              0.194397 |                  0.194397 |                   0.194397 |                  1.292505 |                 2.540669 |    -1.098108 |        -2.346272 |         -1.098108 |            152.146905 |                 100.0 |                82.308727 |               98.518331 |    0.128 |    0.128 |  0.019406 |  0.019406 |      0.0 | -0.177762 |  -0.00198 |   -11.458198 |     0.755438 |       0.0 |       0.0 | 1.504547 |      0.0 |  -0.47417 | -1.330832 |  0.541269 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.040594 | 23.738117 | 23.738117 | -0.502519 | 0.848175 | -2.550483 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000047 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                     0.494633 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 | -8.501611 | -1.235869 |              0.688094 |                  0.688094 |                   0.688094 |                  3.475519 |                 4.119004 |    -2.787425 |         -3.43091 |         -2.787425 |             50.715635 |                 100.0 |                82.308727 |               98.517542 | 0.826594 | 0.061995 | -0.125295 | -0.125295 |      0.0 | -0.724213 |  0.005662 |    -2.856407 |     0.751606 |       0.0 |       0.0 | 0.969722 |   0.7646 | -1.931802 | -2.907946 | -1.408838 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | nan | nan | 0.185295 | 25.288925 | 25.288925 | -0.068511 | 0.890684 | -0.573015 | 93.270889 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000025 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

>>> ktschnee(inf)

hourly simulation
_________________

Now we change the simulation time step size from one day to one hour.
Consequently, we need to change the values of all time-dependent control
parameters by the factor 1/24 (or, in single cases, 24).  Also, we need
to |FixedParameter.restore| all time-dependent fixed parameter values:

>>> pub.timegrids = "1997-08-03", "1997-08-06", "1h"
>>> for parameter in control:
...     if parameter.TIME == True:
...         parameter.value = parameter.value/24
...     if parameter.TIME == False:
...         parameter.value = parameter.value*24
>>> for parameter in fixed:
...     parameter.restore()
>>> parameterstep("1h")
>>> test = IntegrationTest(land)

.. _lland_v3_acker_summer_hourly:

acre (summer)
-------------

The following input values agree with the ones of the respective sub-period
of the daily integration test :ref:`lland_v3_acker_summer_daily`:

>>> inputs.nied.series = 0.0
>>> inputs.teml.series = (
...     16.9, 16.6, 16.4, 16.3, 16.0, 15.9, 16.0, 16.6, 17.4, 19.0, 20.3,
...     21.4, 21.3, 21.8, 22.9, 22.7, 22.5, 21.9, 21.4, 20.7, 19.4, 17.8,
...     17.0, 16.4, 15.9, 15.4, 15.2, 14.9, 14.3, 14.1, 14.3, 15.4, 16.8,
...     18.2, 19.5, 20.3, 21.0, 21.8, 21.9, 21.8, 22.2, 21.8, 20.9, 19.7,
...     17.9, 15.8, 14.8, 14.1, 13.4, 12.8, 12.3, 12.0, 11.7, 11.6, 12.6,
...     14.8, 17.1, 19.4, 20.6, 21.4, 22.5, 23.1, 23.4, 23.8, 24.0, 24.0,
...     23.7, 22.8, 21.3, 19.3, 18.3, 17.4)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5, 0.7, 0.8, 0.5,
...     0.4, 0.5, 0.5, 0.3, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.9, 1.0, 1.0, 0.9, 0.8, 0.9, 0.8, 0.9, 0.9, 0.9, 1.0,
...     1.0, 1.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.9,
...     1.0, 0.9, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     0.8, 0.8, 0.8, 0.8, 0.8, 0.6, 0.9, 0.9, 0.9, 1.3, 1.5, 1.2, 1.3, 1.5,
...     1.9, 1.9, 2.3, 2.4, 2.5, 2.5, 2.2, 1.7, 1.7, 2.3, 2.3, 2.2, 2.3, 2.0,
...     2.3, 2.5, 2.4, 2.5, 2.5, 2.9, 3.1, 3.3, 3.3, 3.2, 2.5, 2.9, 3.6, 4.2,
...     4.2, 3.6, 3.2, 2.2, 2.2, 2.1, 1.8, 1.8, 1.6, 1.2, 1.9, 1.3, 1.6, 1.9,
...     3.2, 4.0, 4.6, 4.1, 4.1, 4.4, 4.6, 4.5, 4.7, 4.5, 4.1, 3.8, 2.5, 1.9,
...     2.2, 2.7)
>>> inputs.relativehumidity.series = (
...     95.1, 94.9, 95.9, 96.7, 97.2, 97.5, 97.7, 97.4, 96.8, 86.1, 76.8,
...     71.8, 67.5, 66.1, 63.4, 62.4, 61.1, 62.1, 67.0, 74.5, 81.2, 86.9,
...     90.1, 90.9, 88.0, 89.2, 88.8, 89.7, 93.0, 93.6, 94.6, 93.3, 86.9,
...     78.9, 75.8, 73.4, 68.3, 61.3, 60.6, 58.8, 57.2, 57.2, 59.9, 66.1,
...     71.7, 80.6, 85.9, 89.3, 91.4, 93.1, 94.6, 95.6, 96.2, 95.7, 95.1,
...     96.3, 89.9, 79.2, 73.4, 67.3, 62.2, 59.6, 55.4, 52.1, 49.5, 48.1,
...     49.4, 57.0, 65.9, 73.0, 76.7, 81.8)
>>> inputs.atmosphericpressure.series = (
...     101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.6, 101.6,
...     101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6,
...     101.6, 101.6, 101.7, 101.7, 101.7, 101.6, 101.6, 101.6, 101.6, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8, 101.8, 101.8, 101.8, 101.9, 101.9, 101.9, 101.9, 101.9,
...     101.9, 101.9, 101.9, 101.9, 101.8, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8)
>>> inlet.sequences.sim.series = 0.0


In this and the following "summer" examples, we use the initial conditions
of August 3rd, calculated by the corresponding daily integration tests.
However, this is possible for the state sequences, but not for the (now
higher resolved) log sequences.  Hence, we need to define them manually:

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2,
...        0.1, 0.2, 0.1, 0.2, 0.2, 0.3, 0.0, 0.0,
...        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.2, 0.5,
...       0.8, 1.1, 1.2, 1.4, 1.9, 1.6, 0.9, 0.8,
...       0.6, 0.4, 0.2, 0.1, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [13.2, 13.2, 13.1, 12.6, 12.7, 13.0, 13.5, 14.8,
...       16.2, 17.7, 18.8, 19.4, 20.4, 21.0, 21.5, 21.2,
...       20.4, 20.7, 20.2, 19.7, 19.0, 18.0, 17.5, 17.1]),
...     (logs.loggedrelativehumidity,
...      [95.1, 94.5, 94.8, 96.4, 96.6, 97.1, 97.1, 96.7,
...       92.2, 88.5, 81.1, 76.5, 75.1, 70.8, 68.9, 69.2,
...       75.0, 74.0, 77.4, 81.4, 85.3, 90.1, 92.3, 93.8]),
...     (logs.loggedwindspeed2m,
...      [0.8, 1.0, 1.2, 1.3, 0.9, 1.1, 1.3, 1.3,
...       1.9, 2.2, 1.8, 2.3, 2.4, 2.5, 2.4, 2.5,
...       2.1, 2.2, 1.7, 1.7, 1.3, 1.3, 0.7, 0.8]))

This integration test deals with a dry situation.  Hence, the soil water
content (|BoWa|) shows a pronounced decline.  This decline is sharpest around
noon, when evapotranspiration out of the soil (|EvB|) reaches its maximum:

.. integration-test::

    >>> test("lland_v3_acker_summer_hourly",
    ...      axis1=(fluxes.evb, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   7.982856 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |           7.08166 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 | -0.002965 | -0.002965 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.256857 | 0.071349 | 0.002965 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.256857 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.071349 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   7.982856 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          7.081012 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 | -0.002685 | -0.002685 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.250222 | 0.069506 |  0.00565 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.250222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069506 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   7.982856 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |            7.0805 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 | -0.004107 | -0.004107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.243758 | 0.067711 | 0.009757 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.243758 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.067711 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   7.982856 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          7.080492 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 | -0.005232 | -0.005232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.237462 | 0.065962 | 0.014989 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.237462 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065962 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.005318 |                  0.005318 |                   7.982856 |                  0.902341 |                      0.0 |     -0.03228 |              0.0 |          7.080514 |            176.542034 |                 100.0 |               100.629116 |               92.771119 |  0.0 |  0.0 | -0.004819 | -0.004819 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.231328 | 0.064258 | 0.019808 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.231328 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064258 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.060007 |                  0.060007 |                   7.982856 |                  0.901482 |                      0.0 |     0.022445 |              0.0 |          7.081373 |            235.389378 |                 100.0 |               100.629116 |              100.447767 |  0.0 |  0.0 |  0.005073 |  0.005073 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.225353 | 0.062598 | 0.014735 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.225353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062598 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.156653 |                  0.156653 |                   7.982856 |                   0.89977 |                      0.0 |     0.119163 |              0.0 |          7.083086 |            156.926252 |                 100.0 |               100.629116 |              100.447767 |  0.0 |  0.0 |  0.026292 |  0.014735 | 0.008821 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.219531 | 0.060981 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 79.614365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.219531 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060981 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.299135 |                  0.299135 |                   7.982856 |                  0.896826 |                      0.0 |     0.261767 |              0.0 |          7.086029 |            156.926252 |                 100.0 |               100.641253 |               100.45639 |  0.0 |  0.0 |  0.057456 |       0.0 | 0.044017 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.213861 | 0.059406 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 79.570349 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.213861 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.059406 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.467717 |                  0.467717 |                   7.982856 |                  0.891371 |                      0.0 |     0.430576 |              0.0 |          7.091484 |            156.926252 |                 100.0 |               100.701823 |              100.499412 |  0.0 |  0.0 |  0.096099 |       0.0 | 0.073977 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.208337 | 0.057871 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 79.496372 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.208337 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.057871 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.852983 |                  0.852983 |                    8.15863 |                  0.922604 |                      0.0 |     0.814541 |              0.0 |          7.236026 |            108.641252 |                 100.0 |               100.803646 |              100.571701 |  0.0 |  0.0 |  0.222763 |       0.0 | 0.161018 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.202955 | 0.056376 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 79.335354 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.202955 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056376 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.372332 |                  1.372332 |                   8.334404 |                  0.955299 |                      0.0 |     1.332528 |              0.0 |          7.379106 |             94.155751 |                 100.0 |               101.025386 |              100.728981 |  0.0 |  0.0 |  0.389923 |       0.0 | 0.275567 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.197713 |  0.05492 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 79.059786 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.197713 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05492 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.682691 |                  1.682691 |                   8.773841 |                  1.039981 |                      0.0 |     1.639359 |              0.0 |          7.733859 |            117.694689 |                 100.0 |               101.405231 |              100.997942 |  0.0 |  0.0 |   0.46628 |       0.0 | 0.348858 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.192605 | 0.053502 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 78.710929 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.192605 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053502 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.714225 |                  1.714225 |                   9.301164 |                  1.142661 |                      0.0 |     1.666614 |              0.0 |          8.158503 |            108.641252 |                 100.0 |               101.886758 |              101.338062 |  0.0 |  0.0 |  0.491066 |       0.0 | 0.360852 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18763 |  0.05212 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 78.350077 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18763 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05212 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.357421 |                  1.357421 |                   9.652713 |                  1.211985 |                      0.0 |     1.306922 |              0.0 |          8.440727 |             94.155751 |                 100.0 |               102.385622 |              101.689443 |  0.0 |  0.0 |  0.424802 |       0.0 |  0.30342 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.182784 | 0.050773 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 78.046657 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.182784 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050773 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.147809 |                  1.147809 |                   9.828487 |                   1.24753 |                      0.0 |     1.095829 |              0.0 |          8.580957 |             74.333488 |                 100.0 |               102.805714 |              101.984567 |  0.0 |  0.0 |  0.415458 |       0.0 | 0.282559 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.178062 | 0.049462 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.764098 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.178062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.049462 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |               1.06126 |                   1.06126 |                  10.180036 |                  1.316893 |                      0.0 |     1.006389 |              0.0 |          8.863143 |             74.333488 |                 100.0 |               103.197445 |              102.259131 |  0.0 |  0.0 |  0.393508 |       0.0 | 0.266848 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.173463 | 0.048184 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 |  77.49725 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.173463 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.048184 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.761967 |                  0.761967 |                  10.443698 |                  1.370145 |                      0.0 |     0.704878 |              0.0 |          9.073553 |             61.405925 |                 100.0 |               103.567861 |              102.518193 |  0.0 |  0.0 |  0.342513 |       0.0 |   0.2192 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.168982 | 0.046939 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 77.278049 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.168982 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046939 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.375211 |                  0.375211 |                   10.70736 |                  1.421011 |                      0.0 |     0.316002 |              0.0 |          9.286348 |             58.847345 |                 100.0 |                103.87248 |               102.73083 |  0.0 |  0.0 |  0.236583 |       0.0 | 0.148202 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.164617 | 0.045727 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 77.129847 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.164617 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045727 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.139757 |                  0.139757 |                  10.795247 |                  1.433696 |                      0.0 |     0.080019 |              0.0 |           9.36155 |             56.493451 |                 100.0 |               104.078611 |              102.874509 |  0.0 |  0.0 |  0.156446 |       0.0 | 0.096042 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.160365 | 0.044546 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 77.033805 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.160365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044546 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.037299 |                  0.037299 |                  10.795247 |                  1.423756 |                      0.0 |    -0.022024 |              0.0 |          9.371491 |             56.493451 |                 100.0 |                104.21227 |              102.967584 |  0.0 |  0.0 |   0.09619 |       0.0 |   0.0585 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.156222 | 0.043395 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 76.975305 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.156222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043395 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000509 |                  0.000509 |                  10.795247 |                  1.412078 |                      0.0 |    -0.058328 |              0.0 |          9.383169 |             64.197103 |                 100.0 |               104.293713 |               89.533534 |  0.0 |  0.0 |   0.04938 |       0.0 | 0.032417 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.152187 | 0.042274 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 76.942888 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.152187 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042274 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  10.795247 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          9.393683 |             83.078604 |                 100.0 |               104.338854 |               87.719298 |  0.0 |  0.0 |  0.019258 |       0.0 | 0.013442 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.148256 | 0.041182 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 76.929445 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.148256 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041182 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  10.795247 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          9.404089 |             83.078604 |                 100.0 |               104.357574 |               87.719298 |  0.0 |  0.0 |  0.010298 |       0.0 | 0.007136 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.144426 | 0.040118 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 |  76.92231 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.144426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040118 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  10.795247 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          9.412407 |             61.405925 |                 100.0 |               104.367512 |               87.719298 |  0.0 |  0.0 |  0.014661 |       0.0 | 0.009245 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.140696 | 0.039082 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.913065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039082 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  10.739718 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          9.343179 |             61.405925 |                 100.0 |               104.380388 |               87.719298 |  0.0 |  0.0 |  0.023989 |       0.0 | 0.015041 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.137061 | 0.038073 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.898024 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.137061 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038073 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  10.739718 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          9.333085 |             64.197103 |                 100.0 |               104.401338 |               87.719298 |  0.0 |  0.0 |  0.018182 |       0.0 |   0.0115 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.133521 | 0.037089 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.886524 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.133521 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037089 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  10.739718 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |           9.32165 |             61.405925 |                 100.0 |               104.417357 |               87.719298 |  0.0 |  0.0 |   0.02055 |       0.0 |  0.01278 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.130072 | 0.036131 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 76.873744 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.130072 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036131 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  10.739718 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          9.309548 |             70.616813 |                 100.0 |               104.435159 |               87.719298 |  0.0 |  0.0 |   0.01363 |       0.0 | 0.008835 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.126712 | 0.035198 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.864909 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.126712 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035198 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.004352 |                  0.004352 |                  10.739718 |                  1.440702 |                      0.0 |    -0.055678 |              0.0 |          9.299016 |             61.405925 |                 100.0 |               104.447467 |               93.267161 |  0.0 |  0.0 |  0.007255 |       0.0 | 0.004359 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.123439 | 0.034289 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 |  76.86055 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.123439 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034289 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.054586 |                  0.054586 |                  10.739718 |                  1.451305 |                      0.0 |    -0.005885 |              0.0 |          9.288414 |             56.493451 |                 100.0 |               104.453539 |              103.135415 |  0.0 |  0.0 |  0.018162 |       0.0 | 0.010113 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.120251 | 0.033403 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.850437 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.120251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033403 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.172329 |                  0.172329 |                   10.82738 |                  1.478175 |                      0.0 |     0.110738 |              0.0 |          9.349205 |             58.847345 |                 100.0 |               104.467628 |              103.145209 |  0.0 |  0.0 |   0.04093 |       0.0 | 0.023217 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.117144 |  0.03254 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.827221 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.117144 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03254 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |              0.781884 |                  0.781884 |                  11.616331 |                   1.64464 |                      0.0 |     0.713358 |              0.0 |          9.971692 |             56.493451 |                 100.0 |               104.499975 |              103.167691 |  0.0 |  0.0 |   0.18972 |       0.0 | 0.107491 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.114118 |   0.0317 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.719729 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114118 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0317 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.300216 |                  1.300216 |                  12.492945 |                   1.83472 |                      0.0 |      1.22377 |              0.0 |         10.658225 |             56.493451 |                 100.0 |               104.649787 |               103.27176 |  0.0 |  0.0 |  0.343734 |       0.0 | 0.198364 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.111171 | 0.030881 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.521365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.111171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030881 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.679922 |                  1.679922 |                  13.194235 |                  1.989515 |                      0.0 |     1.597026 |              0.0 |          11.20472 |             48.701251 |                 100.0 |               104.926453 |              103.463719 |  0.0 |  0.0 |  0.494183 |       0.0 | 0.274461 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.108299 | 0.030083 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.246905 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.108299 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030083 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              1.761455 |                  1.761455 |                   13.54488 |                   2.06621 |                      0.0 |     1.675363 |              0.0 |          11.47867 |             45.559235 |                 100.0 |               105.309691 |              103.729123 |  0.0 |  0.0 |  0.550449 |       0.0 | 0.303114 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.105502 | 0.029306 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 75.943791 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.105502 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029306 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |               1.64232 |                   1.64232 |                  13.632542 |                  2.087634 |                      0.0 |     1.555335 |              0.0 |         11.544908 |             42.798069 |                 100.0 |               105.733535 |              104.021981 |  0.0 |  0.0 |  0.548384 |       0.0 | 0.297669 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.102776 | 0.028549 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 75.646122 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.102776 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028549 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.697727 |                  1.697727 |                  13.720203 |                  2.106001 |                      0.0 |     1.609977 |              0.0 |         11.614202 |             42.798069 |                 100.0 |               106.150385 |              104.309322 |  0.0 |  0.0 |  0.597559 |       0.0 | 0.327471 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.100122 | 0.027812 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.318651 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.100122 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027812 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |              1.640072 |                  1.640072 |                  13.983187 |                  2.166369 |                      0.0 |     1.549806 |              0.0 |         11.816818 |             44.135508 |                 100.0 |               106.609685 |              104.625144 |  0.0 |  0.0 |  0.621801 |       0.0 | 0.348686 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.097535 | 0.027093 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 74.969965 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.097535 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027093 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |              1.678676 |                  1.678676 |                  14.421494 |                  2.265756 |                      0.0 |      1.58427 |              0.0 |         12.155738 |             56.493451 |                 100.0 |               107.099579 |              104.961105 |  0.0 |  0.0 |  0.589183 |       0.0 | 0.361258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.095016 | 0.026393 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 74.608707 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.095016 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026393 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.414805 |                  1.414805 |                  14.772139 |                  2.348437 |                      0.0 |     1.316953 |              0.0 |         12.423701 |             48.701251 |                 100.0 |               107.608064 |              105.308836 |  0.0 |  0.0 |  0.551424 |       0.0 | 0.320065 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.092562 | 0.025712 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 | 74.288642 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.092562 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025712 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.033761 |                  1.033761 |                  15.122784 |                  2.428674 |                      0.0 |     0.932566 |              0.0 |          12.69411 |             39.231563 |                 100.0 |               108.059369 |              105.616633 |  0.0 |  0.0 |  0.505761 |       0.0 | 0.270996 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.090171 | 0.025047 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.017646 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.090171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025047 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.690032 |                  0.690032 |                  15.736413 |                  2.564427 |                      0.0 |     0.583181 |              0.0 |         13.171986 |             33.627054 |                 100.0 |               108.442082 |               105.87704 |  0.0 |  0.0 |  0.444603 |       0.0 | 0.220974 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.087841 |   0.0244 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 73.796672 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.087841 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0244 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.351786 |                  0.351786 |                  16.525365 |                  2.744116 |                      0.0 |     0.237448 |              0.0 |         13.781249 |             33.627054 |                 100.0 |               108.754562 |              106.089245 |  0.0 |  0.0 |   0.32478 |       0.0 | 0.158718 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.085572 |  0.02377 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 73.637953 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.085572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02377 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.115111 |                  0.115111 |                  17.401978 |                  2.950104 |                      0.0 |     -0.00781 |              0.0 |         14.451875 |             39.231563 |                 100.0 |               108.979235 |              106.241591 |  0.0 |  0.0 |  0.183091 |       0.0 | 0.093928 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.083362 | 0.023156 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 73.544026 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083362 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023156 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002224 |                  0.002224 |                  17.664962 |                  3.034376 |                      0.0 |    -0.124208 |              0.0 |         14.630586 |             44.135508 |                 100.0 |               109.112286 |               89.357093 |  0.0 |  0.0 |  0.100631 |       0.0 | 0.057096 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.081209 | 0.022558 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 |  73.48693 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.081209 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022558 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  17.664962 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         14.604285 |             64.197103 |                 100.0 |               109.193197 |               87.719298 |  0.0 |  0.0 |  0.029981 |       0.0 |  0.01905 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079111 | 0.021975 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 |  73.46788 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079111 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021975 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  17.664962 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         14.580638 |             64.197103 |                 100.0 |               109.220198 |               87.719298 |  0.0 |  0.0 |  0.012064 |       0.0 |  0.00758 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077068 | 0.021408 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077068 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021408 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  17.664962 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         14.561323 |             67.254108 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 | -0.000111 | -0.000111 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075077 | 0.020855 | 0.000111 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020855 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   17.58988 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         14.465269 |             78.463126 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 | -0.009107 | -0.009107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073138 | 0.020316 | 0.009218 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073138 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020316 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   17.58988 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         14.454373 |             78.463126 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 | -0.013348 | -0.013348 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.071248 | 0.019791 | 0.022566 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.071248 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019791 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   17.58988 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         14.444982 |             88.271017 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 | -0.017799 | -0.017799 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.069408 |  0.01928 | 0.040365 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.069408 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01928 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   17.58988 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          14.43577 |            117.694689 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 | -0.020794 | -0.020794 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.067615 | 0.018782 | 0.061159 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.067615 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018782 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004056 |                  0.004056 |                   17.58988 |                  3.166293 |                      0.0 |    -0.127873 |              0.0 |         14.423587 |             74.333488 |                 100.0 |               109.230942 |               93.746935 |  0.0 |  0.0 | -0.019617 | -0.019617 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.065868 | 0.018297 | 0.080776 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.065868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018297 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.057827 |                  0.057827 |                   17.58988 |                  3.179808 |                      0.0 |    -0.074665 |              0.0 |         14.410072 |            108.641252 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |   -0.0093 |   -0.0093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064167 | 0.017824 | 0.090076 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064167 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017824 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.344417 |                  0.344417 |                  18.027018 |                  3.288825 |                      0.0 |     0.207383 |              0.0 |         14.738193 |             88.271017 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |   0.05232 |   0.05232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06251 | 0.017364 | 0.037756 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017364 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.837093 |                  0.837093 |                  18.027018 |                  3.287363 |                      0.0 |     0.700119 |              0.0 |         14.739656 |             74.333488 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |  0.166735 |  0.037756 | 0.079061 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.060895 | 0.016915 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 73.381239 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060895 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016915 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.398382 |                  1.398382 |                  18.027018 |                  3.279458 |                      0.0 |     1.261738 |              0.0 |         14.747561 |             44.135508 |                 100.0 |               109.343038 |              106.487872 |  0.0 |  0.0 |  0.358814 |       0.0 | 0.185653 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.059322 | 0.016478 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 73.195586 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.059322 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016478 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.722365 |                  1.722365 |                  17.939591 |                  3.250753 |                      0.0 |     1.586917 |              0.0 |         14.688838 |             35.308407 |                 100.0 |               109.606454 |              106.665883 |  0.0 |  0.0 |  0.541756 |       0.0 | 0.263073 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05779 | 0.016053 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 72.932513 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05779 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016053 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.095274 |                  2.095274 |                  18.027018 |                  3.266966 |                      0.0 |     1.959151 |              0.0 |         14.760052 |             30.702962 |                 100.0 |               109.980185 |              106.917992 |  0.0 |  0.0 |  0.716789 |       0.0 | 0.332173 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056297 | 0.015638 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 |  72.60034 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056297 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015638 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.098911 |                  2.098911 |                  18.201874 |                  3.309837 |                      0.0 |     1.961001 |              0.0 |         14.892037 |             34.447226 |                 100.0 |               110.452867 |                107.2361 |  0.0 |  0.0 |  0.746035 |       0.0 | 0.369918 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054843 | 0.015234 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 72.230422 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054843 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015234 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.014418 |                  2.014418 |                  18.289301 |                  3.330376 |                      0.0 |     1.875652 |              0.0 |         14.958925 |             34.447226 |                 100.0 |               110.980311 |              107.590074 |  0.0 |  0.0 |  0.771401 |       0.0 | 0.389258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.053426 | 0.014841 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 71.841164 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.053426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014841 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.101265 |                  2.101265 |                  18.464157 |                  3.363175 |                      0.0 |     1.961133 |              0.0 |         15.100981 |             32.098552 |                 100.0 |               111.536545 |              107.962247 |  0.0 |  0.0 |  0.840981 |       0.0 | 0.414609 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052046 | 0.014457 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 71.426555 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052046 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014457 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.013503 |                  2.013503 |                  18.551584 |                   3.38194 |                      0.0 |     1.872589 |              0.0 |         15.169644 |             30.702962 |                 100.0 |               112.130402 |              108.358325 |  0.0 |  0.0 |  0.865172 |       0.0 | 0.419297 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.050702 | 0.014084 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 71.007258 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.050702 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014084 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.697346 |                  1.697346 |                  18.639012 |                  3.400029 |                      0.0 |     1.555678 |              0.0 |         15.238983 |              31.38525 |                 100.0 |               112.732468 |              108.758548 |  0.0 |  0.0 |  0.798412 |       0.0 | 0.393038 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049392 |  0.01372 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 |  70.61422 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049392 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01372 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.231461 |                  1.231461 |                   18.72644 |                   3.42173 |                      0.0 |     1.088889 |              0.0 |         15.304709 |             30.049708 |                 100.0 |               113.298222 |              109.133419 |  0.0 |  0.0 |  0.702711 |       0.0 | 0.339609 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048116 | 0.013366 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 70.274611 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013366 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |              0.748608 |                  0.748608 |                   18.72644 |                  3.423912 |                      0.0 |     0.605945 |              0.0 |         15.302528 |              31.38525 |                 100.0 |               113.788174 |              109.457117 |  0.0 |  0.0 |   0.55931 |       0.0 | 0.275426 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046873 |  0.01302 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 69.999185 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046873 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01302 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |              0.364895 |                  0.364895 |                   18.72644 |                   3.42458 |                      0.0 |     0.222204 |              0.0 |          15.30186 |             34.447226 |                 100.0 |               114.186294 |                109.7195 |  0.0 |  0.0 |  0.406398 |       0.0 | 0.207367 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.045662 | 0.012684 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 69.791818 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.045662 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012684 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.110333 |                  0.110333 |                   18.72644 |                  3.420162 |                      0.0 |    -0.032173 |              0.0 |         15.306277 |             37.166744 |                 100.0 |               114.486496 |              109.916969 |  0.0 |  0.0 |  0.257469 |       0.0 | 0.133625 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044483 | 0.012356 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 69.658193 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044483 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012356 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002372 |                  0.002372 |                  18.901295 |                  3.445356 |                      0.0 |    -0.141184 |              0.0 |         15.455939 |             56.493451 |                 100.0 |               114.680153 |                89.03222 |  0.0 |  0.0 |  0.102212 |       0.0 | 0.066116 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043334 | 0.012037 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 69.592077 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043334 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012037 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                  18.901295 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         15.466597 |             74.333488 |                 100.0 |               114.776034 |               87.719298 |  0.0 |  0.0 |  0.046259 |       0.0 | 0.031824 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042215 | 0.011726 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 69.560253 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011726 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                  18.901295 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         15.473739 |             64.197103 |                 100.0 |               114.822199 |               87.719298 |  0.0 |  0.0 |  0.042514 |       0.0 | 0.027773 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041124 | 0.011423 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 69.532479 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041124 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011423 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                  18.901295 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         15.482962 |             52.308751 |                 100.0 |               114.862497 |               87.719298 |  0.0 |  0.0 |  0.035583 |       0.0 | 0.021518 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040062 | 0.011128 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 | 69.510962 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011128 |

.. _lland_v3_acker_heavy_rain_hourly:

acre (heavy rain)
-----------------

When comparing the soil moisture content (|BoWa|) at August 6th, 00:00, we
notice only small differences between the daily
(:ref:`lland_v3_acker_summer_daily`, 70.2 mm) and the hourly
(:ref:`lland_v3_acker_summer_daily`, 69.5) simulation results for a dry
situation.  However, when we compare the differences between our daily
(:ref:`lland_v3_acker_heavy_rain_daily`, 152.8 mm) and our hourly (next
table) test for a wet situation, we see much more pronounced differences,
which are mainly due to the prioritisation of throughfall (|NBes|) over
interception evaporation (|EvI|):

.. integration-test::

    >>> inputs.nied.series = 20.0/24.0
    >>> test("lland_v3_acker_heavy_rain_hourly",
    ...      axis1=(inputs.nied, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_heavy_rain)
    |                date |     nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |     nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |     wada |      qdb |     qib1 | qib2 |      qbb | qkap |     qdgz |      qah |       qa |     inzp | wats | waes | esnow | taus |      ebdn |       bowa |    qdgz1 | qdgz2 |    qigz1 | qigz2 |     qbgz |    qdga1 | qdga2 |    qiga1 | qiga2 |     qbga | inlet |   outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.833333 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 | 0.783333 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   7.982856 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |           7.08166 |            176.542034 |                 100.0 |                70.327061 |               87.719298 | 0.083333 |  0.0 | -0.002965 | -0.002965 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.007403 |      0.0 |  0.0 |      0.0 |  0.0 | 0.007403 |  0.80139 | 0.222608 | 0.702965 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 |   102.5625 | 0.007403 |   0.0 |      0.0 |   0.0 |      0.0 |  0.80139 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.222608 |
    | 1997-08-03 01:00:00 | 0.833333 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 | 0.783333 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   7.982856 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          7.081012 |            176.542034 |                 100.0 |                70.229321 |               87.719298 | 0.786298 |  0.0 | -0.002685 | -0.002685 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.786298 | 0.070172 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070172 | 0.781695 | 0.217137 | 0.702685 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 103.278626 | 0.070172 |   0.0 |      0.0 |   0.0 |      0.0 | 0.781695 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.217137 |
    | 1997-08-03 02:00:00 | 0.833333 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 | 0.783333 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   7.982856 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |            7.0805 |            176.542034 |                 100.0 |                  69.3082 |               87.719298 | 0.786019 |  0.0 | -0.004107 | -0.004107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.786019 | 0.070722 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070722 | 0.763323 | 0.212034 | 0.704107 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 103.993923 | 0.070722 |   0.0 |      0.0 |   0.0 |      0.0 | 0.763323 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.212034 |
    | 1997-08-03 03:00:00 | 0.833333 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 | 0.783333 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   7.982856 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          7.080492 |            176.542034 |                 100.0 |                68.389393 |               87.719298 | 0.787441 |  0.0 | -0.005232 | -0.005232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.787441 | 0.071427 |      0.0 |  0.0 |      0.0 |  0.0 | 0.071427 | 0.745441 | 0.207067 | 0.705232 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 104.709937 | 0.071427 |   0.0 |      0.0 |   0.0 |      0.0 | 0.745441 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.207067 |
    | 1997-08-03 04:00:00 | 0.833333 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 | 0.783333 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.005318 |                  0.005318 |                   7.982856 |                  0.902341 |                      0.0 |     -0.03228 |              0.0 |          7.080514 |            176.542034 |                 100.0 |                67.470886 |                81.46212 | 0.788565 |  0.0 | -0.004819 | -0.004819 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.788565 | 0.072109 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072109 |  0.72804 | 0.202233 | 0.704819 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 105.426394 | 0.072109 |   0.0 |      0.0 |   0.0 |      0.0 |  0.72804 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.202233 |
    | 1997-08-03 05:00:00 | 0.833333 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 | 0.783333 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.060007 |                  0.060007 |                   7.982856 |                  0.901482 |                      0.0 |     0.022445 |              0.0 |          7.081373 |            235.389378 |                 100.0 |                66.553006 |               73.619304 | 0.788152 |  0.0 |  0.005073 |  0.005073 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.788152 | 0.072652 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072652 | 0.711104 | 0.197529 | 0.694927 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 106.141894 | 0.072652 |   0.0 |      0.0 |   0.0 |      0.0 | 0.711104 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.197529 |
    | 1997-08-03 06:00:00 | 0.833333 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 | 0.783333 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.156653 |                  0.156653 |                   7.982856 |                   0.89977 |                      0.0 |     0.119163 |              0.0 |          7.083086 |            156.926252 |                 100.0 |                65.637518 |               72.818235 |  0.77826 |  0.0 |  0.026292 |  0.026292 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.77826 | 0.072311 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072311 | 0.694607 | 0.192946 | 0.673708 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 106.847843 | 0.072311 |   0.0 |      0.0 |   0.0 |      0.0 | 0.694607 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.192946 |
    | 1997-08-03 07:00:00 | 0.833333 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 | 0.783333 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.299135 |                  0.299135 |                   7.982856 |                  0.896826 |                      0.0 |     0.261767 |              0.0 |          7.086029 |            156.926252 |                 100.0 |                64.735367 |               72.024255 | 0.757042 |  0.0 |  0.057456 |  0.057456 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.757042 | 0.070885 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070885 | 0.678514 | 0.188476 | 0.642544 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 |    107.534 | 0.070885 |   0.0 |      0.0 |   0.0 |      0.0 | 0.678514 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.188476 |
    | 1997-08-03 08:00:00 | 0.833333 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 | 0.783333 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.467717 |                  0.467717 |                   7.982856 |                  0.891371 |                      0.0 |     0.430576 |              0.0 |          7.091484 |            156.926252 |                 100.0 |                63.859549 |               71.249063 | 0.725878 |  0.0 |  0.096099 |  0.096099 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.725878 | 0.068473 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068473 | 0.662788 | 0.184108 | 0.603901 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 108.191405 | 0.068473 |   0.0 |      0.0 |   0.0 |      0.0 | 0.662788 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.184108 |
    | 1997-08-03 09:00:00 | 0.833333 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 | 0.783333 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.852983 |                  0.852983 |                    8.15863 |                  0.922604 |                      0.0 |     0.814541 |              0.0 |          7.236026 |            108.641252 |                 100.0 |                63.021373 |               70.503105 | 0.687235 |  0.0 |  0.222763 |  0.222763 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.687235 | 0.065285 |      0.0 |  0.0 |      0.0 |  0.0 | 0.065285 | 0.647395 | 0.179832 | 0.477237 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 108.813354 | 0.065285 |   0.0 |      0.0 |   0.0 |      0.0 | 0.647395 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.179832 |
    | 1997-08-03 10:00:00 | 0.833333 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 | 0.783333 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.372332 |                  1.372332 |                   8.334404 |                  0.955299 |                      0.0 |     1.332528 |              0.0 |          7.379106 |             94.155751 |                 100.0 |                62.229232 |               69.794418 |  0.56057 |  0.0 |  0.389923 |  0.389923 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.56057 | 0.053582 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053582 | 0.632206 | 0.175613 | 0.310077 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 109.320342 | 0.053582 |   0.0 |      0.0 |   0.0 |      0.0 | 0.632206 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.175613 |
    | 1997-08-03 11:00:00 | 0.833333 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 | 0.783333 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.682691 |                  1.682691 |                   8.773841 |                  1.039981 |                      0.0 |     1.639359 |              0.0 |          7.733859 |            117.694689 |                 100.0 |                61.584098 |               69.214576 |  0.39341 |  0.0 |   0.46628 |   0.46628 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.39341 | 0.037782 |      0.0 |  0.0 |      0.0 |  0.0 | 0.037782 | 0.617055 | 0.171404 |  0.23372 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 109.675971 | 0.037782 |   0.0 |      0.0 |   0.0 |      0.0 | 0.617055 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.171404 |
    | 1997-08-03 12:00:00 | 0.833333 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 | 0.783333 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.714225 |                  1.714225 |                   9.301164 |                  1.142661 |                      0.0 |     1.666614 |              0.0 |          8.158503 |            108.641252 |                 100.0 |                61.131877 |               68.806682 | 0.317053 |  0.0 |  0.491066 |  0.491066 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.317053 | 0.030555 |      0.0 |  0.0 |      0.0 |  0.0 | 0.030555 | 0.601998 | 0.167222 | 0.208934 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 109.962469 | 0.030555 |   0.0 |      0.0 |   0.0 |      0.0 | 0.601998 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.167222 |
    | 1997-08-03 13:00:00 | 0.833333 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 | 0.783333 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.357421 |                  1.357421 |                   9.652713 |                  1.211985 |                      0.0 |     1.306922 |              0.0 |          8.440727 |             94.155751 |                 100.0 |                60.767747 |               68.477376 | 0.292267 |  0.0 |  0.424802 |  0.424802 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.292267 | 0.028251 |      0.0 |  0.0 |      0.0 |  0.0 | 0.028251 | 0.587207 | 0.163113 | 0.275198 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 110.226486 | 0.028251 |   0.0 |      0.0 |   0.0 |      0.0 | 0.587207 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.163113 |
    | 1997-08-03 14:00:00 | 0.833333 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 | 0.783333 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.147809 |                  1.147809 |                   9.828487 |                   1.24753 |                      0.0 |     1.095829 |              0.0 |          8.580957 |             74.333488 |                 100.0 |                60.432333 |               68.173354 | 0.358532 |  0.0 |  0.415458 |  0.415458 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.358532 | 0.034766 |      0.0 |  0.0 |      0.0 |  0.0 | 0.034766 | 0.572854 | 0.159126 | 0.284542 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 110.550251 | 0.034766 |   0.0 |      0.0 |   0.0 |      0.0 | 0.572854 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.159126 |
    | 1997-08-03 15:00:00 | 0.833333 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 | 0.783333 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |               1.06126 |                   1.06126 |                  10.180036 |                  1.316893 |                      0.0 |     1.006389 |              0.0 |          8.863143 |             74.333488 |                 100.0 |                  60.0212 |               67.799797 | 0.367875 |  0.0 |  0.393508 |  0.393508 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.367875 | 0.035799 |      0.0 |  0.0 |      0.0 |  0.0 | 0.035799 | 0.558968 | 0.155269 | 0.306492 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 110.882327 | 0.035799 |   0.0 |      0.0 |   0.0 |      0.0 | 0.558968 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.155269 |
    | 1997-08-03 16:00:00 | 0.833333 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 | 0.783333 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.761967 |                  0.761967 |                  10.443698 |                  1.370145 |                      0.0 |     0.704878 |              0.0 |          9.073553 |             61.405925 |                 100.0 |                59.599725 |               67.415809 | 0.389826 |  0.0 |  0.342513 |  0.342513 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.389826 | 0.038075 |      0.0 |  0.0 |      0.0 |  0.0 | 0.038075 | 0.545483 | 0.151523 | 0.357487 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 111.234078 | 0.038075 |   0.0 |      0.0 |   0.0 |      0.0 | 0.545483 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.151523 |
    | 1997-08-03 17:00:00 | 0.833333 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 | 0.783333 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.375211 |                  0.375211 |                   10.70736 |                  1.421011 |                      0.0 |     0.316002 |              0.0 |          9.286348 |             58.847345 |                 100.0 |                59.153511 |               67.008135 | 0.440821 |  0.0 |  0.236583 |  0.236583 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.440821 |  0.04323 |      0.0 |  0.0 |      0.0 |  0.0 |  0.04323 | 0.532443 | 0.147901 | 0.463417 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 111.631669 |  0.04323 |   0.0 |      0.0 |   0.0 |      0.0 | 0.532443 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.147901 |
    | 1997-08-03 18:00:00 | 0.833333 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 | 0.783333 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.139757 |                  0.139757 |                  10.795247 |                  1.433696 |                      0.0 |     0.080019 |              0.0 |           9.36155 |             56.493451 |                 100.0 |                58.649429 |               66.546166 |  0.54675 |  0.0 |  0.156446 |  0.156446 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.54675 | 0.053875 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053875 | 0.519945 | 0.144429 | 0.543554 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 112.124544 | 0.053875 |   0.0 |      0.0 |   0.0 |      0.0 | 0.519945 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.144429 |
    | 1997-08-03 19:00:00 | 0.833333 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 | 0.783333 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.037299 |                  0.037299 |                  10.795247 |                  1.423756 |                      0.0 |    -0.022024 |              0.0 |          9.371491 |             56.493451 |                 100.0 |                58.024955 |               65.971756 | 0.626887 |  0.0 |   0.09619 |   0.09619 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.626887 | 0.062121 |      0.0 |  0.0 |      0.0 |  0.0 | 0.062121 | 0.508013 | 0.141115 |  0.60381 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 |  112.68931 | 0.062121 |   0.0 |      0.0 |   0.0 |      0.0 | 0.508013 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.141115 |
    | 1997-08-03 20:00:00 | 0.833333 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 | 0.783333 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000509 |                  0.000509 |                  10.795247 |                  1.412078 |                      0.0 |    -0.058328 |              0.0 |          9.383169 |             64.197103 |                 100.0 |                 57.30995 |               83.797685 | 0.687143 |  0.0 |   0.04938 |   0.04938 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.687143 | 0.068523 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068523 | 0.496578 | 0.137938 |  0.65062 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 113.307931 | 0.068523 |   0.0 |      0.0 |   0.0 |      0.0 | 0.496578 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.137938 |
    | 1997-08-03 21:00:00 | 0.833333 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 | 0.783333 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  10.795247 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          9.393683 |             83.078604 |                 100.0 |                56.527431 |               87.719298 | 0.733954 |  0.0 |  0.019258 |  0.019258 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.733954 | 0.073689 |      0.0 |  0.0 |      0.0 |  0.0 | 0.073689 | 0.485588 | 0.134886 | 0.680742 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 113.968195 | 0.073689 |   0.0 |      0.0 |   0.0 |      0.0 | 0.485588 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.134886 |
    | 1997-08-03 22:00:00 | 0.833333 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 | 0.783333 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  10.795247 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          9.404089 |             83.078604 |                 100.0 |                 55.69299 |               87.719298 | 0.764075 |  0.0 |  0.010298 |  0.010298 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.764075 | 0.077261 |      0.0 |  0.0 |      0.0 |  0.0 | 0.077261 | 0.474995 | 0.131943 | 0.689702 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 |  114.65501 | 0.077261 |   0.0 |      0.0 |   0.0 |      0.0 | 0.474995 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.131943 |
    | 1997-08-03 23:00:00 | 0.833333 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 | 0.783333 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  10.795247 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          9.412407 |             61.405925 |                 100.0 |                 54.82581 |               87.719298 | 0.773036 |  0.0 |  0.014661 |  0.014661 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.773036 | 0.078736 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078736 |  0.46474 | 0.129094 | 0.685339 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 115.349309 | 0.078736 |   0.0 |      0.0 |   0.0 |      0.0 |  0.46474 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.129094 |
    | 1997-08-04 00:00:00 | 0.833333 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 | 0.783333 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  10.739718 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          9.343179 |             61.405925 |                 100.0 |                53.950005 |               87.719298 | 0.768673 |  0.0 |  0.023989 |  0.023989 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.768673 | 0.078861 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078861 | 0.454771 | 0.126325 | 0.676011 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 116.039121 | 0.078861 |   0.0 |      0.0 |   0.0 |      0.0 | 0.454771 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.126325 |
    | 1997-08-04 01:00:00 | 0.833333 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 | 0.783333 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  10.739718 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          9.333085 |             64.197103 |                 100.0 |                53.080668 |               87.719298 | 0.759344 |  0.0 |  0.018182 |  0.018182 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.759344 | 0.078462 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078462 | 0.445055 | 0.123627 | 0.681818 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 116.720003 | 0.078462 |   0.0 |      0.0 |   0.0 |      0.0 | 0.445055 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.123627 |
    | 1997-08-04 02:00:00 | 0.833333 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 | 0.783333 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  10.739718 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |           9.32165 |             61.405925 |                 100.0 |                52.223358 |               87.719298 | 0.765151 |  0.0 |   0.02055 |   0.02055 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.765151 | 0.079625 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079625 | 0.435601 |    0.121 |  0.67945 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 117.405529 | 0.079625 |   0.0 |      0.0 |   0.0 |      0.0 | 0.435601 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 |    0.121 |
    | 1997-08-04 03:00:00 | 0.833333 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 | 0.783333 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  10.739718 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          9.309548 |             70.616813 |                 100.0 |                 51.36096 |               87.719298 | 0.762784 |  0.0 |   0.01363 |   0.01363 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.762784 | 0.079941 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079941 |  0.42641 | 0.118447 |  0.68637 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 118.088372 | 0.079941 |   0.0 |      0.0 |   0.0 |      0.0 |  0.42641 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.118447 |
    | 1997-08-04 04:00:00 | 0.833333 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 | 0.783333 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.004352 |                  0.004352 |                  10.739718 |                  1.440702 |                      0.0 |    -0.055678 |              0.0 |          9.299016 |             61.405925 |                 100.0 |                50.502684 |               73.397763 | 0.769703 |  0.0 |  0.007255 |  0.007255 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.769703 | 0.081237 |      0.0 |  0.0 |      0.0 |  0.0 | 0.081237 | 0.417477 | 0.115966 | 0.692745 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 118.776838 | 0.081237 |   0.0 |      0.0 |   0.0 |      0.0 | 0.417477 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.115966 |
    | 1997-08-04 05:00:00 | 0.833333 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 | 0.783333 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.054586 |                  0.054586 |                  10.739718 |                  1.451305 |                      0.0 |    -0.005885 |              0.0 |          9.288414 |             56.493451 |                 100.0 |                49.638078 |               58.024051 | 0.776079 |  0.0 |  0.018162 |  0.018162 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.776079 | 0.082492 |      0.0 |  0.0 |      0.0 |  0.0 | 0.082492 | 0.408808 | 0.113558 | 0.681838 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 119.470425 | 0.082492 |   0.0 |      0.0 |   0.0 |      0.0 | 0.408808 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.113558 |
    | 1997-08-04 06:00:00 | 0.833333 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 | 0.783333 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.172329 |                  0.172329 |                   10.82738 |                  1.478175 |                      0.0 |     0.110738 |              0.0 |          9.349205 |             58.847345 |                 100.0 |                 48.85611 |               57.260179 | 0.765171 |  0.0 |   0.04093 |   0.04093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.765171 | 0.081905 | 0.003897 |  0.0 | 0.000003 |  0.0 | 0.081905 | 0.400397 | 0.111221 |  0.65907 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 |  120.14979 | 0.081905 |   0.0 | 0.003897 |   0.0 | 0.000003 | 0.400372 |   0.0 | 0.000025 |   0.0 |      0.0 |   0.0 | 0.111221 |
    | 1997-08-04 07:00:00 | 0.833333 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 | 0.783333 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |              0.781884 |                  0.781884 |                  11.616331 |                   1.64464 |                      0.0 |     0.713358 |              0.0 |          9.971692 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.742403 |  0.0 |   0.18972 |   0.18972 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.742403 |  0.08001 | 0.003919 |  0.0 | 0.000031 |  0.0 |  0.08001 | 0.392197 | 0.108943 |  0.51028 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 120.808233 |  0.08001 |   0.0 | 0.003919 |   0.0 | 0.000031 | 0.392121 |   0.0 | 0.000076 |   0.0 |      0.0 |   0.0 | 0.108943 |
    | 1997-08-04 08:00:00 | 0.833333 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 | 0.783333 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.300216 |                  1.300216 |                  12.492945 |                   1.83472 |                      0.0 |      1.22377 |              0.0 |         10.658225 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.593614 |  0.0 |  0.343734 |  0.343734 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.593614 | 0.064359 | 0.003941 |  0.0 | 0.000059 |  0.0 | 0.064359 | 0.383982 | 0.106662 | 0.356266 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 121.333488 | 0.064359 |   0.0 | 0.003941 |   0.0 | 0.000059 | 0.383856 |   0.0 | 0.000126 |   0.0 |      0.0 |   0.0 | 0.106662 |
    | 1997-08-04 09:00:00 | 0.833333 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 | 0.783333 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.679922 |                  1.679922 |                  13.194235 |                  1.989515 |                      0.0 |     1.597026 |              0.0 |          11.20472 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.439599 |  0.0 |  0.494183 |  0.494183 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.439599 | 0.047881 | 0.003958 |  0.0 | 0.000081 |  0.0 | 0.047881 | 0.375565 | 0.104324 | 0.205817 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 121.721168 | 0.047881 |   0.0 | 0.003958 |   0.0 | 0.000081 | 0.375389 |   0.0 | 0.000176 |   0.0 |      0.0 |   0.0 | 0.104324 |
    | 1997-08-04 10:00:00 | 0.833333 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 | 0.783333 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              1.761455 |                  1.761455 |                   13.54488 |                   2.06621 |                      0.0 |     1.675363 |              0.0 |          11.47867 |             45.559235 |                 100.0 |                 48.85611 |               57.260179 | 0.289151 |  0.0 |  0.550449 |  0.550449 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.289151 | 0.031596 | 0.003971 |  0.0 | 0.000097 |  0.0 | 0.031596 | 0.366943 | 0.101929 | 0.149551 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 121.974655 | 0.031596 |   0.0 | 0.003971 |   0.0 | 0.000097 | 0.366718 |   0.0 | 0.000225 |   0.0 |      0.0 |   0.0 | 0.101929 |
    | 1997-08-04 11:00:00 | 0.833333 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 | 0.783333 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |               1.64232 |                   1.64232 |                  13.632542 |                  2.087634 |                      0.0 |     1.555335 |              0.0 |         11.544908 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 | 0.232884 |  0.0 |  0.548384 |  0.548384 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.232884 | 0.025506 | 0.003979 |  0.0 | 0.000107 |  0.0 | 0.025506 | 0.358256 | 0.099516 | 0.151616 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 122.177947 | 0.025506 |   0.0 | 0.003979 |   0.0 | 0.000107 | 0.357982 |   0.0 | 0.000274 |   0.0 |      0.0 |   0.0 | 0.099516 |
    | 1997-08-04 12:00:00 | 0.833333 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 | 0.783333 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.697727 |                  1.697727 |                  13.720203 |                  2.106001 |                      0.0 |     1.609977 |              0.0 |         11.614202 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 |  0.23495 |  0.0 |  0.597559 |  0.597559 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.23495 | 0.025785 | 0.003986 |  0.0 | 0.000116 |  0.0 | 0.025785 |  0.34972 | 0.097144 | 0.102441 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 122.383011 | 0.025785 |   0.0 | 0.003986 |   0.0 | 0.000116 | 0.349398 |   0.0 | 0.000322 |   0.0 |      0.0 |   0.0 | 0.097144 |
    | 1997-08-04 13:00:00 | 0.833333 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 | 0.783333 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |              1.640072 |                  1.640072 |                  13.983187 |                  2.166369 |                      0.0 |     1.549806 |              0.0 |         11.816818 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.185775 |  0.0 |  0.621801 |  0.621801 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.185775 | 0.020426 | 0.003992 |  0.0 | 0.000124 |  0.0 | 0.020426 | 0.341339 | 0.094816 | 0.078199 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 122.544243 | 0.020426 |   0.0 | 0.003992 |   0.0 | 0.000124 | 0.340969 |   0.0 |  0.00037 |   0.0 |      0.0 |   0.0 | 0.094816 |
    | 1997-08-04 14:00:00 | 0.833333 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 | 0.783333 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |              1.678676 |                  1.678676 |                  14.421494 |                  2.265756 |                      0.0 |      1.58427 |              0.0 |         12.155738 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.161533 |  0.0 |  0.589183 |  0.589183 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.161533 | 0.017787 | 0.003998 |  0.0 | 0.000131 |  0.0 | 0.017787 | 0.333072 |  0.09252 | 0.110817 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 |  122.68386 | 0.017787 |   0.0 | 0.003998 |   0.0 | 0.000131 | 0.332655 |   0.0 | 0.000417 |   0.0 |      0.0 |   0.0 |  0.09252 |
    | 1997-08-04 15:00:00 | 0.833333 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 | 0.783333 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.414805 |                  1.414805 |                  14.772139 |                  2.348437 |                      0.0 |     1.316953 |              0.0 |         12.423701 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 |  0.19415 |  0.0 |  0.551424 |  0.551424 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.19415 | 0.021412 | 0.004002 |  0.0 | 0.000137 |  0.0 | 0.021412 | 0.325032 | 0.090287 | 0.148576 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 |  122.85246 | 0.021412 |   0.0 | 0.004002 |   0.0 | 0.000137 | 0.324569 |   0.0 | 0.000463 |   0.0 |      0.0 |   0.0 | 0.090287 |
    | 1997-08-04 16:00:00 | 0.833333 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 | 0.783333 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.033761 |                  1.033761 |                  15.122784 |                  2.428674 |                      0.0 |     0.932566 |              0.0 |          12.69411 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.231909 |  0.0 |  0.505761 |  0.505761 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.231909 | 0.025623 | 0.004008 |  0.0 | 0.000144 |  0.0 | 0.025623 | 0.317302 | 0.088139 | 0.194239 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 123.054595 | 0.025623 |   0.0 | 0.004008 |   0.0 | 0.000144 | 0.316793 |   0.0 | 0.000509 |   0.0 |      0.0 |   0.0 | 0.088139 |
    | 1997-08-04 17:00:00 | 0.833333 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 | 0.783333 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.690032 |                  0.690032 |                  15.736413 |                  2.564427 |                      0.0 |     0.583181 |              0.0 |         13.171986 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.277572 |  0.0 |  0.444603 |  0.444603 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.277572 | 0.030737 | 0.004014 |  0.0 | 0.000152 |  0.0 | 0.030737 | 0.309893 | 0.086081 | 0.255397 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 123.297264 | 0.030737 |   0.0 | 0.004014 |   0.0 | 0.000152 | 0.309338 |   0.0 | 0.000555 |   0.0 |      0.0 |   0.0 | 0.086081 |
    | 1997-08-04 18:00:00 | 0.833333 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 | 0.783333 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.351786 |                  0.351786 |                  16.525365 |                  2.744116 |                      0.0 |     0.237448 |              0.0 |         13.781249 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 |  0.33873 |  0.0 |   0.32478 |   0.32478 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.33873 |  0.03761 | 0.004022 |  0.0 | 0.000162 |  0.0 |  0.03761 |  0.30283 |  0.08412 |  0.37522 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 |   123.5942 |  0.03761 |   0.0 | 0.004022 |   0.0 | 0.000162 |  0.30223 |   0.0 |   0.0006 |   0.0 |      0.0 |   0.0 |  0.08412 |
    | 1997-08-04 19:00:00 | 0.833333 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 | 0.783333 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.115111 |                  0.115111 |                  17.401978 |                  2.950104 |                      0.0 |     -0.00781 |              0.0 |         14.451875 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.458554 |  0.0 |  0.183091 |  0.183091 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.458554 | 0.051092 | 0.004032 |  0.0 | 0.000175 |  0.0 | 0.051092 | 0.296214 | 0.082282 | 0.516909 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 123.997455 | 0.051092 |   0.0 | 0.004032 |   0.0 | 0.000175 |  0.29557 |   0.0 | 0.000644 |   0.0 |      0.0 |   0.0 | 0.082282 |
    | 1997-08-04 20:00:00 | 0.833333 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 | 0.783333 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002224 |                  0.002224 |                  17.664962 |                  3.034376 |                      0.0 |    -0.124208 |              0.0 |         14.630586 |             44.135508 |                 100.0 |                 48.85611 |               83.091125 | 0.600242 |  0.0 |  0.100631 |  0.100631 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.600242 | 0.067188 | 0.004045 |  0.0 | 0.000192 |  0.0 | 0.067188 | 0.290152 | 0.080598 | 0.599369 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 124.526272 | 0.067188 |   0.0 | 0.004045 |   0.0 | 0.000192 | 0.289464 |   0.0 | 0.000689 |   0.0 |      0.0 |   0.0 | 0.080598 |
    | 1997-08-04 21:00:00 | 0.833333 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 | 0.783333 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  17.664962 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         14.604285 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.682702 |  0.0 |  0.029981 |  0.029981 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.682702 | 0.076847 | 0.004062 |  0.0 | 0.000214 |  0.0 | 0.076847 |  0.28458 |  0.07905 | 0.670019 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 125.127851 | 0.076847 |   0.0 | 0.004062 |   0.0 | 0.000214 | 0.283847 |   0.0 | 0.000732 |   0.0 |      0.0 |   0.0 |  0.07905 |
    | 1997-08-04 22:00:00 | 0.833333 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 | 0.783333 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  17.664962 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         14.580638 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.753353 |  0.0 |  0.012064 |  0.012064 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.753353 | 0.085331 | 0.004082 |  0.0 | 0.000239 |  0.0 | 0.085331 | 0.279386 | 0.077607 | 0.687936 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 | 125.791553 | 0.085331 |   0.0 | 0.004082 |   0.0 | 0.000239 |  0.27861 |   0.0 | 0.000776 |   0.0 |      0.0 |   0.0 | 0.077607 |
    | 1997-08-04 23:00:00 | 0.833333 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 | 0.783333 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  17.664962 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         14.561323 |             67.254108 |                 100.0 |                 48.85611 |               87.719298 | 0.771269 |  0.0 | -0.000111 | -0.000111 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.771269 | 0.087938 | 0.004103 |  0.0 | 0.000266 |  0.0 | 0.087938 | 0.274471 | 0.076242 | 0.700111 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 | 126.470515 | 0.087938 |   0.0 | 0.004103 |   0.0 | 0.000266 | 0.273652 |   0.0 | 0.000819 |   0.0 |      0.0 |   0.0 | 0.076242 |
    | 1997-08-05 00:00:00 | 0.833333 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 | 0.783333 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   17.58988 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         14.465269 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.783444 |  0.0 | -0.009107 | -0.009107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.783444 | 0.089926 | 0.004126 |  0.0 | 0.000295 |  0.0 | 0.089926 | 0.269742 | 0.074928 | 0.709107 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 | 127.159612 | 0.089926 |   0.0 | 0.004126 |   0.0 | 0.000295 |  0.26888 |   0.0 | 0.000862 |   0.0 |      0.0 |   0.0 | 0.074928 |
    | 1997-08-05 01:00:00 | 0.833333 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 | 0.783333 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   17.58988 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         14.454373 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.792441 |  0.0 | -0.013348 | -0.013348 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.792441 | 0.091575 | 0.004148 |  0.0 | 0.000323 |  0.0 | 0.091575 | 0.265184 | 0.073662 | 0.713348 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 | 127.856006 | 0.091575 |   0.0 | 0.004148 |   0.0 | 0.000323 | 0.264279 |   0.0 | 0.000904 |   0.0 |      0.0 |   0.0 | 0.073662 |
    | 1997-08-05 02:00:00 | 0.833333 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 | 0.783333 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   17.58988 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         14.444982 |             88.271017 |                 100.0 |                 48.85611 |               87.719298 | 0.796681 |  0.0 | -0.017799 | -0.017799 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.796681 | 0.092692 | 0.004171 |  0.0 | 0.000352 |  0.0 | 0.092692 |  0.26078 | 0.072439 | 0.717799 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 | 128.555473 | 0.092692 |   0.0 | 0.004171 |   0.0 | 0.000352 | 0.259832 |   0.0 | 0.000947 |   0.0 |      0.0 |   0.0 | 0.072439 |
    | 1997-08-05 03:00:00 | 0.833333 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 | 0.783333 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   17.58988 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          14.43577 |            117.694689 |                 100.0 |                 48.85611 |               87.719298 | 0.801132 |  0.0 | -0.020794 | -0.020794 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.801132 | 0.093844 | 0.004194 |  0.0 | 0.000381 |  0.0 | 0.093844 | 0.256519 | 0.071255 | 0.720794 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 | 129.258186 | 0.093844 |   0.0 | 0.004194 |   0.0 | 0.000381 |  0.25553 |   0.0 | 0.000989 |   0.0 | 0.000001 |   0.0 | 0.071255 |
    | 1997-08-05 04:00:00 | 0.833333 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 | 0.783333 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004056 |                  0.004056 |                   17.58988 |                  3.166293 |                      0.0 |    -0.127873 |              0.0 |         14.423587 |             74.333488 |                 100.0 |                 48.85611 |               73.423526 | 0.804127 |  0.0 | -0.019617 | -0.019617 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.804127 | 0.094836 | 0.004217 |  0.0 | 0.000411 |  0.0 | 0.094836 | 0.252397 |  0.07011 | 0.719617 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |  129.96285 | 0.094836 |   0.0 | 0.004217 |   0.0 | 0.000411 | 0.251366 |   0.0 |  0.00103 |   0.0 | 0.000001 |   0.0 |  0.07011 |
    | 1997-08-05 05:00:00 | 0.833333 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 | 0.783333 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.057827 |                  0.057827 |                   17.58988 |                  3.179808 |                      0.0 |    -0.074665 |              0.0 |         14.410072 |            108.641252 |                 100.0 |                 48.85611 |               57.260179 |  0.80295 |  0.0 |   -0.0093 |   -0.0093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.80295 | 0.095339 |  0.00424 |  0.0 |  0.00044 |  0.0 | 0.095339 | 0.248402 | 0.069001 |   0.7093 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 | 130.665781 | 0.095339 |   0.0 |  0.00424 |   0.0 |  0.00044 |  0.24733 |   0.0 | 0.001072 |   0.0 | 0.000001 |   0.0 | 0.069001 |
    | 1997-08-05 06:00:00 | 0.833333 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 | 0.783333 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.344417 |                  0.344417 |                  18.027018 |                  3.288825 |                      0.0 |     0.207383 |              0.0 |         14.738193 |             88.271017 |                 100.0 |                 48.85611 |               57.260179 | 0.792634 |  0.0 |   0.05232 |   0.05232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.792634 | 0.094745 | 0.004262 |  0.0 | 0.000469 |  0.0 | 0.094745 |  0.24451 | 0.067919 |  0.64768 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 131.358937 | 0.094745 |   0.0 | 0.004262 |   0.0 | 0.000469 | 0.243396 |   0.0 | 0.001113 |   0.0 | 0.000001 |   0.0 | 0.067919 |
    | 1997-08-05 07:00:00 | 0.833333 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 | 0.783333 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.837093 |                  0.837093 |                  18.027018 |                  3.287363 |                      0.0 |     0.700119 |              0.0 |         14.739656 |             74.333488 |                 100.0 |                 48.85611 |               57.260179 | 0.731013 |  0.0 |  0.166735 |  0.166735 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.731013 | 0.087936 | 0.004285 |  0.0 | 0.000498 |  0.0 | 0.087936 | 0.240623 |  0.06684 | 0.533265 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 131.997231 | 0.087936 |   0.0 | 0.004285 |   0.0 | 0.000498 | 0.239468 |   0.0 | 0.001154 |   0.0 | 0.000001 |   0.0 |  0.06684 |
    | 1997-08-05 08:00:00 | 0.833333 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 | 0.783333 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.398382 |                  1.398382 |                  18.027018 |                  3.279458 |                      0.0 |     1.261738 |              0.0 |         14.747561 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.616599 |  0.0 |  0.358814 |  0.358814 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.616599 | 0.074588 | 0.004306 |  0.0 | 0.000525 |  0.0 | 0.074588 | 0.236577 | 0.065716 | 0.341186 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 132.534411 | 0.074588 |   0.0 | 0.004306 |   0.0 | 0.000525 |  0.23538 |   0.0 | 0.001195 |   0.0 | 0.000001 |   0.0 | 0.065716 |
    | 1997-08-05 09:00:00 | 0.833333 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 | 0.783333 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.722365 |                  1.722365 |                  17.939591 |                  3.250753 |                      0.0 |     1.586917 |              0.0 |         14.688838 |             35.308407 |                 100.0 |                 48.85611 |               57.260179 | 0.424519 |  0.0 |  0.541756 |  0.541756 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.424519 | 0.051574 | 0.004323 |  0.0 | 0.000547 |  0.0 | 0.051574 | 0.232165 |  0.06449 | 0.158244 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 132.902486 | 0.051574 |   0.0 | 0.004323 |   0.0 | 0.000547 | 0.230928 |   0.0 | 0.001236 |   0.0 | 0.000001 |   0.0 |  0.06449 |
    | 1997-08-05 10:00:00 | 0.833333 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 | 0.783333 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.095274 |                  2.095274 |                  18.027018 |                  3.266966 |                      0.0 |     1.959151 |              0.0 |         14.760052 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.241578 |  0.0 |  0.716789 |       0.7 | 0.010363 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.241578 | 0.029429 | 0.004335 |  0.0 | 0.000563 |  0.0 | 0.029429 | 0.227285 | 0.063135 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 | 133.099374 | 0.029429 |   0.0 | 0.004335 |   0.0 | 0.000563 | 0.226008 |   0.0 | 0.001276 |   0.0 | 0.000001 |   0.0 | 0.063135 |
    | 1997-08-05 11:00:00 | 0.833333 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 | 0.783333 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.098911 |                  2.098911 |                  18.201874 |                  3.309837 |                      0.0 |     1.961001 |              0.0 |         14.892037 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.746035 |       0.7 | 0.029836 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 | 0.221998 | 0.061666 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 133.137795 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.220681 |   0.0 | 0.001316 |   0.0 | 0.000001 |   0.0 | 0.061666 |
    | 1997-08-05 12:00:00 | 0.833333 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 | 0.783333 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.014418 |                  2.014418 |                  18.289301 |                  3.330376 |                      0.0 |     1.875652 |              0.0 |         14.958925 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.771401 |       0.7 | 0.046898 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010167 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010167 | 0.216599 | 0.060166 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 133.159147 | 0.010167 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.215243 |   0.0 | 0.001355 |   0.0 | 0.000001 |   0.0 | 0.060166 |
    | 1997-08-05 13:00:00 | 0.833333 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 | 0.783333 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.101265 |                  2.101265 |                  18.464157 |                  3.363175 |                      0.0 |     1.961133 |              0.0 |         15.100981 |             32.098552 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.840981 |       0.7 | 0.091225 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010169 | 0.004344 |  0.0 | 0.000573 |  0.0 | 0.010169 | 0.211341 | 0.058706 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 133.136169 | 0.010169 |   0.0 | 0.004344 |   0.0 | 0.000573 | 0.209946 |   0.0 | 0.001394 |   0.0 | 0.000001 |   0.0 | 0.058706 |
    | 1997-08-05 14:00:00 | 0.833333 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 | 0.783333 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.013503 |                  2.013503 |                  18.551584 |                   3.38194 |                      0.0 |     1.872589 |              0.0 |         15.169644 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.865172 |       0.7 | 0.105748 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010167 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010167 | 0.206219 | 0.057283 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 133.098671 | 0.010167 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.204785 |   0.0 | 0.001432 |   0.0 | 0.000001 |   0.0 | 0.057283 |
    | 1997-08-05 15:00:00 | 0.833333 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 | 0.783333 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.697346 |                  1.697346 |                  18.639012 |                  3.400029 |                      0.0 |     1.555678 |              0.0 |         15.238983 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.798412 |       0.7 | 0.063779 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 |  0.20123 | 0.055897 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 133.103149 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.199758 |   0.0 |  0.00147 |   0.0 | 0.000001 |   0.0 | 0.055897 |
    | 1997-08-05 16:00:00 | 0.833333 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 | 0.783333 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.231461 |                  1.231461 |                   18.72644 |                   3.42173 |                      0.0 |     1.088889 |              0.0 |         15.304709 |             30.049708 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.702711 |       0.7 | 0.001736 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 |  0.19637 | 0.054547 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 133.169669 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.194861 |   0.0 | 0.001508 |   0.0 | 0.000002 |   0.0 | 0.054547 |
    | 1997-08-05 17:00:00 | 0.833333 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 | 0.783333 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |              0.748608 |                  0.748608 |                   18.72644 |                  3.423912 |                      0.0 |     0.605945 |              0.0 |         15.302528 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |   0.55931 |   0.55931 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 |  0.01017 | 0.004344 |  0.0 | 0.000574 |  0.0 |  0.01017 | 0.191636 | 0.053232 |  0.14069 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 133.237914 |  0.01017 |   0.0 | 0.004344 |   0.0 | 0.000574 |  0.19009 |   0.0 | 0.001544 |   0.0 | 0.000002 |   0.0 | 0.053232 |
    | 1997-08-05 18:00:00 | 0.833333 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 | 0.783333 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |              0.364895 |                  0.364895 |                   18.72644 |                   3.42458 |                      0.0 |     0.222204 |              0.0 |          15.30186 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.224023 |  0.0 |  0.406398 |  0.406398 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.224023 | 0.027375 | 0.004346 |  0.0 | 0.000577 |  0.0 | 0.027375 | 0.187248 | 0.052013 | 0.293602 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 |  133.42964 | 0.027375 |   0.0 | 0.004346 |   0.0 | 0.000577 | 0.185665 |   0.0 | 0.001581 |   0.0 | 0.000002 |   0.0 | 0.052013 |
    | 1997-08-05 19:00:00 | 0.833333 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 | 0.783333 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.110333 |                  0.110333 |                   18.72644 |                  3.420162 |                      0.0 |    -0.032173 |              0.0 |         15.306277 |             37.166744 |                 100.0 |                 48.85611 |               57.260179 | 0.376935 |  0.0 |  0.257469 |  0.257469 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.376935 | 0.046172 | 0.004353 |  0.0 | 0.000585 |  0.0 | 0.046172 | 0.183439 | 0.050955 | 0.442531 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 133.755465 | 0.046172 |   0.0 | 0.004353 |   0.0 | 0.000585 | 0.181821 |   0.0 | 0.001617 |   0.0 | 0.000002 |   0.0 | 0.050955 |
    | 1997-08-05 20:00:00 | 0.833333 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 | 0.783333 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002372 |                  0.002372 |                  18.901295 |                  3.445356 |                      0.0 |    -0.141184 |              0.0 |         15.455939 |             56.493451 |                 100.0 |                 48.85611 |               84.453771 | 0.525865 |  0.0 |  0.102212 |  0.102212 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.525865 | 0.064653 | 0.004363 |  0.0 | 0.000598 |  0.0 | 0.064653 | 0.180211 | 0.050059 | 0.597788 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 134.211715 | 0.064653 |   0.0 | 0.004363 |   0.0 | 0.000598 | 0.178556 |   0.0 | 0.001652 |   0.0 | 0.000002 |   0.0 | 0.050059 |
    | 1997-08-05 21:00:00 | 0.833333 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 | 0.783333 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                  18.901295 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         15.466597 |             74.333488 |                 100.0 |                 48.85611 |               87.719298 | 0.681121 |  0.0 |  0.046259 |  0.046259 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.681121 | 0.084155 | 0.004378 |  0.0 | 0.000617 |  0.0 | 0.084155 | 0.177557 | 0.049321 | 0.653741 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 134.803686 | 0.084155 |   0.0 | 0.004378 |   0.0 | 0.000617 | 0.175867 |   0.0 | 0.001688 |   0.0 | 0.000002 |   0.0 | 0.049321 |
    | 1997-08-05 22:00:00 | 0.833333 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 | 0.783333 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                  18.901295 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         15.473739 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.737074 |  0.0 |  0.042514 |  0.042514 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.737074 | 0.091595 | 0.004397 |  0.0 | 0.000642 |  0.0 | 0.091595 |  0.17532 |   0.0487 | 0.657486 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 135.444126 | 0.091595 |   0.0 | 0.004397 |   0.0 | 0.000642 | 0.173595 |   0.0 | 0.001723 |   0.0 | 0.000002 |   0.0 |   0.0487 |
    | 1997-08-05 23:00:00 | 0.833333 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 | 0.783333 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                  18.901295 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         15.482962 |             52.308751 |                 100.0 |                 48.85611 |               87.719298 | 0.740819 |  0.0 |  0.035583 |  0.035583 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.740819 | 0.092614 | 0.004418 |  0.0 | 0.000669 |  0.0 | 0.092614 |  0.17325 | 0.048125 | 0.664417 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 | 136.087245 | 0.092614 |   0.0 | 0.004418 |   0.0 | 0.000669 |  0.17149 |   0.0 | 0.001758 |   0.0 | 0.000002 |   0.0 | 0.048125 |

.. _lland_v3_water_hourly:

water
-----

For our hourly integration test on water type |WASSER|, we deviate from the
input data of the daily simulation (:ref:`lland_v3_wasser_daily`) in setting
precipitation to the constant value of 0.05 mm/h.  Neither the actual (|EvI|)
or the potential evaporation (|EvPo|) does show a diurnal pattern due to
method |Return_Penman_V1| relying on daily aggregated input data (e.g.
|DailyNetRadiation|) only (the same holds for water types |FLUSS| and |SEE|):

.. integration-test::

    >>> lnk(WASSER)
    >>> inputs.nied.series = 0.05
    >>> test("lland_v3_water_hourly",
    ...      axis1=(fluxes.nkor, fluxes.evpo, fluxes.evi, fluxes.qah),
    ...      use_conditions=conditions_wasser)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.05 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 | 0.047 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |          2.249931 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036205 | 0.036205 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010795 | 0.002999 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002999 |
    | 1997-08-03 01:00:00 | 0.05 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 | 0.047 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          2.249283 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036073 | 0.036073 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010927 | 0.003035 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003035 |
    | 1997-08-03 02:00:00 | 0.05 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 | 0.047 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |          2.248772 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.035724 | 0.035724 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011276 | 0.003132 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003132 |
    | 1997-08-03 03:00:00 | 0.05 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 | 0.047 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          2.248763 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.035169 | 0.035169 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011831 | 0.003286 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003286 |
    | 1997-08-03 04:00:00 | 0.05 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 | 0.047 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | 0.0 | nan | 0.0 |              0.002099 |                  0.002099 |                   3.151127 |                  0.902341 |                      0.0 |    -0.035498 |              0.0 |          2.248786 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.034397 | 0.034397 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.012603 | 0.003501 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003501 |
    | 1997-08-03 05:00:00 | 0.05 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 | 0.047 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | 0.0 | nan | 0.0 |              0.023687 |                  0.023687 |                   3.151127 |                  0.901482 |                      0.0 |    -0.013875 |              0.0 |          2.249645 |            355.009444 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.033483 | 0.033483 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013517 | 0.003755 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003755 |
    | 1997-08-03 06:00:00 | 0.05 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 | 0.047 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | 0.0 | nan | 0.0 |              0.061837 |                  0.061837 |                   3.151127 |                   0.89977 |                      0.0 |     0.024346 |              0.0 |          2.251357 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.032456 | 0.032456 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.014544 |  0.00404 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00404 |
    | 1997-08-03 07:00:00 | 0.05 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 | 0.047 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | 0.0 | nan | 0.0 |              0.118079 |                  0.118079 |                   3.151127 |                  0.896826 |                      0.0 |     0.080712 |              0.0 |          2.254301 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.031557 | 0.031557 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015443 |  0.00429 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00429 |
    | 1997-08-03 08:00:00 | 0.05 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 | 0.047 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | 0.0 | nan | 0.0 |              0.184625 |                  0.184625 |                   3.151127 |                  0.891371 |                      0.0 |     0.147485 |              0.0 |          2.259756 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030534 | 0.030534 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016466 | 0.004574 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004574 |
    | 1997-08-03 09:00:00 | 0.05 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 | 0.047 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | 0.0 | nan | 0.0 |              0.336704 |                  0.336704 |                   3.220512 |                  0.922604 |                      0.0 |     0.298262 |              0.0 |          2.297908 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030275 | 0.030275 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016725 | 0.004646 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004646 |
    | 1997-08-03 10:00:00 | 0.05 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 | 0.047 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | 0.0 | nan | 0.0 |               0.54171 |                   0.54171 |                   3.289897 |                  0.955299 |                      0.0 |     0.501906 |              0.0 |          2.334598 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030379 | 0.030379 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016621 | 0.004617 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004617 |
    | 1997-08-03 11:00:00 | 0.05 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 | 0.047 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | 0.0 | nan | 0.0 |               0.66422 |                   0.66422 |                   3.463358 |                  1.039981 |                      0.0 |     0.620888 |              0.0 |          2.423377 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.031338 | 0.031338 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015662 | 0.004351 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004351 |
    | 1997-08-03 12:00:00 | 0.05 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 | 0.047 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | 0.0 | nan | 0.0 |              0.676668 |                  0.676668 |                   3.671512 |                  1.142661 |                      0.0 |     0.629057 |              0.0 |          2.528851 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.032665 | 0.032665 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.014335 | 0.003982 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003982 |
    | 1997-08-03 13:00:00 | 0.05 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 | 0.047 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | 0.0 | nan | 0.0 |              0.535824 |                  0.535824 |                   3.810281 |                  1.211985 |                      0.0 |     0.485325 |              0.0 |          2.598296 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.033843 | 0.033843 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013157 | 0.003655 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003655 |
    | 1997-08-03 14:00:00 | 0.05 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 | 0.047 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | 0.0 | nan | 0.0 |              0.453083 |                  0.453083 |                   3.879666 |                   1.24753 |                      0.0 |     0.401102 |              0.0 |          2.632136 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.034948 | 0.034948 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.012052 | 0.003348 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003348 |
    | 1997-08-03 15:00:00 | 0.05 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 | 0.047 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | 0.0 | nan | 0.0 |              0.418918 |                  0.418918 |                   4.018435 |                  1.316893 |                      0.0 |     0.364048 |              0.0 |          2.701542 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036637 | 0.036637 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010363 | 0.002879 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002879 |
    | 1997-08-03 16:00:00 | 0.05 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 | 0.047 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | 0.0 | nan | 0.0 |              0.300777 |                  0.300777 |                   4.122512 |                  1.370145 |                      0.0 |     0.243687 |              0.0 |          2.752367 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.038445 | 0.038445 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.008555 | 0.002376 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002376 |
    | 1997-08-03 17:00:00 | 0.05 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 | 0.047 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | 0.0 | nan | 0.0 |               0.14811 |                   0.14811 |                   4.226589 |                  1.421011 |                      0.0 |     0.088901 |              0.0 |          2.805578 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.040342 | 0.040342 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.006658 |  0.00185 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00185 |
    | 1997-08-03 18:00:00 | 0.05 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 | 0.047 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | 0.0 | nan | 0.0 |              0.055167 |                  0.055167 |                   4.261282 |                  1.433696 |                      0.0 |     -0.00457 |              0.0 |          2.827585 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.041848 | 0.041848 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005152 | 0.001431 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001431 |
    | 1997-08-03 19:00:00 | 0.05 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 | 0.047 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | 0.0 | nan | 0.0 |              0.014723 |                  0.014723 |                   4.261282 |                  1.423756 |                      0.0 |      -0.0446 |              0.0 |          2.837526 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.043061 | 0.043061 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.003939 | 0.001094 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001094 |
    | 1997-08-03 20:00:00 | 0.05 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 | 0.047 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | 0.0 | nan | 0.0 |              0.000201 |                  0.000201 |                   4.261282 |                  1.412078 |                      0.0 |    -0.058636 |              0.0 |          2.849203 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.043962 | 0.043962 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.003038 | 0.000844 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000844 |
    | 1997-08-03 21:00:00 | 0.05 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 | 0.047 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          2.859717 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.044514 | 0.044514 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.002486 |  0.00069 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00069 |
    | 1997-08-03 22:00:00 | 0.05 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 | 0.047 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          2.870124 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.044975 | 0.044975 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.002025 | 0.000563 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000563 |
    | 1997-08-03 23:00:00 | 0.05 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 | 0.047 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          2.878441 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.045525 | 0.045525 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.001475 |  0.00041 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00041 |
    | 1997-08-04 00:00:00 | 0.05 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 | 0.047 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          2.842823 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.04554 |  0.04554 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.00146 | 0.000406 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000406 |
    | 1997-08-04 01:00:00 | 0.05 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 | 0.047 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |           2.83273 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.045787 | 0.045787 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.001213 | 0.000337 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000337 |
    | 1997-08-04 02:00:00 | 0.05 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 | 0.047 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |          2.821294 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046085 | 0.046085 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000915 | 0.000254 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000254 |
    | 1997-08-04 03:00:00 | 0.05 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 | 0.047 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          2.809192 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046319 | 0.046319 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000681 | 0.000189 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000189 |
    | 1997-08-04 04:00:00 | 0.05 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 | 0.047 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | 0.0 | nan | 0.0 |              0.001718 |                  0.001718 |                   4.239363 |                  1.440702 |                      0.0 |    -0.058312 |              0.0 |           2.79866 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046528 | 0.046528 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000472 | 0.000131 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000131 |
    | 1997-08-04 05:00:00 | 0.05 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 | 0.047 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | 0.0 | nan | 0.0 |              0.021547 |                  0.021547 |                   4.239363 |                  1.451305 |                      0.0 |    -0.038924 |              0.0 |          2.788058 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046795 | 0.046795 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000205 | 0.000057 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000057 |
    | 1997-08-04 06:00:00 | 0.05 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 | 0.047 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | 0.0 | nan | 0.0 |              0.068025 |                  0.068025 |                   4.273966 |                  1.478175 |                      0.0 |     0.006434 |              0.0 |          2.795791 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.047182 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 07:00:00 | 0.05 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 | 0.047 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | 0.0 | nan | 0.0 |              0.308638 |                  0.308638 |                   4.585394 |                   1.64464 |                      0.0 |     0.240112 |              0.0 |          2.940754 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |   0.0492 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 08:00:00 | 0.05 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 | 0.047 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | 0.0 | nan | 0.0 |              0.513243 |                  0.513243 |                   4.931426 |                   1.83472 |                      0.0 |     0.436797 |              0.0 |          3.096706 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.051571 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 09:00:00 | 0.05 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 | 0.047 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | 0.0 | nan | 0.0 |              0.663127 |                  0.663127 |                   5.208251 |                  1.989515 |                      0.0 |     0.580231 |              0.0 |          3.218735 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.053462 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 10:00:00 | 0.05 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 | 0.047 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | 0.0 | nan | 0.0 |              0.695311 |                  0.695311 |                   5.346663 |                   2.06621 |                      0.0 |     0.609219 |              0.0 |          3.280453 |             68.711505 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.054455 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 11:00:00 | 0.05 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 | 0.047 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | 0.0 | nan | 0.0 |              0.648284 |                  0.648284 |                   5.381266 |                  2.087634 |                      0.0 |     0.561299 |              0.0 |          3.293633 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.054881 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 12:00:00 | 0.05 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 | 0.047 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | 0.0 | nan | 0.0 |              0.670155 |                  0.670155 |                    5.41587 |                  2.106001 |                      0.0 |     0.582405 |              0.0 |          3.309868 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.055387 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 13:00:00 | 0.05 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 | 0.047 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | 0.0 | nan | 0.0 |              0.647397 |                  0.647397 |                   5.519679 |                  2.166369 |                      0.0 |     0.557131 |              0.0 |           3.35331 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.056372 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 14:00:00 | 0.05 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 | 0.047 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | 0.0 | nan | 0.0 |              0.662635 |                  0.662635 |                   5.692695 |                  2.265756 |                      0.0 |     0.568229 |              0.0 |          3.426939 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.057374 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 15:00:00 | 0.05 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 | 0.047 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | 0.0 | nan | 0.0 |              0.558476 |                  0.558476 |                   5.831107 |                  2.348437 |                      0.0 |     0.460624 |              0.0 |           3.48267 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.058283 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 16:00:00 | 0.05 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 | 0.047 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | 0.0 | nan | 0.0 |              0.408064 |                  0.408064 |                    5.96952 |                  2.428674 |                      0.0 |     0.306869 |              0.0 |          3.540846 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.05932 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 17:00:00 | 0.05 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 | 0.047 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | 0.0 | nan | 0.0 |              0.272381 |                  0.272381 |                   6.211742 |                  2.564427 |                      0.0 |      0.16553 |              0.0 |          3.647315 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061051 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 18:00:00 | 0.05 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 | 0.047 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | 0.0 | nan | 0.0 |              0.138863 |                  0.138863 |                    6.52317 |                  2.744116 |                      0.0 |     0.024525 |              0.0 |          3.779055 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063121 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 19:00:00 | 0.05 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 | 0.047 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | 0.0 | nan | 0.0 |              0.045438 |                  0.045438 |                   6.869202 |                  2.950104 |                      0.0 |    -0.077483 |              0.0 |          3.919098 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |   0.0652 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 20:00:00 | 0.05 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 | 0.047 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | 0.0 | nan | 0.0 |              0.000878 |                  0.000878 |                   6.973011 |                  3.034376 |                      0.0 |    -0.125554 |              0.0 |          3.938635 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.06592 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 21:00:00 | 0.05 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 | 0.047 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |          3.912334 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065866 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 22:00:00 | 0.05 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 | 0.047 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |          3.888687 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065746 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 23:00:00 | 0.05 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 | 0.047 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |          3.869372 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065422 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 00:00:00 | 0.05 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 | 0.047 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |          3.818763 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064471 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 01:00:00 | 0.05 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 | 0.047 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |          3.807867 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063963 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 02:00:00 | 0.05 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 | 0.047 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |          3.798476 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063323 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 03:00:00 | 0.05 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 | 0.047 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          3.789264 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062667 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 04:00:00 | 0.05 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 | 0.047 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | 0.0 | nan | 0.0 |              0.001601 |                  0.001601 |                   6.943374 |                  3.166293 |                      0.0 |    -0.130328 |              0.0 |          3.777081 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062183 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 05:00:00 | 0.05 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 | 0.047 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | 0.0 | nan | 0.0 |              0.022826 |                  0.022826 |                   6.943374 |                  3.179808 |                      0.0 |    -0.109666 |              0.0 |          3.763566 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061583 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 06:00:00 | 0.05 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 | 0.047 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | 0.0 | nan | 0.0 |              0.135954 |                  0.135954 |                   7.115928 |                  3.288825 |                      0.0 |     -0.00108 |              0.0 |          3.827103 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062021 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 07:00:00 | 0.05 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 | 0.047 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | 0.0 | nan | 0.0 |              0.330431 |                  0.330431 |                   7.115928 |                  3.287363 |                      0.0 |     0.193458 |              0.0 |          3.828566 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061775 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 08:00:00 | 0.05 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 | 0.047 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | 0.0 | nan | 0.0 |              0.551993 |                  0.551993 |                   7.115928 |                  3.279458 |                      0.0 |     0.415349 |              0.0 |          3.836471 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061896 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 09:00:00 | 0.05 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 | 0.047 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | 0.0 | nan | 0.0 |              0.679881 |                  0.679881 |                   7.081417 |                  3.250753 |                      0.0 |     0.544433 |              0.0 |          3.830665 |             53.251417 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062093 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 10:00:00 | 0.05 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 | 0.047 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | 0.0 | nan | 0.0 |              0.827082 |                  0.827082 |                   7.115928 |                  3.266966 |                      0.0 |     0.690958 |              0.0 |          3.848962 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062733 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 11:00:00 | 0.05 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 | 0.047 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | 0.0 | nan | 0.0 |              0.828518 |                  0.828518 |                    7.18495 |                  3.309837 |                      0.0 |     0.690608 |              0.0 |          3.875113 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063478 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 12:00:00 | 0.05 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 | 0.047 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | 0.0 | nan | 0.0 |              0.795165 |                  0.795165 |                   7.219461 |                  3.330376 |                      0.0 |     0.656399 |              0.0 |          3.889085 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064117 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 13:00:00 | 0.05 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 | 0.047 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | 0.0 | nan | 0.0 |              0.829447 |                  0.829447 |                   7.288483 |                  3.363175 |                      0.0 |     0.689314 |              0.0 |          3.925307 |             48.410379 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064896 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 14:00:00 | 0.05 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 | 0.047 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | 0.0 | nan | 0.0 |              0.794804 |                  0.794804 |                   7.322994 |                   3.38194 |                      0.0 |      0.65389 |              0.0 |          3.941053 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065782 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 15:00:00 | 0.05 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 | 0.047 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | 0.0 | nan | 0.0 |              0.670005 |                  0.670005 |                   7.357505 |                  3.400029 |                      0.0 |     0.528337 |              0.0 |          3.957476 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.066681 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 16:00:00 | 0.05 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 | 0.047 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | 0.0 | nan | 0.0 |              0.486103 |                  0.486103 |                   7.392016 |                   3.42173 |                      0.0 |     0.343531 |              0.0 |          3.970285 |             45.320355 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.06748 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 17:00:00 | 0.05 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 | 0.047 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | 0.0 | nan | 0.0 |              0.295503 |                  0.295503 |                   7.392016 |                  3.423912 |                      0.0 |      0.15284 |              0.0 |          3.968104 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.068039 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 18:00:00 | 0.05 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 | 0.047 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | 0.0 | nan | 0.0 |              0.144038 |                  0.144038 |                   7.392016 |                   3.42458 |                      0.0 |     0.001347 |              0.0 |          3.967436 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.068633 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 19:00:00 | 0.05 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 | 0.047 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | 0.0 | nan | 0.0 |              0.043553 |                  0.043553 |                   7.392016 |                  3.420162 |                      0.0 |    -0.098954 |              0.0 |          3.971854 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.069308 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 20:00:00 | 0.05 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 | 0.047 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | 0.0 | nan | 0.0 |              0.000936 |                  0.000936 |                   7.461037 |                  3.445356 |                      0.0 |     -0.14262 |              0.0 |          4.015681 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.070109 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 21:00:00 | 0.05 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 | 0.047 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |          4.026339 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.070708 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 22:00:00 | 0.05 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 | 0.047 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |          4.033482 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.071405 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 23:00:00 | 0.05 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 | 0.047 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |          4.042705 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.07218 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

    >>> inputs.nied.series = 0.0

.. _lland_v3_nadelw_summer_hourly:

conifers (summer)
-----------------

Now we repeat the daily integration test :ref:`lland_v3_nadelw_summer_daily`
on an hourly basis. In contrast to the calculation for water areas (see
:ref:`lland_v3_water_hourly`), both the interception evaporation (|EvI|)
and the soil evapotranspiration (|EvB|) rely on hourly input data (see method
|Return_PenmanMonteith_V1|) and thus show a diurnal pattern.  |lland_v3|
always adds condensed water to the interception storage (|Inzp|).  Hence,
|EvI| is sometimes negative during the nighttime, but |Inzp| dries out as
soon as the sun starts to rise:

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v3_nadelw_summer_hourly",
    ...      axis1=(fluxes.evi, fluxes.evb),
    ...      axis2=states.inzp,
    ...      use_conditions=conditions_nadelw_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.283992 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   9.243306 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |          8.342111 |                 117.5 |                 100.0 |               170.875609 |               69.444444 |  0.0 |  0.0 | -0.001025 | -0.001025 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.221929 | 0.061647 | 0.001025 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.221929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061647 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.283992 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   9.243306 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          8.341463 |                 117.5 |                 100.0 |               173.430933 |               69.444444 |  0.0 |  0.0 | -0.000655 | -0.000655 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.216196 | 0.060055 |  0.00168 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.216196 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060055 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.283992 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   9.243306 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |          8.340951 |                 117.5 |                 100.0 |               173.240523 |               69.444444 |  0.0 |  0.0 | -0.002635 | -0.002635 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.210612 | 0.058503 | 0.004315 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.210612 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058503 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.283992 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   9.243306 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          8.340943 |                 117.5 |                 100.0 |               172.677019 |               69.444444 |  0.0 |  0.0 | -0.004198 | -0.004198 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.205172 | 0.056992 | 0.008513 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.205172 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056992 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.283992 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.006158 |                  0.001144 |                   9.243306 |                  0.902341 |                      0.0 |     -0.03144 |              0.0 |          8.340965 |                 117.5 |                 100.0 |                174.22122 |               93.330305 |  0.0 |  0.0 | -0.003798 | -0.003798 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.199872 |  0.05552 | 0.012311 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.199872 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05552 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.212994 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.069482 |                  0.012904 |                   9.243306 |                  0.901482 |                      0.0 |      0.03192 |              0.0 |          8.341824 |            156.666667 |                 100.0 |               174.539602 |              172.004443 |  0.0 |  0.0 |  0.008314 |  0.008314 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.194709 | 0.054086 | 0.003996 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.194709 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.054086 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.319491 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.181388 |                  0.033686 |                   9.243306 |                   0.89977 |                      0.0 |     0.143897 |              0.0 |          8.343537 |            104.444444 |                 100.0 |               173.411109 |              170.929929 |  0.0 |  0.0 |  0.034791 |  0.003996 | 0.017757 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.189679 | 0.052689 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 78.706923 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.189679 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052689 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.319491 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.346366 |                  0.064325 |                   9.243306 |                  0.896826 |                      0.0 |     0.308999 |              0.0 |           8.34648 |            104.444444 |                 100.0 |                169.26657 |              166.979573 |  0.0 |  0.0 |  0.073429 |       0.0 | 0.043065 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18478 | 0.051328 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 78.663858 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18478 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.051328 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.319491 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.541567 |                  0.100577 |                   9.243306 |                  0.891371 |                      0.0 |     0.504426 |              0.0 |          8.351935 |            104.444444 |                 100.0 |               164.539821 |              162.466484 |  0.0 |  0.0 |  0.121304 |       0.0 | 0.072626 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.180007 | 0.050002 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 78.591232 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.180007 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050002 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           0.461488 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.987664 |                  0.183423 |                   9.446835 |                  0.922604 |                      0.0 |     0.949222 |              0.0 |          8.524231 |             72.307692 |                 100.0 |               175.701188 |              173.109975 |  0.0 |  0.0 |  0.278008 |       0.0 | 0.144385 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.175357 |  0.04871 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 78.446847 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.175357 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04871 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           0.532486 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.589016 |                  0.295103 |                   9.650363 |                  0.955299 |                      0.0 |     1.549212 |              0.0 |          8.695064 |             62.666667 |                 100.0 |                 197.9318 |              194.171796 |  0.0 |  0.0 |  0.484138 |       0.0 | 0.227779 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.170827 | 0.047452 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 78.219068 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.170827 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047452 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.425988 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.948379 |                  0.361842 |                  10.159184 |                  1.039981 |                      0.0 |     1.905047 |              0.0 |          9.119203 |             78.333333 |                 100.0 |               224.347905 |              218.964094 |  0.0 |  0.0 |  0.579474 |       0.0 | 0.287656 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.166415 | 0.046226 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 77.931412 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166415 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046226 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           0.461488 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.984892 |                  0.368623 |                  10.769769 |                  1.142661 |                      0.0 |     1.937282 |              0.0 |          9.627108 |             72.307692 |                 100.0 |               245.282288 |              238.432816 |  0.0 |  0.0 |  0.611134 |       0.0 |  0.27972 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.162116 | 0.045032 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 77.651692 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.162116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045032 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           0.532486 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.571751 |                  0.291897 |                  11.176825 |                  1.211985 |                      0.0 |     1.521251 |              0.0 |           9.96484 |             62.666667 |                 100.0 |               259.971088 |              251.999948 |  0.0 |  0.0 |  0.532257 |       0.0 | 0.222625 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.157929 | 0.043869 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 77.429067 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043869 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           0.674482 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.329043 |                  0.246822 |                  11.380354 |                   1.24753 |                      0.0 |     1.277062 |              0.0 |         10.132824 |             49.473684 |                 100.0 |               297.295394 |              286.132852 |  0.0 |  0.0 |  0.526288 |       0.0 | 0.183586 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.153849 | 0.042736 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.245482 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.153849 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042736 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           0.674482 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |              1.228827 |                  0.228211 |                   11.78741 |                  1.316893 |                      0.0 |     1.173957 |              0.0 |         10.470517 |             49.473684 |                 100.0 |               302.773995 |              291.102238 |  0.0 |  0.0 |  0.500938 |       0.0 | 0.172023 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.149875 | 0.041632 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 77.073458 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.149875 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041632 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |           0.816478 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.882278 |                  0.163852 |                  12.092703 |                  1.370145 |                      0.0 |     0.825189 |              0.0 |         10.722558 |             40.869565 |                 100.0 |               310.898811 |              298.452854 |  0.0 |  0.0 |  0.446874 |       0.0 | 0.133947 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.146004 | 0.040557 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 76.939511 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146004 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040557 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           0.851977 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.434455 |                  0.080685 |                  12.397995 |                  1.421011 |                      0.0 |     0.375246 |              0.0 |         10.976984 |             39.166667 |                 100.0 |               291.447874 |              280.817391 |  0.0 |  0.0 |  0.321464 |       0.0 | 0.096503 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.142232 | 0.039509 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 76.843008 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.142232 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039509 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           0.887476 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.161824 |                  0.030053 |                  12.499759 |                  1.433696 |                      0.0 |     0.102086 |              0.0 |         11.066063 |                  37.6 |                 100.0 |               252.719347 |              245.311527 |  0.0 |  0.0 |  0.222652 |       0.0 | 0.070699 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.138558 | 0.038488 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 |  76.77231 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.138558 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038488 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           0.887476 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.043189 |                  0.008021 |                  12.499759 |                  1.423756 |                      0.0 |    -0.016134 |              0.0 |         11.076003 |                  37.6 |                 100.0 |               212.737062 |              208.098179 |  0.0 |  0.0 |  0.143053 |       0.0 | 0.049883 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.134979 | 0.037494 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 76.722426 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.134979 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037494 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           0.780979 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000589 |                  0.000109 |                  12.499759 |                  1.412078 |                      0.0 |    -0.058248 |              0.0 |         11.087681 |             42.727273 |                 100.0 |               190.515534 |               75.961268 |  0.0 |  0.0 |  0.077666 |       0.0 | 0.047501 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.131493 | 0.036526 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 76.674926 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.131493 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036526 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           0.603484 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  12.499759 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |         11.098195 |             55.294118 |                 100.0 |               185.552232 |               69.444444 |  0.0 |  0.0 |  0.033668 |       0.0 | 0.022649 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.128096 | 0.035582 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 76.652277 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.128096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035582 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           0.603484 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  12.499759 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |         11.108602 |             55.294118 |                 100.0 |               184.347096 |               69.444444 |  0.0 |  0.0 |  0.020838 |       0.0 | 0.013901 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.124787 | 0.034663 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 | 76.638376 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.124787 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034663 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |           0.816478 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  12.499759 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |         11.116919 |             40.869565 |                 100.0 |               187.258263 |               69.444444 |  0.0 |  0.0 |  0.027773 |       0.0 | 0.016657 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.121564 | 0.033768 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.621719 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033768 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |           0.816478 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  12.435463 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |         11.038924 |             40.869565 |                 100.0 |               196.850637 |               69.444444 |  0.0 |  0.0 |  0.041216 |       0.0 |  0.02456 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.118424 | 0.032895 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.597159 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.118424 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032895 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           0.780979 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  12.435463 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          11.02883 |             42.727273 |                 100.0 |               198.635637 |               69.444444 |  0.0 |  0.0 |  0.032732 |       0.0 | 0.019695 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.115365 | 0.032046 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.577464 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.115365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032046 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |           0.816478 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  12.435463 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |         11.017395 |             40.869565 |                 100.0 |               201.155121 |               69.444444 |  0.0 |  0.0 |  0.036288 |       0.0 | 0.021425 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.112385 | 0.031218 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 |  76.55604 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.112385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031218 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           0.709981 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  12.435463 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |         11.005293 |                  47.0 |                 100.0 |               202.122033 |               69.444444 |  0.0 |  0.0 |  0.025988 |       0.0 | 0.016073 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.109482 | 0.030412 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.539966 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.109482 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030412 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |           0.816478 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.005039 |                  0.000936 |                  12.435463 |                  1.440702 |                      0.0 |    -0.054991 |              0.0 |         10.994761 |             40.869565 |                 100.0 |               201.760902 |               93.626696 |  0.0 |  0.0 |  0.016861 |       0.0 | 0.008591 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.106654 | 0.029626 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 76.531375 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.106654 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029626 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           0.887476 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.063204 |                  0.011738 |                  12.435463 |                  1.451305 |                      0.0 |     0.002733 |              0.0 |         10.984159 |                  37.6 |                 100.0 |               202.699444 |              198.665144 |  0.0 |  0.0 |  0.029878 |       0.0 | 0.009282 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.103899 | 0.028861 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.522093 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.103899 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028861 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           0.851977 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.199539 |                  0.037057 |                  12.536966 |                  1.478175 |                      0.0 |     0.137948 |              0.0 |         11.058791 |             39.166667 |                 100.0 |               199.052037 |              195.228329 |  0.0 |  0.0 |  0.056009 |       0.0 | 0.018144 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.101215 | 0.028115 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.503949 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.101215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028115 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           0.887476 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |               0.90534 |                  0.168134 |                  13.450489 |                   1.64464 |                      0.0 |     0.836813 |              0.0 |         11.805849 |                  37.6 |                 100.0 |               191.292078 |              187.900211 |  0.0 |  0.0 |  0.234858 |       0.0 | 0.077952 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.098601 | 0.027389 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.425997 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.098601 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027389 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           0.887476 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.505514 |                  0.279595 |                  14.465515 |                   1.83472 |                      0.0 |     1.429067 |              0.0 |         12.630795 |                  37.6 |                 100.0 |               192.882798 |              189.404207 |  0.0 |  0.0 |  0.424426 |       0.0 | 0.144453 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.096054 | 0.026682 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.281544 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.096054 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026682 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           1.029472 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.945173 |                  0.361246 |                  15.277535 |                  1.989515 |                      0.0 |     1.862277 |              0.0 |          13.28802 |             32.413793 |                 100.0 |               203.810564 |              199.711145 |  0.0 |  0.0 |  0.613278 |       0.0 | 0.189597 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.093572 | 0.025992 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.091948 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.093572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025992 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |            1.10047 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              2.039579 |                  0.378779 |                  15.683546 |                   2.06621 |                      0.0 |     1.953487 |              0.0 |         13.617336 |             30.322581 |                 100.0 |               207.855185 |              203.514925 |  0.0 |  0.0 |  0.685505 |       0.0 | 0.206707 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.091155 | 0.025321 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 |  75.88524 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.091155 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025321 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           1.171468 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |              1.901633 |                   0.35316 |                  15.785048 |                  2.087634 |                      0.0 |     1.814648 |              0.0 |         13.697414 |             28.484848 |                 100.0 |               216.999554 |              212.092863 |  0.0 |  0.0 |  0.688399 |       0.0 | 0.197407 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.088801 | 0.024667 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 75.687833 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088801 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024667 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           1.171468 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.965789 |                  0.365075 |                  15.886551 |                  2.106001 |                      0.0 |     1.878039 |              0.0 |          13.78055 |             28.484848 |                 100.0 |               245.428784 |              238.568502 |  0.0 |  0.0 |  0.754061 |       0.0 | 0.202161 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.086507 |  0.02403 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.485672 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.086507 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02403 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           1.135969 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |               1.89903 |                  0.352677 |                  16.191059 |                  2.166369 |                      0.0 |     1.808765 |              0.0 |          14.02469 |                29.375 |                 100.0 |               301.593569 |              290.032403 |  0.0 |  0.0 |  0.791859 |       0.0 | 0.191425 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.084272 | 0.023409 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 75.294247 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.084272 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023409 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           0.887476 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |               1.94373 |                  0.360978 |                  16.698571 |                  2.265756 |                      0.0 |     1.849324 |              0.0 |         14.432816 |                  37.6 |                 100.0 |               310.085747 |              297.718286 |  0.0 |  0.0 |  0.744343 |       0.0 | 0.208898 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.082096 | 0.022804 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 75.085349 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.082096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022804 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           1.029472 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.638195 |                  0.304236 |                  17.104582 |                  2.348437 |                      0.0 |     1.540344 |              0.0 |         14.756144 |             32.413793 |                 100.0 |               324.688901 |              310.877147 |  0.0 |  0.0 |  0.707659 |       0.0 |  0.17376 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079975 | 0.022215 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 | 74.911589 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079975 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022215 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           1.277965 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.196986 |                  0.222297 |                  17.510592 |                  2.428674 |                      0.0 |     1.095792 |              0.0 |         15.081918 |             26.111111 |                 100.0 |               351.951333 |              335.249533 |  0.0 |  0.0 |  0.668848 |       0.0 |  0.13417 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077909 | 0.021641 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.777419 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077909 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021641 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |            1.49096 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.798984 |                  0.148383 |                   18.22111 |                  2.564427 |                      0.0 |     0.692133 |              0.0 |         15.656683 |             22.380952 |                 100.0 |               341.504263 |              325.939578 |  0.0 |  0.0 |  0.608166 |       0.0 | 0.109849 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075897 | 0.021082 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 |  74.66757 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075897 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021082 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |            1.49096 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.407331 |                  0.075647 |                  19.134633 |                  2.744116 |                      0.0 |     0.292993 |              0.0 |         16.390518 |             22.380952 |                 100.0 |               298.906719 |              287.595486 |  0.0 |  0.0 |  0.461333 |       0.0 | 0.089872 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073936 | 0.020538 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 74.577699 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073936 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020538 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           1.277965 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.133286 |                  0.024753 |                  20.149659 |                  2.950104 |                      0.0 |     0.010365 |              0.0 |         17.199555 |             26.111111 |                 100.0 |               249.928887 |              242.732867 |  0.0 |  0.0 |  0.272979 |       0.0 | 0.065761 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.072026 | 0.020007 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 74.511937 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.072026 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020007 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           1.135969 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002575 |                  0.000478 |                  20.454167 |                  3.034376 |                      0.0 |    -0.123857 |              0.0 |         17.419791 |                29.375 |                 100.0 |               232.156834 |               74.878388 |  0.0 |  0.0 |  0.160746 |       0.0 | 0.083418 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.070166 | 0.019491 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 74.428519 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.070166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019491 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           0.780979 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  20.454167 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         17.393489 |             42.727273 |                 100.0 |               219.836505 |               69.444444 |  0.0 |  0.0 |  0.056353 |       0.0 | 0.034084 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.068353 | 0.018987 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 74.394435 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.068353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018987 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           0.780979 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  20.454167 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         17.369842 |             42.727273 |                 100.0 |               216.481047 |               69.444444 |  0.0 |  0.0 |  0.030452 |       0.0 | 0.018184 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.066588 | 0.018497 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 |  74.37625 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.066588 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018497 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |            0.74548 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  20.454167 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         17.350527 |             44.761905 |                 100.0 |               216.446595 |               69.444444 |  0.0 |  0.0 |  0.012642 |       0.0 | 0.007611 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064868 | 0.018019 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018019 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           0.638983 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   20.36723 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         17.242619 |             52.222222 |                 100.0 |               219.889739 |               69.444444 |  0.0 |  0.0 | -0.000942 | -0.000942 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.063192 | 0.017553 | 0.000942 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.063192 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017553 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           0.638983 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   20.36723 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         17.231723 |             52.222222 |                 100.0 |               223.677492 |               69.444444 |  0.0 |  0.0 | -0.007156 | -0.007156 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06156 |   0.0171 | 0.008098 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06156 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0171 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           0.567985 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   20.36723 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         17.222331 |                 58.75 |                 100.0 |               227.175049 |               69.444444 |  0.0 |  0.0 |  -0.01403 |  -0.01403 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05997 | 0.016658 | 0.022128 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05997 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016658 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.425988 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   20.36723 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          17.21312 |             78.333333 |                 100.0 |               229.328233 |               69.444444 |  0.0 |  0.0 | -0.019316 | -0.019316 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.058421 | 0.016228 | 0.041443 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.058421 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016228 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           0.674482 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004696 |                  0.000872 |                   20.36723 |                  3.166293 |                      0.0 |    -0.127233 |              0.0 |         17.200937 |             49.473684 |                 100.0 |               232.312366 |               93.057374 |  0.0 |  0.0 | -0.016434 | -0.016434 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056912 | 0.015809 | 0.057877 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056912 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015809 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           0.461488 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.066957 |                  0.012435 |                   20.36723 |                  3.179808 |                      0.0 |    -0.065535 |              0.0 |         17.187422 |             72.307692 |                 100.0 |               234.493634 |              228.419105 |  0.0 |  0.0 | -0.005056 | -0.005056 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.055442 |   0.0154 | 0.062933 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.055442 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0154 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           0.567985 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.398799 |                  0.074063 |                   20.87339 |                  3.288825 |                      0.0 |     0.261764 |              0.0 |         17.584565 |                 58.75 |                 100.0 |               222.635333 |              217.364459 |  0.0 |  0.0 |   0.07078 |  0.062933 | 0.002912 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054009 | 0.015003 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 74.365727 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054009 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015003 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           0.674482 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.969265 |                  0.180006 |                   20.87339 |                  3.287363 |                      0.0 |     0.832292 |              0.0 |         17.586027 |             49.473684 |                 100.0 |               196.965142 |              193.259739 |  0.0 |  0.0 |  0.207904 |       0.0 | 0.078639 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052614 | 0.014615 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 74.287088 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014615 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           1.135969 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.619179 |                  0.300705 |                   20.87339 |                  3.279458 |                      0.0 |     1.482535 |              0.0 |         17.593932 |                29.375 |                 100.0 |               189.864158 |              186.549347 |  0.0 |  0.0 |  0.442855 |       0.0 | 0.131304 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.051255 | 0.014238 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 74.155785 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.051255 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014238 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           1.419962 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.994317 |                  0.370373 |                  20.772158 |                  3.250753 |                      0.0 |     1.858869 |              0.0 |         17.521405 |                  23.5 |                 100.0 |               202.881364 |              198.836433 |  0.0 |  0.0 |  0.679013 |       0.0 | 0.175093 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049931 |  0.01387 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 73.980691 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049931 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01387 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |           1.632956 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.426107 |                  0.450563 |                   20.87339 |                  3.266966 |                      0.0 |     2.289983 |              0.0 |         17.606424 |             20.434783 |                 100.0 |               224.441968 |              219.051923 |  0.0 |  0.0 |  0.901883 |       0.0 | 0.201952 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048642 | 0.013512 |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 |  73.77874 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048642 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013512 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           1.455461 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.430318 |                  0.451345 |                  21.075854 |                  3.309837 |                      0.0 |     2.292408 |              0.0 |         17.766017 |             22.926829 |                 100.0 |               261.395836 |              253.311837 |  0.0 |  0.0 |  0.942976 |       0.0 | 0.209722 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047385 | 0.013163 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 73.569018 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013163 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           1.455461 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.332484 |                  0.433176 |                  21.177086 |                  3.330376 |                      0.0 |     2.193718 |              0.0 |         17.846709 |             22.926829 |                 100.0 |               315.484528 |              302.591605 |  0.0 |  0.0 |  0.983393 |       0.0 | 0.196078 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046161 | 0.012823 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 73.372939 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046161 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012823 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           1.561958 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.433044 |                  0.451851 |                   21.37955 |                  3.363175 |                      0.0 |     2.292911 |              0.0 |         18.016374 |             21.363636 |                 100.0 |               357.583565 |              340.253538 |  0.0 |  0.0 |  1.076015 |       0.0 | 0.188119 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044969 | 0.012491 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 73.184821 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044969 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012491 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |           1.632956 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.331424 |                  0.432979 |                  21.480782 |                   3.38194 |                      0.0 |      2.19051 |              0.0 |         18.098841 |             20.434783 |                 100.0 |               429.701201 |              403.402315 |  0.0 |  0.0 |  1.117473 |       0.0 | 0.165048 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043807 | 0.012169 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 73.019773 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043807 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012169 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           1.597457 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.965348 |                  0.364993 |                  21.582014 |                  3.400029 |                      0.0 |      1.82368 |              0.0 |         18.181985 |             20.888889 |                 100.0 |               525.329939 |              484.576285 |  0.0 |  0.0 |  1.044793 |       0.0 | 0.135822 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042676 | 0.011854 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 72.883951 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042676 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011854 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           1.668455 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.425902 |                   0.26481 |                  21.683246 |                   3.42173 |                      0.0 |      1.28333 |              0.0 |         18.261515 |                  20.0 |                 100.0 |               627.911114 |              568.561649 |  0.0 |  0.0 |  0.943159 |       0.0 | 0.103418 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041573 | 0.011548 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 72.780532 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041573 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011548 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           1.597457 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |               0.86681 |                  0.160979 |                  21.683246 |                  3.423912 |                      0.0 |     0.724147 |              0.0 |         18.259334 |             20.888889 |                 100.0 |               687.355785 |              615.833261 |  0.0 |  0.0 |  0.773152 |       0.0 | 0.081869 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040499 |  0.01125 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 72.698663 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040499 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01125 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           1.455461 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |               0.42251 |                  0.078466 |                  21.683246 |                   3.42458 |                      0.0 |     0.279819 |              0.0 |         18.258666 |             22.926829 |                 100.0 |               600.040442 |              546.050102 |  0.0 |  0.0 |  0.580608 |       0.0 | 0.073132 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039453 | 0.010959 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 72.625531 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039453 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010959 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           1.348964 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.127754 |                  0.023726 |                  21.683246 |                  3.420162 |                      0.0 |    -0.014752 |              0.0 |         18.263084 |             24.736842 |                 100.0 |               384.261969 |              363.812822 |  0.0 |  0.0 |   0.38361 |       0.0 | 0.070385 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.038434 | 0.010676 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 72.555146 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.038434 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010676 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           0.887476 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002747 |                   0.00051 |                   21.88571 |                  3.445356 |                      0.0 |    -0.140809 |              0.0 |         18.440354 |                  37.6 |                 100.0 |               272.690342 |                73.37327 |  0.0 |  0.0 |  0.163075 |       0.0 | 0.099136 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.037441 |   0.0104 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 72.456009 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037441 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0104 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           0.674482 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                   21.88571 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         18.451012 |             49.473684 |                 100.0 |               228.581729 |               69.444444 |  0.0 |  0.0 |  0.080802 |       0.0 | 0.053476 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.036474 | 0.010132 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 72.402533 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036474 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010132 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           0.780979 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                   21.88571 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         18.458154 |             42.727273 |                 100.0 |               220.437523 |               69.444444 |  0.0 |  0.0 |  0.076279 |       0.0 | 0.047612 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035532 |  0.00987 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 72.354921 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.035532 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00987 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           0.958474 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                   21.88571 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         18.467377 |             34.814815 |                 100.0 |               211.424945 |               69.444444 |  0.0 |  0.0 |  0.067295 |       0.0 | 0.038511 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.034614 | 0.009615 |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 |  72.31641 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009615 |

.. _lland_v3_acker_winter_hourly:

acre (winter)
-------------

In our hourly integration tests covering snow processes, we focus on the
subperiod from December 10th to December 12th, where, according to the
daily simulations, some melting occurs:

>>> pub.timegrids = "2010-12-10", "2010-12-13", "1h"

The following input values agree with the ones of the respective sub-period
of the daily integration test :ref:`lland_v3_acker_winter_daily`:

>>> inputs.nied.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.3, 0.2, 0.4, 0.7, 0.5, 1.0, 2.0, 2.1, 1.8, 1.3,
...     0.5, 0.2, 0.5, 0.2, 0.1, 0.3, 0.1, 0.1, 0.0, 0.1, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.1, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.teml.series = (
...     -3.7, -4.8, -5.6, -6.2, -7.3, -8.3, -8.6, -8.5, -9.5, -9.2, -6.9,
...     -4.0, -2.8, -1.4, -0.4, 0.5, 1.0, 1.1, 1.0, 1.4, 1.6, 2.0, 2.4, 2.5,
...     2.7, 3.2, 3.9, 4.4, 4.8, 4.8, 5.1, 5.1, 5.4, 5.3, 5.5, 6.1, 6.4, 6.4,
...     6.5, 6.3, 5.9, 5.5, 5.1, 4.8, 4.9, 4.9, 4.7, 3.8, 3.9, 3.7, 3.5, 3.3,
...     3.3, 3.0, 1.9, 1.2, 0.8, 0.8, 0.5, 0.3, -0.1, -0.4, -0.6, -0.8, -1.3,
...     -1.6, -2.0, -2.1, -2.2, -2.6, -3.1, -3.9)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.2, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     3.4, 2.9, 2.5, 3.0, 2.5, 2.3, 2.1, 1.4, 1.1, 1.0, 1.4, 1.3, 1.3, 1.4,
...     2.3, 2.8, 3.1, 3.5, 4.4, 3.8, 5.7, 5.6, 5.8, 6.2, 5.6, 5.2, 5.7, 6.6,
...     7.0, 7.3, 7.2, 7.6, 7.3, 7.1, 7.1, 8.9, 9.5, 7.9, 9.6, 9.1, 8.7, 8.9,
...     7.9, 7.1, 6.4, 7.2, 7.0, 6.8, 6.4, 6.3, 5.7, 5.2, 4.6, 4.9, 6.3, 5.8,
...     6.1, 5.9, 6.4, 6.8, 7.0, 6.3, 7.3, 7.5, 7.0, 6.0, 4.8, 4.2, 4.9, 4.0,
...     3.4, 3.4)
>>> inputs.relativehumidity.series = (
...     87.5, 90.7, 89.7, 91.3, 94.2, 93.2, 96.5, 95.9, 96.8, 94.3, 96.7,
...     97.4, 93.9, 93.4, 93.3, 94.1, 94.6, 94.4, 96.9, 98.3, 98.7, 99.6,
...     99.4, 99.2, 99.2, 99.6, 99.8, 99.8, 99.0, 97.6, 97.4, 96.3, 95.2,
...     96.3, 95.5, 91.6, 90.0, 85.7, 82.3, 82.5, 82.2, 76.5, 81.7, 83.4,
...     82.8, 85.8, 87.6, 85.6, 85.3, 86.2, 89.3, 91.4, 90.9, 89.2, 85.6,
...     86.3, 91.2, 87.5, 84.5, 76.7, 74.8, 69.7, 66.9, 64.4, 65.4, 67.6,
...     70.2, 70.4, 69.5, 71.6, 74.5, 80.2)
>>> inputs.atmosphericpressure.series = (
...     101.7, 101.8, 101.8, 101.9, 102.0, 102.0, 102.1, 102.1, 102.2, 102.2,
...     102.2, 102.2, 102.2, 102.1, 102.0, 102.0, 101.9, 101.8, 101.6, 101.5,
...     101.4, 101.3, 101.1, 101.0, 100.8, 100.7, 100.6, 100.5, 100.5, 100.4,
...     100.4, 100.4, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2, 100.2, 100.3,
...     100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2,
...     100.2, 100.3, 100.4, 100.5, 100.6, 100.8, 100.9, 101.0, 101.1, 101.2,
...     101.3, 101.4, 101.5, 101.5, 101.7, 101.7, 101.8, 101.9, 101.9, 102.0,
...     102.0, 102.0)

In this and the following "winter" examples, we use the initial conditions
of December 10th, calculated by the corresponding daily integration tests.
However, this is possible for the state sequences, but not for the (now
higher resolved) log sequences.  Hence, we need to define them manually:

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.0, 0.04, 0.25, 0.59, 0.91, 0.97, 1.0,
...       0.65, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.1, 0.2, 0.4, 0.6, 0.5, 0.2, 0.0,
...       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [-2.8, -2.6, -2.2, -1.8, -1.8, -2.4, -1.5, -1.4,
...       -0.8, -0.8, -0.9, 0.9, -0.8, -1.0, -1.2, -1.8,
...       -2.8, -3.8, -4.3, -4, -3.9, -1.6, -1.6, 2.6]),
...     (logs.loggedrelativehumidity,
...      [99.4, 99.0, 99.6, 99.6, 99.3, 99.3, 99.8, 99.7,
...       100.0, 99.6, 99.3, 98.9, 96.6, 93.0, 91.4, 92.0,
...       94.5, 96.5, 98.2, 96.9, 98.0, 95.0, 91.8, 88.5]),
...     (logs.loggedwindspeed2m,
...      [2.2, 2.0, 2.5, 2.6, 2.3, 1.7, 2.8, 1.9,
...       2.6, 2.9, 3.5, 3.4, 3.5, 3.0, 3.9, 3.6,
...       2.2, 2.4, 2.5, 1.5, 2.5, 3.0, 3.2, 3.1]))

Compared to the daily simulation :ref:`lland_v3_acker_winter_daily`, the
hourly simulation estimates a slightly more pronounced melting rate for
land-use type |ACKER|, resulting in a complete depletion of the snow-cover:

.. integration-test::

    >>> lnk(ACKER)
    >>> test("lland_v3_acker_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                      0.41098 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.338514 | -0.000646 |                   0.0 |                       0.0 |                   0.701061 |                  4.014396 |                 0.266716 |    -0.167266 |        -0.266716 |         -3.313335 |             62.648725 |                 100.0 |               112.102145 |               98.814229 | 0.234885 | 0.234885 |  0.001835 |  0.001835 |      0.0 |  0.000713 | -0.001816 |    -5.308595 |     0.755856 |      0.0 |      0.0 | 0.701337 | 0.701337 |  0.001902 | -0.036834 |  0.231784 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.058165 | 20.378044 |  20.39327 |       0.0 | 0.843656 | -1.570008 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                     0.396005 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.233359 |  0.012102 |                   0.0 |                       0.0 |                   0.702998 |                  4.024777 |                 0.276594 |    -0.167699 |        -0.276594 |         -3.321779 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 | -0.004543 | -0.004543 |      0.0 |  0.001157 |       0.0 |    -5.793625 |     0.755182 |      0.0 |      0.0 | 0.739882 | 0.015226 |  0.003087 | -0.020459 |  0.259222 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062708 | 20.392113 | 20.392113 | -0.242035 | 0.850964 |  -1.55121 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                     0.322887 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.170698 |  0.070286 |                   0.0 |                       0.0 |                   0.703629 |                  4.040774 |                 0.252682 |    -0.168366 |        -0.252682 |         -3.337145 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.002708 | 0.002708 | -0.004605 | -0.004605 |      0.0 | -0.004655 | -0.000032 |    -8.424686 |     0.754962 |      0.0 |      0.0 | 1.075161 |      0.0 | -0.012416 | -0.052943 |  0.187323 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.064605 | 20.399475 | 20.399475 | -0.359104 | 0.853356 | -1.590596 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                     0.292362 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.301986 |  0.094629 |                   0.0 |                       0.0 |                   0.704086 |                  4.052814 |                  0.24323 |    -0.168867 |         -0.24323 |         -3.348728 |             71.001889 |                 100.0 |               112.102145 |               98.814229 | 0.004605 | 0.004605 | -0.004915 | -0.004915 |      0.0 | -0.008067 |  -0.00006 |    -9.682955 |     0.754803 |      0.0 |      0.0 | 1.236302 |      0.0 | -0.021518 | -0.073321 |  0.148391 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.064915 | 20.412147 | 20.412147 | -0.412925 | 0.855092 | -1.654325 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                     0.268467 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.514415 |  0.113135 |                   0.0 |                       0.0 |                   0.704487 |                  4.064535 |                 0.244637 |    -0.169356 |        -0.244637 |         -3.360048 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.004915 | 0.004915 | -0.008645 | -0.008645 |      0.0 | -0.007867 | -0.000075 |   -10.752156 |     0.754663 |      0.0 |      0.0 |  1.37369 |      0.0 | -0.020986 | -0.064703 |  0.158948 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.068645 | 20.424929 | 20.424929 | -0.458812 | 0.856622 |  -1.73656 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                     0.247996 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.788533 |    0.1285 |                   0.0 |                       0.0 |                   0.704768 |                  4.080005 |                 0.249591 |        -0.17 |        -0.249591 |         -3.375237 |             92.611159 |                 100.0 |               112.102145 |               98.814229 | 0.008645 | 0.008645 | -0.007726 | -0.007726 |      0.0 | -0.006572 |  -0.00015 |   -11.737605 |     0.754565 |      0.0 |      0.0 | 1.500803 |      0.0 | -0.017529 | -0.061256 |  0.170806 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.067726 | 20.440145 | 20.440145 | -0.501268 | 0.857693 |  -1.83416 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                     0.236558 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -6.113866 |  0.134056 |                   0.0 |                       0.0 |                   0.705038 |                  4.093695 |                 0.243066 |    -0.170571 |        -0.243066 |         -3.388658 |             101.43127 |                 100.0 |               112.102145 |               98.814229 | 0.007726 | 0.007726 | -0.010331 | -0.010331 |      0.0 | -0.007889 | -0.000139 |   -12.320179 |     0.754471 |      0.0 |      0.0 | 1.576395 |      0.0 | -0.021044 | -0.062856 |  0.159165 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.070331 | 20.455761 | 20.455761 | -0.526516 | 0.858722 | -1.937316 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                     0.229932 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -6.45772 |  0.134168 |                   0.0 |                       0.0 |                   0.648768 |                  3.682358 |                 0.225082 |    -0.153432 |        -0.225082 |          -3.03359 |            152.146905 |                 100.0 |               112.102145 |               98.814229 | 0.010331 | 0.010331 | -0.008377 | -0.008377 |      0.0 | -0.007022 | -0.000184 |   -12.669195 |     0.754404 |      0.0 |      0.0 | 1.622496 |      0.0 | -0.018732 | -0.056968 |  0.149382 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.068377 | 20.473114 | 20.473114 | -0.541914 | 0.859461 | -2.040584 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                     0.221942 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 | -6.801947 |  0.136086 |               0.00098 |                   0.00098 |                   0.562082 |                  3.039527 |                 0.214164 |    -0.125667 |        -0.213184 |         -2.477445 |            193.641515 |                 100.0 |               112.102145 |               99.616909 | 0.008377 | 0.008377 | -0.005976 | -0.005976 |      0.0 | -0.005282 | -0.000166 |   -13.102204 |     0.754325 |      0.0 |      0.0 | 1.679213 |      0.0 |  -0.01409 | -0.044231 |  0.154863 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.065976 | 20.486773 | 20.486773 | -0.560857 | 0.860328 |  -2.14577 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                      0.22582 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 | -7.152565 |  0.123936 |              0.016513 |                  0.016513 |                   0.477976 |                  2.412046 |                 0.187786 |    -0.083989 |        -0.171273 |          -1.93407 |            213.005666 |                 100.0 |               112.102145 |              101.107639 | 0.005976 | 0.005976 | -0.001967 | -0.001967 |      0.0 | -0.004727 | -0.000115 |   -12.890333 |     0.754232 |      0.0 |      0.0 | 1.652967 |      0.0 | -0.012609 | -0.043609 |  0.115055 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061967 | 20.497476 | 20.497476 | -0.552091 | 0.861354 | -2.238805 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                      0.26043 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -7.462684 |  0.079231 |               0.04981 |                   0.04981 |                   0.399077 |                  1.815487 |                 0.133551 |    -0.025836 |        -0.083742 |          -1.41641 |            152.146905 |                 100.0 |               112.102145 |              101.107639 | 0.001967 | 0.001967 |  0.002293 |  0.002293 |      0.0 | -0.008223 | -0.000028 |   -11.130785 |     0.754105 |      0.0 |      0.0 | 1.427801 |      0.0 | -0.021935 |  -0.05781 |  0.003997 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.057707 | 20.507667 | 20.507667 | -0.476886 | 0.862747 | -2.287136 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                     0.321088 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -7.623788 |  0.018837 |              0.094929 |                  0.094929 |                   0.356693 |                  1.493002 |                   0.0885 |     0.032721 |          0.00643 |         -1.136309 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.008207 |  0.008207 |      0.0 | -0.010542 |       0.0 |    -8.495873 |     0.753945 |      0.0 |      0.0 | 1.090245 |      0.0 | -0.028121 | -0.059356 | -0.093907 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.049499 | 20.518209 | 20.518209 | -0.364142 | 0.864511 | -2.275073 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                     0.362884 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -7.583578 | -0.014184 |               0.08109 |                   0.08109 |                   0.335197 |                  1.331032 |                 0.078354 |      0.02563 |         0.002736 |         -0.995835 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.010188 |  0.010188 |      0.0 | -0.009019 |       0.0 |    -6.926904 |     0.753753 |      0.0 |      0.0 | 0.889362 |      0.0 | -0.024058 | -0.054485 | -0.081279 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.039311 | 20.527228 | 20.527228 | -0.297047 |  0.86663 | -2.229989 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                     0.400478 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -7.433297 | -0.038583 |              0.071403 |                  0.071403 |                   0.332002 |                  1.306141 |                 0.063562 |      0.01698 |         0.007841 |         -0.974139 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.010647 |  0.010647 |      0.0 | -0.010304 |       0.0 |    -5.647063 |     0.753539 |      0.0 |      0.0 | 0.725359 |      0.0 | -0.027487 | -0.058032 |  -0.09336 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.028664 | 20.537532 | 20.537532 |  -0.24227 | 0.869006 | -2.160507 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                     0.431939 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -7.201688 | -0.055001 |              0.038296 |                  0.038296 |                   0.332321 |                  1.307301 |                 0.054785 |    -0.016175 |        -0.016489 |          -0.97498 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |   0.01032 |   0.01032 |      0.0 | -0.014286 |       0.0 |    -4.655362 |     0.753302 |      0.0 |      0.0 | 0.598276 |      0.0 | -0.038107 | -0.075828 | -0.097446 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.018345 | 20.551818 | 20.551818 | -0.199824 | 0.871636 | -2.074606 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                     0.459304 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -6.915353 | -0.066352 |              0.009175 |                  0.009175 |                   0.332669 |                  1.307629 |                 0.043958 |     -0.04531 |        -0.034783 |         -0.974959 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.007065 |  0.007065 |      0.0 |  -0.01816 |       0.0 |    -3.843504 |     0.753043 |      0.0 |      0.0 | 0.494285 |      0.0 | -0.048441 | -0.087426 | -0.101085 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.011279 | 20.569978 | 20.569978 | -0.165091 | 0.874517 | -1.977354 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                     0.482766 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 |  -6.59118 | -0.073664 |              0.000025 |                  0.000025 |                    0.33293 |                   1.30686 |                 0.041117 |    -0.054428 |        -0.041092 |          -0.97393 |             68.711505 |                 100.0 |               112.102145 |               98.953162 | 0.045279 |  0.01132 |  0.005889 |  0.005889 |      0.0 | -0.019668 |  0.000166 |    -3.180823 |      0.75285 |      0.0 |      0.0 | 0.411003 |  0.03396 | -0.052463 | -0.089943 | -0.101314 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.054111 | 20.634926 | 20.634926 | -0.125933 | 0.876679 |  -1.87279 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                     0.502679 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -6.242634 | -0.077809 |                   0.0 |                       0.0 |                   0.333358 |                  1.305996 |                 0.047594 |    -0.054417 |        -0.047594 |         -0.972638 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 |  0.007388 |  0.007388 |      0.0 | -0.018778 |       0.0 |    -2.640366 |     0.752532 |      0.0 |      0.0 | 0.340931 |      0.0 | -0.050089 | -0.087375 | -0.089871 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.046723 | 20.653704 | 20.653704 | -0.113871 | 0.880239 | -1.764081 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                     0.518258 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.880271 | -0.078834 |                   0.0 |                       0.0 |                   0.333101 |                  1.303739 |                 0.052878 |    -0.054322 |        -0.052878 |         -0.970637 |             48.410379 |                 100.0 |               112.102145 |               98.814229 | 0.268723 | 0.067181 |  0.003845 |  0.003845 |      0.0 | -0.021454 |  0.000984 |    -2.230559 |     0.752723 |      0.0 |      0.0 | 0.294851 | 0.201542 | -0.057228 | -0.088891 |  -0.09324 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.056155 | 20.943881 | 20.943881 | -0.031165 |   0.8781 | -1.654347 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                      0.55373 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.514492 |  -0.09024 |                   0.0 |                       0.0 |                    0.33351 |                  1.301296 |                 0.055039 |    -0.054221 |        -0.055039 |         -0.967785 |             56.054123 |                 100.0 |               112.102145 |               98.814229 | 0.184155 | 0.027623 |  0.000004 |  0.000004 |      0.0 | -0.018093 |  0.000998 |    -1.336706 |     0.752419 |      0.0 |      0.0 | 0.178037 | 0.156532 | -0.048261 |  -0.06772 | -0.060943 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.059996 | 21.146129 | 21.146129 | -0.007183 | 0.881505 | -1.533207 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                     0.582171 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.110691 | -0.096225 |                   0.0 |                       0.0 |                   0.333956 |                  1.298642 |                  0.06046 |     -0.05411 |         -0.06046 |         -0.964687 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 0.375996 |   0.0376 |  0.000346 |  0.000346 |      0.0 | -0.020956 |  0.002393 |    -0.655845 |     0.752088 |      0.0 |      0.0 | 0.089719 | 0.089719 | -0.055898 | -0.075611 | -0.071049 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.059654 | 21.294161 | 21.543081 |       0.0 | 0.885229 | -1.406083 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                      0.59998 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -4.686942 |  -0.09596 |                   0.0 |                       0.0 |                   0.334828 |                  1.295068 |                 0.055481 |    -0.053961 |        -0.055481 |          -0.96024 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 0.657654 |      0.0 | -0.002532 | -0.002532 |      0.0 | -0.022472 |  0.005507 |    -0.244342 |     0.751441 |      0.0 |      0.0 | 0.035335 | 0.035335 | -0.059943 | -0.074189 | -0.078651 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062532 | 21.351086 | 22.223207 |       0.0 | 0.892565 | -1.279222 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -4.264075 | -0.092104 |                   0.0 |                       0.0 |                   0.335693 |                  1.290674 |                 0.049365 |    -0.053778 |        -0.049365 |          -0.95498 |             36.725115 |                 100.0 |               112.102145 |               98.814229 | 0.472532 |      0.0 | -0.001838 | -0.001838 |      0.0 | -0.024931 |  0.004748 |          0.0 |     0.750799 | 0.033933 | 0.033933 |      0.0 |      0.0 | -0.066503 | -0.081551 |  -0.09869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061838 | 21.340569 |  22.72067 |       0.0 | 0.899896 | -1.156218 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -3.854062 | -0.083248 |                   0.0 |                       0.0 |                   0.336554 |                   1.28612 |                 0.047044 |    -0.053588 |        -0.047044 |         -0.949566 |             34.355753 |                 100.0 |               112.102145 |               98.814229 | 0.941838 |      0.0 | -0.000987 | -0.000987 |      0.0 | -0.027164 |  0.009858 |          0.0 |     0.750159 | 0.124524 | 0.124524 |      0.0 |      0.0 | -0.072459 | -0.089566 | -0.114981 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.060987 | 21.240401 | 23.689672 |       0.0 | 0.907248 | -1.042071 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -3.473569 | -0.075029 |                   0.0 |                       0.0 |                   0.335096 |                  1.277745 |                 0.041776 |    -0.053239 |        -0.041776 |         -0.942648 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 1.880987 |      0.0 | -0.001175 | -0.001175 |      0.0 | -0.027338 |  0.021263 |          0.0 |     0.749522 | 0.199503 | 0.199503 |      0.0 |      0.0 | -0.072924 | -0.089252 |   -0.1204 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061175 | 21.063393 | 25.597998 |       0.0 | 0.914638 | -0.936142 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -3.120472 | -0.067402 |                   0.0 |                       0.0 |                   0.335944 |                  1.268373 |                 0.027852 |    -0.052849 |        -0.027852 |          -0.93243 |             40.962628 |                 100.0 |               112.102145 |               98.814229 | 1.975175 |      0.0 | -0.002517 | -0.002517 |      0.0 | -0.031933 |  0.026463 |          0.0 |     0.748889 | 0.348078 | 0.348078 |      0.0 |      0.0 |  -0.08518 |  -0.09987 | -0.157197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062517 | 20.739306 | 27.605106 |       0.0 | 0.922037 | -0.837839 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.792798 | -0.060324 |                   0.0 |                       0.0 |                   0.336784 |                  1.257174 |                 0.008709 |    -0.052382 |        -0.008709 |          -0.92039 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 1.694517 |      0.0 | -0.003094 | -0.003094 |      0.0 | -0.043255 |  0.027669 |          0.0 |     0.748261 | 0.612982 | 0.612982 |      0.0 |      0.0 | -0.115381 |  -0.13072 | -0.237391 |      1.0 |      0.3 | 0.519747 | 0.177426 |  0.0 |  0.0 | 0.0 |  0.0 | 0.177426 | 0.002302 | 0.000639 | 0.063094 | 20.156036 | 28.823132 |       0.0 |  0.92943 | -0.746615 | 71.788581 | 0.177426 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000639 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.488717 | -0.053756 |                   0.0 |                       0.0 |                   0.337608 |                  1.245134 |                -0.004886 |    -0.051881 |         0.004886 |         -0.907526 |             32.273586 |                 100.0 |               111.611781 |               98.814229 | 1.225094 |      0.0 | -0.002985 | -0.002985 |      0.0 | -0.055801 |       0.0 |          0.0 |     0.747645 | 0.795628 | 0.795628 |      0.0 |      0.0 | -0.148847 | -0.165762 | -0.319496 |      1.0 |      0.3 | 2.365045 | 0.809191 |  0.0 |  0.0 | 0.0 |  0.0 | 0.809191 | 0.026643 | 0.007401 | 0.062985 |  19.39789 | 27.738983 |       0.0 | 0.936738 | -0.661959 | 73.344435 |      0.5 | 0.309191 |   0.0 |   0.0 |  0.0 | 0.011009 | 0.015633 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007401 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.206529 | -0.047661 |                   0.0 |                       0.0 |                   0.338426 |                  1.232336 |                -0.014528 |    -0.051347 |         0.014528 |          -0.89391 |             30.429381 |                 100.0 |               109.395237 |               98.814229 | 0.472985 |      0.0 |  0.000597 |  0.000597 |      0.0 |  -0.06294 |       0.0 |          0.0 |     0.747033 | 0.971418 | 0.971418 |      0.0 |      0.0 |  -0.16789 | -0.189697 | -0.372115 |      1.0 |      0.3 | 1.865265 | 0.639779 |  0.0 |  0.0 | 0.0 |  0.0 | 0.639779 | 0.059884 | 0.016635 | 0.059403 | 18.468282 | 26.409643 |       0.0 | 0.944046 | -0.583398 |  74.56992 |      0.5 | 0.139779 |   0.0 |   0.0 |  0.0 | 0.023641 | 0.036244 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016635 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.944659 | -0.042005 |                   0.0 |                       0.0 |                   0.339238 |                  1.218349 |                -0.012189 |    -0.050765 |         0.012189 |         -0.879111 |             29.178858 |                 100.0 |               107.662715 |               98.814229 | 0.187403 |      0.0 |  0.007215 |  0.007215 |      0.0 | -0.061894 |       0.0 |          0.0 |     0.746426 |   0.9929 |   0.9929 |      0.0 |      0.0 | -0.165098 | -0.196345 | -0.373633 |      1.0 |      0.3 | 1.610577 | 0.553574 |  0.0 |  0.0 | 0.0 |  0.0 | 0.553574 | 0.078121 |   0.0217 | 0.052785 | 17.516338 | 25.048363 |       0.0 | 0.951354 | -0.510493 | 75.626924 |      0.5 | 0.053574 |   0.0 |   0.0 |  0.0 | 0.035945 | 0.042175 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0217 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.701644 | -0.036755 |                   0.0 |                       0.0 |                   0.340043 |                  1.204312 |                -0.020102 |     -0.05018 |         0.020102 |         -0.864269 |              29.58412 |                 100.0 |                106.17729 |               98.814229 | 0.462785 |      0.0 |  0.008179 |  0.008179 |      0.0 |  -0.06548 |       0.0 |          0.0 |     0.745824 |  1.09064 |  1.09064 |      0.0 |      0.0 | -0.174665 | -0.206262 | -0.401029 |      1.0 |      0.3 | 2.026477 |  0.69817 |  0.0 |  0.0 | 0.0 |  0.0 |  0.69817 | 0.098551 | 0.027375 | 0.051821 | 16.468637 | 23.550151 |       0.0 | 0.958662 | -0.442838 | 76.955231 |      0.5 |  0.19817 |   0.0 |   0.0 |  0.0 | 0.047932 | 0.050619 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027375 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.476125 | -0.031884 |                   0.0 |                       0.0 |                   0.340843 |                  1.190153 |                -0.018227 |     -0.04959 |         0.018227 |         -0.849309 |             28.027061 |                 100.0 |               104.321666 |               98.814229 | 0.179821 |      0.0 |  0.014145 |  0.014145 |      0.0 | -0.065772 |       0.0 |          0.0 |     0.745226 | 1.130145 | 1.130145 |      0.0 |      0.0 | -0.175444 | -0.215681 | -0.409353 |      1.0 |      0.3 | 1.800442 | 0.621729 |  0.0 |  0.0 | 0.0 |  0.0 | 0.621729 | 0.121029 | 0.033619 | 0.045855 |  15.38133 | 21.995303 |       0.0 |  0.96597 | -0.380053 | 78.133943 |      0.5 | 0.121729 |   0.0 |   0.0 |  0.0 |  0.05961 | 0.061419 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033619 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 | -1.266844 | -0.027364 |              0.000828 |                  0.000828 |                   0.341637 |                  1.174779 |                -0.024601 |    -0.048121 |         0.025429 |         -0.833142 |             29.178858 |                 100.0 |               102.684805 |               99.313275 | 0.079855 |      0.0 |  0.019007 |  0.019007 |      0.0 | -0.065786 |       0.0 |          0.0 |     0.744633 | 1.180942 | 1.180942 |      0.0 |      0.0 | -0.175481 | -0.220889 | -0.421799 |      1.0 |      0.3 | 1.773653 | 0.613832 |  0.0 |  0.0 | 0.0 |  0.0 | 0.613832 | 0.137999 | 0.038333 | 0.040993 |  14.24286 |  20.36729 |       0.0 | 0.973278 | -0.321789 | 79.293765 |      0.5 | 0.113832 |   0.0 |   0.0 |  0.0 | 0.070985 | 0.067014 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038333 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 | -1.072631 | -0.023169 |              0.015911 |                  0.015911 |                   0.342425 |                   1.15845 |                -0.023742 |    -0.032358 |         0.039653 |         -0.816025 |             30.000798 |                 100.0 |               101.082683 |              100.108808 | 0.262993 |      0.0 |  0.016361 |  0.016361 |      0.0 | -0.065321 |       0.0 |          0.0 |     0.744044 | 1.205479 | 1.205479 |      0.0 |      0.0 | -0.174242 | -0.211904 | -0.425799 |      1.0 |      0.3 | 1.992357 | 0.691173 |  0.0 |  0.0 | 0.0 |  0.0 | 0.691173 | 0.157644 |  0.04379 | 0.043639 | 13.079194 | 18.703248 |       0.0 | 0.980586 | -0.267721 | 80.594949 |      0.5 | 0.191173 |   0.0 |   0.0 |  0.0 | 0.082067 | 0.075577 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04379 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 | -0.892402 | -0.019276 |              0.048595 |                  0.048595 |                   0.343207 |                  1.144953 |                 -0.02789 |     0.000888 |         0.076485 |         -0.801746 |             30.000798 |                 100.0 |                99.294813 |               99.927984 | 0.077639 |      0.0 |  0.026372 |  0.026372 |      0.0 | -0.066592 |       0.0 |          0.0 |     0.743459 | 1.361497 | 1.361497 |      0.0 |      0.0 | -0.177631 |   -0.2199 | -0.474016 |      1.0 |      0.3 | 2.031512 | 0.706524 |  0.0 |  0.0 | 0.0 |  0.0 | 0.706524 | 0.180705 | 0.050196 | 0.033628 | 11.759417 | 16.815966 |       0.0 | 0.987894 | -0.217545 | 81.919937 |      0.5 | 0.206524 |   0.0 |   0.0 |  0.0 | 0.092863 | 0.087843 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050196 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 | -0.725149 | -0.015663 |              0.074176 |                  0.074176 |                   0.334932 |                  1.076886 |                -0.043295 |     0.029306 |         0.117471 |         -0.741954 |             23.933221 |                 100.0 |                97.484039 |               99.738784 | 0.067628 |      0.0 |  0.059957 |  0.059957 |      0.0 | -0.080429 |       0.0 |          0.0 |     0.742879 | 1.829142 | 1.829142 |      0.0 |      0.0 |  -0.21454 | -0.294586 | -0.626597 |      1.0 |      0.3 | 2.695811 | 0.940477 |  0.0 |  0.0 | 0.0 |  0.0 | 0.940477 | 0.214846 |  0.05968 | 0.000043 |   9.97777 | 14.268212 |       0.0 | 0.995202 | -0.170981 | 83.675271 |      0.5 | 0.440477 |   0.0 |   0.0 |  0.0 | 0.103379 | 0.111467 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05968 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 | -0.569938 | -0.012311 |              0.074473 |                  0.074473 |                   0.335683 |                  1.067855 |                -0.048721 |     0.029979 |         0.123194 |         -0.732172 |             22.421649 |                 100.0 |                95.099467 |               99.479805 |      0.0 |      0.0 |  0.073828 |  0.000043 |      0.0 | -0.085956 |       0.0 |          0.0 |     0.742303 | 1.996921 | 1.996921 |      0.0 |      0.0 | -0.229285 | -0.326804 | -0.679282 |      1.0 |      0.3 |   2.8728 | 1.005742 |  0.0 |  0.0 | 0.0 |  0.0 | 1.005742 | 0.261091 | 0.072525 |      0.0 |  8.028929 | 11.481368 |       0.0 |  1.00251 | -0.127771 | 85.542329 |      0.5 | 0.505742 |   0.0 |   0.0 |  0.0 | 0.113624 | 0.147467 |   0.0 |   0.0 |  0.0 |   0.0 | 0.072525 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 | -0.425903 | -0.009199 |              0.070509 |                  0.070509 |                   0.336428 |                   1.06128 |                -0.040462 |     0.026289 |         0.110972 |         -0.724852 |             26.962743 |                 100.0 |                92.579867 |               99.193243 |    0.034 |      0.0 |  0.085284 |      0.06 |      0.0 | -0.061599 |       0.0 |          0.0 |     0.741731 | 1.633556 | 1.633556 |      0.0 |      0.0 | -0.164312 | -0.279524 | -0.554807 |      1.0 |      0.3 | 2.382518 | 0.836717 |  0.0 |  0.0 | 0.0 |  0.0 | 0.836717 | 0.298134 | 0.082815 |      0.0 |  6.429685 |  9.194449 |       0.0 | 1.009818 | -0.087671 |  87.08813 |      0.5 | 0.336717 |   0.0 |   0.0 |  0.0 | 0.123605 | 0.174529 |   0.0 |   0.0 |  0.0 |   0.0 | 0.082815 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 | -0.292238 | -0.006312 |              0.039349 |                  0.039349 |                   0.337168 |                  1.056546 |                -0.036448 |    -0.004674 |         0.075797 |         -0.719378 |              22.18809 |                 100.0 |                90.505993 |               98.946731 |      0.0 |      0.0 |  0.116226 |       0.0 |      0.0 | -0.063414 |       0.0 |          0.0 |     0.741163 | 1.717216 | 1.717216 |      0.0 |      0.0 | -0.169154 | -0.334911 | -0.579862 |      1.0 |      0.3 | 2.472555 | 0.870985 |  0.0 |  0.0 | 0.0 |  0.0 | 0.870985 | 0.325709 | 0.090475 |      0.0 |  4.744971 |  6.785309 |       0.0 | 1.017126 | -0.050459 |   88.6897 |      0.5 | 0.370985 |   0.0 |   0.0 |  0.0 | 0.133327 | 0.192381 |   0.0 |   0.0 |  0.0 |   0.0 | 0.090475 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 | -0.168197 | -0.003633 |              0.011779 |                  0.011779 |                   0.347034 |                  1.110113 |                -0.025663 |    -0.034475 |         0.037442 |          -0.76308 |             23.407216 |                 100.0 |                88.368197 |               98.681882 |    0.034 |      0.0 |  0.102939 |      0.06 |      0.0 | -0.057639 |       0.0 |          0.0 |     0.740599 | 1.489884 | 1.489884 |      0.0 |      0.0 |  -0.15375 | -0.310062 | -0.501254 |      1.0 |      0.3 | 2.182633 | 0.771028 |  0.0 |  0.0 | 0.0 |  0.0 | 0.771028 | 0.347877 | 0.096633 |      0.0 |  3.282738 |  4.694315 |       0.0 | 1.024434 | -0.015926 | 90.101304 |      0.5 | 0.271028 |   0.0 |   0.0 |  0.0 | 0.142799 | 0.205079 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096633 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 | -0.053086 | -0.001147 |              0.000021 |                  0.000021 |                   0.347782 |                  1.107879 |                -0.014096 |    -0.046141 |         0.014116 |         -0.760097 |              24.48341 |                 100.0 |                86.492644 |               98.794057 |      0.0 |      0.0 |  0.096543 |       0.0 |      0.0 | -0.047871 |       0.0 |          0.0 |     0.740039 |  1.25791 |  1.25791 |      0.0 |      0.0 | -0.127693 | -0.279479 | -0.421289 | 0.999397 | 0.299638 | 1.817155 | 0.642853 |  0.0 |  0.0 | 0.0 |  0.0 | 0.642853 | 0.357178 | 0.099216 |      0.0 |  2.045476 |  2.925031 |       0.0 | 1.031742 |  0.016121 | 91.275606 |      0.5 | 0.142853 |   0.0 |   0.0 |  0.0 | 0.152026 | 0.205153 |   0.0 |   0.0 |  0.0 |   0.0 | 0.099216 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.348526 |                   1.10789 |                 0.008426 |    -0.046162 |        -0.008426 |         -0.759364 |             23.933221 |                 100.0 |                84.938259 |               98.814229 |      0.0 |      0.0 |  0.129358 |       0.0 |      0.0 | -0.025712 |       0.0 |          0.0 |     0.739484 | 0.975359 | 0.975359 |      0.0 |      0.0 | -0.068585 |  -0.26561 |  -0.32577 |  0.99824 | 0.298945 | 1.407023 | 0.497811 |  0.0 |  0.0 | 0.0 |  0.0 | 0.497811 | 0.352724 | 0.097979 |      0.0 |  1.079524 |   1.54372 |       0.0 |  1.03905 |  0.047021 | 92.184819 | 0.497811 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160986 | 0.191738 |   0.0 |   0.0 |  0.0 |   0.0 | 0.097979 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                      0.61078 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.349264 |                  1.107494 |                 0.008574 |    -0.046146 |        -0.008574 |          -0.75823 |             26.962743 |                 100.0 |                83.738273 |               98.814229 |      0.0 |      0.0 |  0.087608 |       0.0 |      0.0 | -0.030943 |       0.0 |          0.0 |     0.738932 | 0.888357 | 0.888357 |      0.0 |      0.0 |  -0.08254 | -0.222746 | -0.296711 | 0.997084 | 0.298253 | 1.295814 | 0.458403 |  0.0 |  0.0 | 0.0 |  0.0 | 0.458403 | 0.341856 |  0.09496 |      0.0 |     0.195 |   0.27885 |       0.0 | 1.046358 |  0.077921 |  93.02223 | 0.458403 |      0.0 |   0.0 |   0.0 |  0.0 | 0.169175 | 0.172681 |   0.0 |   0.0 |  0.0 |   0.0 |  0.09496 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                      0.61078 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.349997 |                  1.107885 |                  0.01332 |    -0.046162 |         -0.01332 |         -0.757888 |             30.000798 |                 100.0 |                82.635663 |               98.814229 |      0.0 |      0.0 |  0.070419 |       0.0 | 0.025143 | -0.028252 |       0.0 |          0.0 |     0.738384 | 0.760344 |    0.195 |      0.0 |      0.0 | -0.075361 | -0.191913 | -0.253955 | 0.995927 | 0.297561 | 0.307102 | 0.108525 |  0.0 |  0.0 | 0.0 |  0.0 | 0.108525 | 0.327626 | 0.091007 |      0.0 |       0.0 |       0.0 |       0.0 | 1.053666 |  0.108821 | 93.195664 | 0.108525 |      0.0 |   0.0 |   0.0 |  0.0 | 0.172108 | 0.155519 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091007 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.108737 |                      0.0 |    -0.046197 |              0.0 |          0.095308 |             33.282135 |                 100.0 |                 82.40761 |               98.814229 |      0.0 |      0.0 |  0.066191 |       0.0 | 0.025199 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |  0.99081 | 0.294511 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.309119 | 0.085866 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.245561 | 93.170465 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.169057 | 0.140062 |   0.0 |   0.0 |  0.0 |   0.0 | 0.085866 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.109465 |                      0.0 |    -0.046228 |              0.0 |          0.094579 |              29.58412 |                 100.0 |                82.440739 |               98.814229 |      0.0 |      0.0 |  0.060894 |       0.0 | 0.021594 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.985692 | 0.291477 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.290832 | 0.080787 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.382301 | 93.148871 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.16469 | 0.126141 |   0.0 |   0.0 |  0.0 |   0.0 | 0.080787 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10152 |                   0.0 |                       0.0 |                   1.204044 |                   1.11033 |                      0.0 |    -0.046264 |              0.0 |          0.093714 |             30.429381 |                 100.0 |                82.469129 |               98.814229 |    0.034 |      0.0 |  0.050931 |  0.050931 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.980736 | 0.288553 |    0.034 | 0.011733 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011733 | 0.274193 | 0.076165 | 0.009069 |       0.0 |       0.0 |       0.0 |      nan |  0.514721 | 93.171138 | 0.011733 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160589 | 0.113604 |   0.0 |   0.0 |  0.0 |   0.0 | 0.076165 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08208 |                   0.0 |                       0.0 |                   1.204044 |                  1.112997 |                      0.0 |    -0.046375 |              0.0 |          0.091047 |             31.324363 |                 100.0 |                82.439853 |               98.814229 | 0.043069 |      0.0 |  0.055782 |  0.055782 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.976508 |  0.28607 | 0.043069 | 0.014765 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014765 | 0.259096 | 0.071971 | 0.004218 |       0.0 |       0.0 |       0.0 |      nan |  0.627701 | 93.199442 | 0.014765 |      0.0 |   0.0 |   0.0 |  0.0 | 0.156783 | 0.102313 |   0.0 |   0.0 |  0.0 |   0.0 | 0.071971 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08424 |                   0.0 |                       0.0 |                   1.196665 |                  1.116023 |                      0.0 |    -0.046501 |              0.0 |          0.080642 |             33.282135 |                 100.0 |                82.402644 |               98.814229 |      0.0 |      0.0 |  0.053849 |  0.004218 | 0.018581 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.972199 | 0.283551 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.245067 | 0.068074 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.742841 |  93.18086 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.152923 | 0.092145 |   0.0 |   0.0 |  0.0 |   0.0 | 0.068074 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07992 |                   0.0 |                       0.0 |                   1.196665 |                  1.119656 |                      0.0 |    -0.046652 |              0.0 |          0.077009 |             33.810423 |                 100.0 |                82.427072 |               98.814229 |      0.0 |      0.0 |  0.049163 |       0.0 | 0.018518 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.968052 | 0.281137 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.231959 | 0.064433 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.853661 | 93.162342 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.148973 | 0.082987 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064433 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0756 |                   0.0 |                       0.0 |                   1.196665 |                  1.123475 |                      0.0 |    -0.046811 |              0.0 |           0.07319 |             37.369415 |                 100.0 |                82.451418 |               98.814229 |    0.034 |      0.0 |  0.033548 |  0.033548 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.964066 | 0.278827 |    0.034 | 0.011429 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011429 | 0.220011 | 0.061114 | 0.026452 |       0.0 |       0.0 |       0.0 |      nan |  0.960161 | 93.184913 | 0.011429 |      0.0 |   0.0 |   0.0 |  0.0 | 0.145273 | 0.074739 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061114 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.127474 |                      0.0 |    -0.046978 |              0.0 |           0.06919 |             40.962628 |                 100.0 |                82.421744 |               98.814229 |      0.0 |      0.0 |  0.023802 |  0.023802 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.960242 | 0.276619 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.208978 | 0.058049 |  0.00265 |       0.0 |       0.0 |       0.0 |      nan |  1.062341 | 93.184913 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.141667 |  0.06731 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058049 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.131856 |                      0.0 |    -0.047161 |              0.0 |          0.064808 |              46.30558 |                 100.0 |                82.421744 |               98.814229 |      0.0 |      0.0 |  0.022296 |   0.00265 | 0.008765 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.956418 |  0.27442 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.198629 | 0.055175 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.164521 | 93.176148 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.138008 | 0.060621 |   0.0 |   0.0 |  0.0 |   0.0 | 0.055175 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0648 |                   0.0 |                       0.0 |                   1.196665 |                  1.136687 |                      0.0 |    -0.047362 |              0.0 |          0.059978 |             43.470544 |                 100.0 |                82.433267 |               98.814229 |      0.0 |      0.0 |  0.028418 |       0.0 |  0.01221 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.952836 | 0.272369 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.189039 | 0.052511 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.260221 | 93.163938 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.134443 | 0.054596 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052511 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.04104 |                   0.0 |                       0.0 |                   1.196665 |                  1.144252 |                      0.0 |    -0.047677 |              0.0 |          0.052412 |             33.810423 |                 100.0 |                 82.44932 |               98.814229 |      0.0 |      0.0 |  0.047697 |       0.0 | 0.017452 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.950144 | 0.270832 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.18014 | 0.050039 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.332161 | 93.146486 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.13097 | 0.049169 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050039 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.02592 |                   0.0 |                       0.0 |                   1.196665 |                  1.152031 |                      0.0 |    -0.048001 |              0.0 |          0.044634 |             36.725115 |                 100.0 |                82.472265 |               98.814229 |      0.0 |      0.0 |  0.040381 |       0.0 | 0.015359 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.948017 | 0.269621 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.17187 | 0.047742 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.388981 | 93.131127 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.127587 | 0.044283 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047742 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 |  -0.01728 |              0.002897 |                  0.002897 |                   1.196665 |                  1.158654 |                      0.0 |     -0.04538 |              0.0 |          0.038011 |             34.918962 |                 100.0 |                 82.49246 |               98.498883 |      0.0 |      0.0 |   0.02629 |       0.0 | 0.009662 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.946214 | 0.268596 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164173 | 0.045604 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.437161 | 93.121465 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124292 | 0.039881 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045604 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 |  -0.01728 |              0.057389 |                  0.057389 |                   1.196665 |                  1.166643 |                      0.0 |     0.008779 |              0.0 |          0.030022 |             36.102655 |                 100.0 |                82.505164 |               97.893562 |      0.0 |      0.0 |  0.045635 |       0.0 | 0.017179 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944411 | 0.267574 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.156999 | 0.043611 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.485341 | 93.104286 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121081 | 0.035918 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043611 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                     0.555459 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |   0.02796 |              0.020003 |                  0.020003 |                   0.136052 |                  1.175846 |                 0.100918 |     -0.02899 |        -0.080915 |         -1.039794 |             33.282135 |                 100.0 |                82.527752 |               97.896789 |    0.034 |  0.01275 |  0.052426 |  0.052426 |      0.0 |  0.004903 |  0.000025 |    -1.294436 |     0.897677 |      0.0 |      0.0 | 0.000192 | 0.000192 |  0.013079 | -0.065945 |  0.028049 | 0.944301 | 0.267511 | 0.015768 | 0.005135 |  0.0 |  0.0 | 0.0 |  0.0 | 0.005135 | 0.150368 | 0.041769 | 0.007574 |  0.009462 |  0.013329 |       0.0 |  0.00743 |  1.488281 | 93.114918 | 0.005135 |      0.0 |   0.0 |   0.0 |  0.0 |  0.11802 | 0.032348 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041769 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                     0.535467 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 |  0.038672 |              0.030684 |                  0.030684 |                   0.139046 |                   1.18712 |                 0.111622 |    -0.018779 |        -0.080938 |         -1.048074 |             31.324363 |                 100.0 |                82.513772 |               97.894792 |      0.0 |      0.0 |  0.084311 |  0.007574 | 0.026267 |  0.013329 |       0.0 |    -1.790392 |     0.895425 |      0.0 |      0.0 | 0.000193 | 0.000193 |  0.038481 | -0.080682 |  0.038737 | 0.944592 | 0.267676 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.14417 | 0.040047 |      0.0 |       0.0 |       0.0 |       0.0 | 0.014738 |  1.480508 | 93.088651 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115037 | 0.029133 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040047 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |   0.00216 |              0.264596 |                  0.264596 |                   1.196665 |                  1.199167 |                      0.0 |      0.21463 |              0.0 |         -0.002502 |             30.429381 |                 100.0 |                82.548312 |               97.899724 |      0.0 |      0.0 |  0.127998 |       0.0 | 0.042738 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.943516 | 0.267067 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.138303 | 0.038418 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.509248 | 93.045913 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.112066 | 0.026237 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038418 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |   0.00864 |              0.249736 |                  0.249736 |                   1.196665 |                   1.21168 |                      0.0 |     0.199249 |              0.0 |         -0.015015 |             33.810423 |                 100.0 |                82.604516 |               97.907741 |      0.0 |      0.0 |  0.131541 |       0.0 | 0.046722 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942683 | 0.266595 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.132801 | 0.036889 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.531508 | 92.999191 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109171 |  0.02363 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036889 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |   0.01296 |              0.195223 |                  0.195223 |                   1.259762 |                  1.349977 |                      0.0 |     0.138974 |              0.0 |         -0.090215 |             29.178858 |                 100.0 |                82.665966 |               97.916495 |      0.0 |      0.0 |  0.146498 |       0.0 | 0.047219 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942012 | 0.266216 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.127632 | 0.035453 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.549448 | 92.951972 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106351 | 0.021281 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035453 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |   0.01728 |              0.048073 |                  0.048073 |                   1.291311 |                  1.428152 |                      0.0 |    -0.011433 |              0.0 |         -0.136842 |             28.400756 |                 100.0 |                82.728077 |               97.925332 |      0.0 |      0.0 |  0.135014 |       0.0 | 0.042609 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941502 | 0.265928 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.12277 | 0.034103 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.563068 | 92.909363 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103604 | 0.019166 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034103 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |   0.02808 |              0.000547 |                  0.000547 |                   1.354408 |                  1.571379 |                      0.0 |    -0.064927 |              0.0 |         -0.216971 |             30.429381 |                 100.0 |                82.784131 |               98.772999 |      0.0 |      0.0 |   0.11254 |       0.0 | 0.036657 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941396 | 0.265868 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.118189 |  0.03283 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.565888 | 92.872706 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.100928 | 0.017261 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03283 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.03456 |                   0.0 |                       0.0 |                   1.354408 |                  1.583999 |                      0.0 |       -0.066 |              0.0 |          -0.22959 |             35.500944 |                 100.0 |                82.832359 |               98.814229 |      0.0 |      0.0 |   0.08847 |       0.0 |  0.03163 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941533 | 0.265946 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.113866 |  0.03163 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.562228 | 92.841076 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098321 | 0.015546 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03163 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |    0.0432 |                   0.0 |                       0.0 |                   1.354408 |                    1.5975 |                      0.0 |    -0.066562 |              0.0 |         -0.243092 |             44.376181 |                 100.0 |                82.873978 |               98.814229 |      0.0 |      0.0 |  0.063209 |       0.0 | 0.025658 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941994 | 0.266206 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.109782 | 0.030495 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.549928 | 92.815418 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095781 | 0.014001 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030495 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04536 |                   0.0 |                       0.0 |                   1.354408 |                  1.611195 |                      0.0 |    -0.067133 |              0.0 |         -0.256786 |             50.715635 |                 100.0 |                82.907741 |               98.814229 |      0.0 |      0.0 |  0.054415 |       0.0 |  0.02375 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942535 | 0.266512 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.105916 | 0.029421 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.535468 | 92.791668 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093307 | 0.012609 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029421 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04752 |                   0.0 |                       0.0 |                   1.354408 |                  1.625096 |                      0.0 |    -0.067712 |              0.0 |         -0.270688 |             43.470544 |                 100.0 |                82.938996 |               98.814229 |      0.0 |      0.0 |  0.065452 |       0.0 |  0.02619 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.943157 | 0.266863 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.102253 | 0.028404 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.518848 | 92.765478 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.090897 | 0.011356 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028404 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.05616 |                   0.0 |                       0.0 |                   1.354408 |                  1.639665 |                      0.0 |    -0.068319 |              0.0 |         -0.285257 |             53.251417 |                 100.0 |                82.973464 |               98.814229 |      0.0 |      0.0 |  0.048111 |       0.0 | 0.021424 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944102 | 0.267399 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.098776 | 0.027438 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.493588 | 92.744053 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088549 | 0.010227 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027438 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.06696 |                   0.0 |                       0.0 |                   1.354408 |                  1.653965 |                      0.0 |    -0.068915 |              0.0 |         -0.299557 |             62.648725 |                 100.0 |                83.001662 |               98.814229 |      0.0 |      0.0 |  0.034966 |       0.0 | 0.016822 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.945452 | 0.268164 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095472 |  0.02652 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.457528 | 92.727231 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086262 | 0.009211 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02652 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.08424 |                   0.0 |                       0.0 |                   1.354408 |                  1.664915 |                      0.0 |    -0.069371 |              0.0 |         -0.310507 |             62.648725 |                 100.0 |                83.023804 |               98.814229 |      0.0 |      0.0 |  0.024921 |       0.0 |  0.01189 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.947448 | 0.269297 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.092329 | 0.025647 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.404188 |  92.71534 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084033 | 0.008295 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025647 |

.. _lland_v3_nadelw_winter_hourly:

conifers (winter)
-----------------

Through comparing the hourly simulation results for |ACKER|
(:ref:`lland_v3_acker_winter_hourly`) and |NADELW| (this example), you can
see the shading effect of the tree-canopies.  Initially, during the
nighttime, the decreased longwave radiation loss (|NetLongwaveRadiationSnow|)
under the canopy results in an earlier start of snow-melt.  But then, during
the daytime, the decreased shortwave radiation gain (|NetShortwaveRadiationSnow|)
results in a lower melting rate, so that the complete melting process finishes
even later for |NADELW| than for |ACKER|:

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v3_nadelw_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_nadelw_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |       wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |           1.206967 |          3.4 |                 0.464297 |                      0.526233 |                     0.529351 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.04202 |                   0.0 |                       0.0 |                   0.723841 |                  4.014396 |                 0.048267 |    -0.167266 |        -0.048267 |         -3.290555 |             27.647059 |                 100.0 |              3113.156712 |               69.444444 | 0.145708 | 0.145708 |  0.023255 |  0.023255 |      0.0 |  0.011494 | -0.001127 |    -1.945364 |     0.747923 |      0.0 |      0.0 |  0.18822 |  0.18822 |  0.030659 |  0.024832 |  0.103759 | 0.990633 | 0.294406 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.176745 | 11.726023 | 13.490212 |       0.0 | 0.933427 |  0.250282 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           1.029472 |          2.9 |                 0.427219 |                      0.521048 |                     0.521839 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.046179 |                   0.0 |                       0.0 |                   0.725602 |                  4.024777 |                 0.048365 |    -0.167699 |        -0.048365 |         -3.299174 |             32.413793 |                 100.0 |             20046.551039 |               69.444444 |      0.0 |      0.0 |   0.00778 |   0.00778 |      0.0 |  0.011639 |       0.0 |     -2.13791 |      0.74731 |      0.0 |      0.0 | 0.204149 | 0.204149 |  0.031047 |  0.034953 |  0.114365 | 0.991205 | 0.294746 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.168965 | 11.919879 | 13.478573 |       0.0 | 0.940735 |  0.235003 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           0.887476 |          2.5 |                 0.401924 |                      0.514631 |                     0.514853 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.050095 |                   0.0 |                       0.0 |                   0.727351 |                  4.040774 |                 0.049173 |    -0.168366 |        -0.049173 |         -3.313423 |                  37.6 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 |   0.00648 |   0.00648 |      0.0 |  0.012537 |       0.0 |    -2.319202 |     0.746701 |      0.0 |      0.0 |   0.2183 |   0.2183 |  0.033441 |  0.040393 |  0.123007 | 0.991923 | 0.295174 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.162485 | 12.126889 | 13.466036 |       0.0 | 0.948043 |  0.215808 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           1.064971 |          3.0 |                 0.383833 |                      0.510972 |                     0.505858 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.055206 |                   0.0 |                       0.0 |                   0.729086 |                  4.052814 |                 0.049248 |    -0.168867 |        -0.049248 |         -3.323727 |             31.333333 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 |  0.006163 |  0.006163 |      0.0 |  0.013674 |       0.0 |    -2.555835 |     0.746096 |      0.0 |      0.0 |  0.23685 |  0.23685 |  0.036474 |  0.048592 |  0.134314 | 0.992833 | 0.295715 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.156322 | 12.351185 | 13.452362 |       0.0 | 0.955351 |  0.191502 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           0.887476 |          2.5 |                 0.352531 |                      0.505764 |                     0.499391 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.058931 |                   0.0 |                       0.0 |                   0.730809 |                  4.064535 |                 0.049397 |    -0.169356 |        -0.049397 |         -3.333726 |                  37.6 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.003516 | -0.003516 |      0.0 |  0.013591 |       0.0 |    -2.728282 |     0.745497 |      0.0 |      0.0 | 0.248522 | 0.248522 |  0.036253 |  0.056286 |  0.141937 | 0.993882 |  0.29634 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.159838 | 12.586977 | 13.438771 |       0.0 | 0.962659 |  0.163472 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |           0.816478 |          2.3 |                 0.326061 |                      0.499513 |                     0.490498 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.064123 |                   0.0 |                       0.0 |                   0.732518 |                  4.080005 |                 0.050254 |        -0.17 |        -0.050254 |         -3.347487 |             40.869565 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.002485 | -0.002485 |      0.0 |  0.014656 |       0.0 |     -2.96865 |     0.744901 |      0.0 |      0.0 | 0.265517 | 0.265517 |  0.039093 |  0.063459 |  0.152806 | 0.995125 | 0.297082 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.162324 | 12.838478 | 13.424116 |       0.0 | 0.969967 |  0.130249 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |            0.74548 |          2.1 |                 0.318472 |                      0.492103 |                     0.490225 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.064284 |                   0.0 |                       0.0 |                   0.734215 |                  4.093695 |                 0.049796 |    -0.170571 |        -0.049796 |          -3.35948 |             44.761905 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.008706 | -0.008706 |      0.0 |  0.013871 |       0.0 |    -2.976098 |      0.74431 |      0.0 |      0.0 | 0.260938 | 0.260938 |     0.037 |  0.064641 |  0.151437 | 0.996375 | 0.297829 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.171029 | 13.085881 | 13.410245 |       0.0 | 0.977275 |  0.096865 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           0.496987 |          1.4 |                 0.320984 |                      0.483429 |                     0.498045 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | 0.059711 |                   0.0 |                       0.0 |                   0.676981 |                  3.682358 |                 0.047666 |    -0.153432 |        -0.047666 |         -3.005377 |             67.142857 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.007626 | -0.007626 |      0.0 |   0.01263 |       0.0 |    -2.764404 |     0.743724 |      0.0 |      0.0 | 0.237602 | 0.237602 |  0.033689 |  0.057715 |   0.13907 | 0.997453 | 0.298474 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.178655 | 13.310935 | 13.397615 |       0.0 | 0.984583 |  0.068054 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |           0.390489 |          1.1 |                 0.296635 |                      0.471927 |                     0.495211 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 |       0.0 | 0.061361 |              0.001024 |                   0.00019 |                    0.58767 |                  3.039527 |                 0.044085 |    -0.125623 |        -0.043895 |         -2.451857 |             85.454545 |                 100.0 |             28069.018521 |              106.754223 |      0.0 |      0.0 | -0.005975 | -0.005975 |      0.0 |  0.012972 |       0.0 |     -2.84079 |     0.743141 |      0.0 |      0.0 | 0.239705 |  0.08668 |  0.034603 |  0.062924 |  0.141422 | 0.998593 | 0.299156 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.18463 | 13.384643 | 13.384643 |  -0.05111 | 0.991891 |  0.037593 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |            0.35499 |          1.0 |                 0.303761 |                      0.460234 |                     0.473204 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 |       0.0 | 0.074463 |              0.017281 |                  0.003209 |                   0.500214 |                  2.412046 |                 0.039431 |    -0.083221 |        -0.036222 |         -1.911833 |                  94.0 |                 100.0 |             28069.018521 |             4298.215505 |      0.0 |      0.0 |  -0.00042 |  -0.00042 |      0.0 |  0.011392 |       0.0 |    -3.447378 |     0.742797 |      0.0 |      0.0 | 0.288732 |      0.0 |  0.030387 |  0.053182 |   0.11979 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.18505 | 13.373251 | 13.373251 | -0.096437 | 0.996231 | -0.005971 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           0.496987 |          1.4 |                 0.363641 |                      0.451775 |                     0.471295 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -0.019902 | 0.075195 |              0.052153 |                  0.009686 |                   0.417852 |                  1.815487 |                  0.03091 |    -0.023492 |        -0.021225 |         -1.397635 |             67.142857 |                 100.0 |             28069.018521 |             4298.215505 |      0.0 |      0.0 |  0.004737 |  0.004737 |      0.0 |  0.007944 |       0.0 |    -3.501169 |     0.742537 |      0.0 |      0.0 | 0.292988 |      0.0 |  0.021191 |  0.034201 |  0.076617 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.180313 | 13.365307 | 13.365307 | -0.097858 | 0.999534 | -0.050266 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           0.461488 |          1.3 |                 0.453913 |                      0.447245 |                     0.502925 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -0.167553 | 0.053271 |               0.09943 |                  0.018466 |                   0.373605 |                  1.493002 |                  0.02343 |     0.037222 |        -0.004965 |         -1.119397 |             72.307692 |                 100.0 |              3952.988639 |             2243.642922 |      0.0 |      0.0 |  0.011652 |  0.011652 |      0.0 |  0.003956 |       0.0 |    -2.633802 |     0.742279 |      0.0 |      0.0 | 0.220273 |      0.0 |  0.010551 |  0.013468 |  0.028984 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.168661 | 13.361352 | 13.361352 | -0.073571 | 1.002813 | -0.072637 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           0.461488 |          1.3 |                 0.496723 |                      0.442058 |                     0.525459 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -0.242123 | 0.038938 |              0.084965 |                  0.015779 |                   0.351214 |                  1.331032 |                 0.020561 |     0.029505 |        -0.004782 |         -0.979819 |             72.307692 |                 100.0 |              1813.687715 |             1354.658798 |      0.0 |      0.0 |   0.01619 |   0.01619 |      0.0 |   0.00384 |       0.0 |    -2.044829 |     0.741987 |      0.0 |      0.0 | 0.170965 |      0.0 |  0.010243 |  0.007445 |  0.022469 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.15247 | 13.357512 | 13.357512 | -0.057102 | 1.006539 | -0.080675 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           0.496987 |          1.4 |                 0.551149 |                      0.441361 |                     0.548501 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -0.268918 |  0.02584 |              0.074843 |                  0.013899 |                      0.348 |                  1.306141 |                 0.018226 |     0.020421 |        -0.004327 |         -0.958141 |             67.142857 |                 100.0 |              1111.794179 |              926.445145 |      0.0 |      0.0 |   0.01755 |   0.01755 |      0.0 |  0.002239 |       0.0 |    -1.465237 |     0.741663 |      0.0 |      0.0 | 0.122471 |      0.0 |  0.005973 | -0.000656 |  0.009643 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.13492 | 13.355272 | 13.355272 | -0.040905 | 1.010687 | -0.075616 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |           0.816478 |          2.3 |                 0.593188 |                      0.441918 |                     0.566519 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -0.252053 | 0.016735 |               0.04016 |                  0.007458 |                   0.348493 |                  1.307301 |                  0.01667 |    -0.014311 |        -0.009212 |         -0.958808 |             40.869565 |                 100.0 |               871.594617 |              756.218729 |      0.0 |      0.0 |  0.020396 |  0.020396 |      0.0 |  0.001027 |       0.0 |    -1.026803 |     0.741297 |      0.0 |      0.0 |  0.08581 |      0.0 |  0.002739 | -0.007461 |   0.00449 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.114524 | 13.354245 | 13.354245 | -0.028661 | 1.015398 |  -0.06145 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           0.993973 |          2.8 |                 0.633417 |                      0.443734 |                     0.584724 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -0.204835 | 0.008453 |              0.009627 |                  0.001788 |                   0.349042 |                  1.307629 |                 0.014965 |    -0.044858 |        -0.013177 |         -0.958586 |             33.571429 |                 100.0 |                728.08904 |              647.654425 |      0.0 |      0.0 |  0.017649 |  0.017649 |      0.0 | -0.000966 |       0.0 |    -0.596195 |     0.740889 |      0.0 |      0.0 |  0.04982 |      0.0 | -0.002576 | -0.014169 | -0.003567 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.096875 | 13.355211 | 13.355211 |  -0.01664 |  1.02067 | -0.039004 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |            1.10047 |          3.1 |                 0.656792 |                      0.447104 |                     0.600343 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 | -0.130013 | 0.002291 |              0.000026 |                  0.000005 |                   0.349662 |                   1.30686 |                 0.013978 |    -0.054426 |        -0.013974 |         -0.957199 |             30.322581 |                 100.0 |               666.767813 |               73.465121 |      0.0 |      0.0 |  0.016667 |  0.016667 |      0.0 | -0.001874 |       0.0 |    -0.236056 |     0.740429 |      0.0 |      0.0 | 0.019727 |      0.0 |    -0.005 | -0.016735 | -0.007761 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.174209 | 13.357085 | 13.357085 | -0.006589 | 1.026649 | -0.010394 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           1.242466 |          3.5 |                 0.661557 |                      0.450781 |                     0.609629 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.034648 | -0.00019 |                   0.0 |                       0.0 |                   0.350354 |                  1.305996 |                 0.013867 |    -0.054417 |        -0.013867 |         -0.955642 |             26.857143 |                 100.0 |               656.365228 |               69.444444 |      0.0 |      0.0 |  0.020103 |  0.020103 |      0.0 |  -0.00141 |       0.0 |    -0.025866 |     0.739915 |      0.0 |      0.0 | 0.002162 |      0.0 |  -0.00376 | -0.016164 | -0.006056 | 0.999225 | 0.299535 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.154106 | 13.358495 | 13.358495 | -0.000722 | 1.033373 |  0.020695 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           1.561958 |          4.4 |                 0.656792 |                      0.455629 |                      0.61078 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.350478 |                  1.303739 |                 0.013313 |    -0.054322 |        -0.013313 |          -0.95326 |             21.363636 |                 100.0 |               661.679848 |               69.444444 | 0.236106 | 0.059027 |  0.012597 |  0.012597 |      0.0 | -0.002741 |  0.000865 |          0.0 |     0.739823 | 0.030956 | 0.030956 |      0.0 |      0.0 | -0.007312 | -0.016197 | -0.010197 | 0.998069 | 0.298843 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.187403 | 13.389265 | 13.597342 |       0.0 | 1.034583 |  0.051595 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           1.348964 |          3.8 |                 0.676036 |                      0.460234 |                      0.61078 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.350948 |                  1.301296 |                  0.01222 |    -0.054221 |         -0.01222 |         -0.950347 |             24.736842 |                 100.0 |               617.558461 |               69.444444 | 0.175403 |  0.02631 |  0.004093 |  0.004093 |      0.0 |  -0.00531 |  0.000951 |          0.0 |     0.739474 | 0.071419 | 0.071419 |      0.0 |      0.0 | -0.014165 | -0.020958 | -0.022903 | 0.996913 |  0.29815 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.195907 | 13.349301 | 13.778056 |       0.0 | 1.039177 |  0.082495 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |           2.023445 |          5.7 |                 0.685844 |                      0.465172 |                      0.61078 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.351301 |                  1.298642 |                 0.011755 |     -0.05411 |        -0.011755 |         -0.947341 |             16.491228 |                 100.0 |               598.069608 |               69.444444 | 0.371907 | 0.037191 |  0.005159 |  0.005159 |      0.0 | -0.008229 |  0.002367 |          0.0 |     0.739212 | 0.127935 | 0.127935 |      0.0 |      0.0 | -0.021951 | -0.030168 | -0.040364 | 0.995756 | 0.297459 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.194841 | 13.266268 | 14.158192 |       0.0 | 1.042638 |  0.113395 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           1.987946 |          5.6 |                 0.705836 |                      0.471336 |                      0.61078 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.352045 |                  1.295068 |                 0.010788 |    -0.053961 |        -0.010788 |         -0.943022 |             16.785714 |                 100.0 |               562.173975 |               69.444444 | 0.652841 |      0.0 | -0.001162 | -0.001162 |      0.0 |  -0.01135 |  0.005467 |          0.0 |      0.73866 | 0.186392 | 0.186392 |      0.0 |      0.0 | -0.030275 | -0.037301 | -0.056788 |   0.9946 | 0.296769 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.201162 | 13.089899 | 14.822383 |       0.0 | 1.049985 |  0.144295 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           2.058944 |          5.8 |                  0.72634 |                      0.478769 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.352784 |                  1.290674 |                 0.010126 |    -0.053778 |        -0.010126 |          -0.93789 |             16.206897 |                 100.0 |               532.195149 |               69.444444 | 0.471162 |      0.0 |  0.000413 |  0.000413 |      0.0 | -0.013984 |  0.004734 |          0.0 |     0.738111 | 0.232495 | 0.232495 |      0.0 |      0.0 | -0.037302 | -0.045743 | -0.072919 | 0.993443 | 0.296079 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.199587 | 12.869162 | 15.307529 |       0.0 | 1.057324 |  0.175195 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           2.200941 |          6.2 |                 0.731547 |                      0.486761 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                    0.35352 |                   1.28612 |                 0.010003 |    -0.053588 |        -0.010003 |         -0.932601 |              15.16129 |                 100.0 |               525.492703 |               69.444444 | 0.939587 |      0.0 |  0.002303 |  0.002303 |      0.0 | -0.015071 |  0.009835 |          0.0 |     0.737565 | 0.268643 | 0.268643 |      0.0 |      0.0 | -0.040202 | -0.049694 | -0.079892 | 0.992287 |  0.29539 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 2.197697 | 12.612207 | 16.262188 |       0.0 | 1.064691 |  0.206095 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           1.987946 |          5.6 |                 0.742061 |                      0.496568 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.351822 |                  1.277745 |                 0.009642 |    -0.053239 |        -0.009642 |         -0.925923 |             16.785714 |                 100.0 |               511.865132 |               69.444444 | 1.877697 |      0.0 |  0.001791 |  0.001791 |      0.0 | -0.015425 |  0.020046 |          0.0 |      0.73702 | 0.305104 | 0.305104 |      0.0 |      0.0 | -0.041144 | -0.050357 | -0.081859 |  0.99113 | 0.294702 | 0.541101 | 0.181979 |  0.0 |  0.0 | 0.0 |  0.0 | 0.181979 | 0.002361 | 0.000656 | 2.198209 | 12.317628 | 17.614208 |       0.0 | 1.072098 |  0.236995 | 71.684129 | 0.181979 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002361 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000656 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |            1.84595 |          5.2 |                  0.76893 |                      0.509073 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.352537 |                  1.268373 |                 0.008604 |    -0.052849 |        -0.008604 |         -0.915837 |             18.076923 |                 100.0 |               477.764673 |               69.444444 | 1.972209 |      0.0 | -0.001204 | -0.001204 |      0.0 | -0.018246 |       0.0 |          0.0 |     0.736486 | 0.290813 | 0.290813 |      0.0 |      0.0 | -0.048671 | -0.057065 | -0.097132 | 0.989974 | 0.294015 | 2.388503 | 0.803637 |  0.0 |  0.0 | 0.0 |  0.0 | 0.803637 | 0.026478 | 0.007355 | 2.201204 | 12.039273 | 17.216161 |       0.0 | 1.079406 |  0.267895 | 73.268995 |      0.5 | 0.303637 |   0.0 |   0.0 |  0.0 | 0.011126 | 0.015352 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007355 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |           2.023445 |          5.7 |                 0.807986 |                      0.524283 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.353246 |                  1.257174 |                  0.00723 |    -0.052382 |         -0.00723 |         -0.903928 |             16.491228 |                 100.0 |               430.975107 |               69.444444 | 1.693204 |      0.0 | -0.002413 | -0.002413 |      0.0 | -0.024333 |       0.0 |          0.0 |     0.735955 | 0.392848 | 0.392848 |      0.0 |      0.0 | -0.064906 | -0.073535 | -0.131211 | 0.988818 | 0.293328 |  2.25577 | 0.759754 |  0.0 |  0.0 | 0.0 |  0.0 | 0.759754 | 0.065539 | 0.018205 | 2.202413 | 11.662886 | 16.677927 |       0.0 | 1.086714 |  0.298795 | 74.765011 |      0.5 | 0.259754 |   0.0 |   0.0 |  0.0 | 0.023754 | 0.041786 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018205 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           2.342937 |          6.6 |                 0.836943 |                      0.541728 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.353951 |                  1.245134 |                 0.006273 |    -0.051881 |        -0.006273 |         -0.891183 |             14.242424 |                 100.0 |               400.183656 |               69.444444 | 1.224413 |      0.0 | -0.002135 | -0.002135 |      0.0 | -0.030654 |       0.0 |          0.0 |     0.735428 | 0.498663 | 0.498663 |      0.0 |      0.0 | -0.081767 | -0.091059 | -0.166554 | 0.987661 | 0.292642 | 1.938812 | 0.653461 |  0.0 |  0.0 | 0.0 |  0.0 | 0.653461 |  0.09413 | 0.026147 | 2.202135 | 11.184742 | 15.994182 |       0.0 | 1.094022 |  0.329695 | 76.050362 |      0.5 | 0.153461 |   0.0 |   0.0 |  0.0 | 0.036055 | 0.058075 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026147 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           2.484933 |          7.0 |                 0.860763 |                      0.562266 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.354651 |                  1.232336 |                 0.005743 |    -0.051347 |        -0.005743 |         -0.877685 |             13.428571 |                 100.0 |               378.717229 |               69.444444 | 0.472135 |      0.0 |  0.005737 |  0.005737 |      0.0 | -0.034262 |       0.0 |          0.0 |     0.734905 | 0.565607 | 0.565607 |      0.0 |      0.0 | -0.091392 | -0.103263 | -0.188913 | 0.986505 | 0.291957 | 1.282686 | 0.432319 |  0.0 |  0.0 | 0.0 |  0.0 | 0.432319 | 0.106958 |  0.02971 | 2.194263 | 10.641883 | 15.217893 |       0.0 |  1.10133 |  0.360595 |  76.90073 | 0.432319 |      0.0 |   0.0 |   0.0 |  0.0 | 0.047162 | 0.059796 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02971 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |            2.59143 |          7.3 |                 0.860763 |                      0.585275 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.355346 |                  1.218349 |                 0.006177 |    -0.050765 |        -0.006177 |         -0.863002 |             12.876712 |                 100.0 |               376.739864 |               69.444444 | 0.182263 |      0.0 |  0.020252 |  0.020252 |      0.0 |  -0.03348 |       0.0 |          0.0 |     0.734385 | 0.566874 | 0.566874 |      0.0 |      0.0 | -0.089306 | -0.106208 | -0.189336 | 0.985348 | 0.291273 | 0.994677 |  0.33508 |  0.0 |  0.0 | 0.0 |  0.0 |  0.33508 | 0.109702 | 0.030473 | 2.179748 | 10.097174 | 14.438959 |       0.0 | 1.108638 |  0.391495 | 77.560326 |  0.33508 |      0.0 |   0.0 |   0.0 |  0.0 | 0.055849 | 0.053853 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030473 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           2.555931 |          7.2 |                 0.879017 |                      0.610223 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.356036 |                  1.204312 |                 0.005657 |     -0.05018 |        -0.005657 |         -0.848276 |             13.055556 |                 100.0 |               362.740081 |               69.444444 | 0.449748 |      0.0 |  0.022273 |  0.022273 |      0.0 | -0.035493 |       0.0 |          0.0 |      0.73387 | 0.601263 | 0.601263 |      0.0 |      0.0 | -0.094676 | -0.111803 | -0.200822 | 0.984192 |  0.29059 | 1.311668 | 0.441729 |  0.0 |  0.0 | 0.0 |  0.0 | 0.441729 | 0.112946 | 0.031374 | 2.177727 |  9.519253 | 13.612532 |       0.0 | 1.115946 |  0.422395 | 78.430265 | 0.441729 |      0.0 |   0.0 |   0.0 |  0.0 | 0.064445 | 0.048501 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031374 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           2.697927 |          7.6 |                 0.879017 |                      0.635914 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.356722 |                  1.190153 |                 0.006005 |     -0.04959 |        -0.006005 |         -0.833431 |             12.368421 |                 100.0 |               360.276004 |               69.444444 | 0.165727 |      0.0 |  0.035379 |  0.035379 |      0.0 | -0.035367 |       0.0 |          0.0 |     0.733357 | 0.611699 | 0.611699 |      0.0 |      0.0 | -0.094338 | -0.115974 | -0.204308 | 0.983035 | 0.289908 |  1.04273 | 0.351009 |  0.0 |  0.0 | 0.0 |  0.0 | 0.351009 | 0.116694 | 0.032415 | 2.164621 |  8.930697 | 12.770896 |       0.0 | 1.123254 |  0.453295 | 79.121986 | 0.351009 |      0.0 |   0.0 |   0.0 |  0.0 | 0.073014 |  0.04368 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032415 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |            2.59143 |          7.3 |                 0.897612 |                      0.665151 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 |       0.0 |      0.0 |              0.000866 |                  0.000161 |                   0.357403 |                  1.174779 |                 0.005774 |    -0.048083 |        -0.005613 |         -0.817376 |             12.876712 |                 100.0 |               348.486262 |               95.717689 | 0.058621 |      0.0 |  0.045768 |  0.045768 |      0.0 | -0.035585 |       0.0 |          0.0 |     0.732848 | 0.625128 | 0.625128 |      0.0 |      0.0 | -0.094922 | -0.119484 | -0.208793 | 0.981879 | 0.289226 | 0.955046 | 0.321298 |  0.0 |  0.0 | 0.0 |  0.0 | 0.321298 | 0.119149 | 0.033097 | 2.154232 |  8.328711 | 11.910057 |       0.0 | 1.130562 |  0.484195 | 79.755734 | 0.321298 |      0.0 |   0.0 |   0.0 |  0.0 |  0.07981 | 0.039339 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033097 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           2.520432 |          7.1 |                 0.891375 |                      0.694736 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 |       0.0 |      0.0 |              0.016638 |                   0.00309 |                   0.358079 |                   1.15845 |                 0.005615 |     -0.03163 |        -0.002525 |         -0.800372 |             13.239437 |                 100.0 |               347.047025 |              330.883603 | 0.236232 |      0.0 |  0.036552 |  0.036552 |      0.0 | -0.035482 |       0.0 |          0.0 |     0.732343 | 0.620432 | 0.620432 |      0.0 |      0.0 | -0.094646 | -0.115104 | -0.207224 | 0.980722 | 0.288545 | 1.126094 | 0.378664 |  0.0 |  0.0 | 0.0 |  0.0 | 0.378664 | 0.122221 |  0.03395 | 2.163448 |  7.731243 | 11.055677 |       0.0 |  1.13787 |  0.515095 | 80.503164 | 0.378664 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086792 | 0.035429 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03395 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           2.520432 |          7.1 |                 0.903886 |                       0.72095 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 |       0.0 |      0.0 |              0.050796 |                  0.009433 |                    0.35875 |                  1.144953 |                 0.005481 |     0.003089 |         0.003953 |         -0.786203 |             13.239437 |                 100.0 |                338.28182 |               323.06047 | 0.057448 |      0.0 |  0.051557 |  0.051557 |      0.0 | -0.036172 |       0.0 |          0.0 |     0.731841 | 0.658344 | 0.658344 |      0.0 |      0.0 | -0.096487 | -0.119447 | -0.219887 | 0.979566 | 0.287865 | 1.001961 |  0.33675 |  0.0 |  0.0 | 0.0 |  0.0 |  0.33675 | 0.125695 | 0.034915 | 2.148443 |   7.09604 | 10.147337 |       0.0 | 1.145178 |  0.545995 | 81.168374 |  0.33675 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093787 | 0.031908 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034915 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           3.159415 |          8.9 |                 0.942351 |                      0.742943 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 |       0.0 |      0.0 |              0.077505 |                  0.014394 |                    0.34996 |                  1.076886 |                 0.004536 |     0.032634 |         0.009858 |         -0.726926 |             10.561798 |                 100.0 |               323.417475 |               309.73436 | 0.042443 |      0.0 |  0.119563 |  0.119563 |      0.0 | -0.042299 |       0.0 |          0.0 |     0.731342 | 0.831193 | 0.831193 |      0.0 |      0.0 | -0.112831 | -0.154929 | -0.277618 |  0.97841 | 0.287186 | 1.236004 | 0.415271 |  0.0 |  0.0 | 0.0 |  0.0 | 0.415271 | 0.129818 | 0.036061 | 2.080437 |  6.290962 |  8.996075 |       0.0 | 1.152486 |  0.576895 | 81.989106 | 0.415271 |      0.0 |   0.0 |   0.0 |  0.0 | 0.101082 | 0.028737 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036061 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           3.372409 |          9.5 |                 0.962118 |                      0.763489 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 |       0.0 |      0.0 |              0.077784 |                  0.014446 |                   0.350605 |                  1.067855 |                  0.00449 |      0.03329 |         0.009955 |          -0.71725 |              9.894737 |                 100.0 |               314.114443 |              301.355814 |      0.0 |      0.0 |  0.150085 |  0.150085 |      0.0 | -0.044819 |       0.0 |          0.0 |     0.730847 | 0.897929 | 0.897929 |      0.0 |      0.0 | -0.119553 |   -0.1704 | -0.299908 | 0.977253 | 0.286507 | 1.290436 | 0.433481 |  0.0 |  0.0 | 0.0 |  0.0 | 0.433481 | 0.135314 | 0.037587 | 1.930352 |  5.419901 |  7.750458 |       0.0 | 1.159794 |  0.607795 | 82.846061 | 0.433481 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109434 |  0.02588 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037587 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           2.804424 |          7.9 |                 0.962118 |                      0.781298 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 |       0.0 |      0.0 |              0.073615 |                  0.013671 |                   0.351246 |                   1.06128 |                 0.006024 |     0.029395 |         0.007647 |         -0.710034 |             11.898734 |                 100.0 |               317.248394 |              304.181631 |      0.0 |      0.0 |  0.175819 |  0.175819 |      0.0 | -0.032937 |       0.0 |          0.0 |     0.730355 | 0.733436 | 0.733436 |      0.0 |      0.0 | -0.087858 | -0.149462 | -0.244968 | 0.976097 | 0.285829 | 1.053271 | 0.353682 |  0.0 |  0.0 | 0.0 |  0.0 | 0.353682 | 0.140077 |  0.03891 | 1.848533 |  4.706381 |  6.730125 |       0.0 | 1.167102 |  0.638695 |  83.54565 | 0.353682 |      0.0 |   0.0 |   0.0 |  0.0 | 0.116769 | 0.023308 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03891 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           3.407908 |          9.6 |                 0.968788 |                      0.797357 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 |       0.0 |      0.0 |              0.041066 |                  0.007627 |                   0.351882 |                  1.056546 |                 0.007091 |    -0.002957 |         0.000536 |         -0.704664 |              9.791667 |                 100.0 |               317.160019 |              304.101992 |      0.0 |      0.0 |  0.251909 |  0.251909 |      0.0 | -0.033021 |       0.0 |          0.0 |     0.729867 |  0.78745 |  0.78745 |      0.0 |      0.0 | -0.088081 | -0.174392 | -0.263008 |  0.97494 | 0.285153 | 1.131578 | 0.379831 |  0.0 |  0.0 | 0.0 |  0.0 | 0.379831 |  0.14422 | 0.040061 | 1.596624 |  3.938159 |  5.631567 |       0.0 |  1.17441 |  0.669595 | 84.297396 | 0.379831 |      0.0 |   0.0 |   0.0 |  0.0 | 0.123228 | 0.020992 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040061 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |           3.230413 |          9.1 |                 0.955489 |                       0.81108 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 |       0.0 |      0.0 |              0.012289 |                  0.002282 |                   0.362041 |                  1.110113 |                 0.008451 |    -0.033966 |        -0.006169 |         -0.748073 |              10.32967 |                 100.0 |               318.585115 |              305.385903 |      0.0 |      0.0 |  0.229011 |  0.229011 |      0.0 | -0.030224 |       0.0 |          0.0 |     0.729382 | 0.709692 | 0.709692 |      0.0 |      0.0 | -0.080621 | -0.162585 | -0.237037 | 0.973784 | 0.284476 | 1.020306 | 0.342328 |  0.0 |  0.0 | 0.0 |  0.0 | 0.342328 | 0.148275 | 0.041188 | 1.461614 |  3.245794 |  4.641485 |       0.0 | 1.181718 |  0.700495 | 84.975374 | 0.342328 |      0.0 |   0.0 |   0.0 |  0.0 |  0.12937 | 0.018905 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041188 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           3.088417 |          8.7 |                 0.929372 |                      0.822835 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 |       0.0 |      0.0 |              0.000021 |                  0.000004 |                   0.362685 |                  1.107879 |                 0.009318 |     -0.04614 |        -0.009314 |         -0.745194 |             10.804598 |                 100.0 |               326.753049 |               72.472463 |      0.0 |      0.0 |  0.217285 |  0.217285 |      0.0 | -0.025254 |       0.0 |          0.0 |       0.7289 | 0.615227 | 0.615227 |      0.0 |      0.0 | -0.067363 | -0.147437 | -0.205486 | 0.972627 | 0.283801 | 0.884561 | 0.296601 |  0.0 |  0.0 | 0.0 |  0.0 | 0.296601 | 0.151304 | 0.042029 | 1.244329 |   2.64488 |  3.782178 |       0.0 | 1.189026 |  0.731395 | 85.563334 | 0.296601 |      0.0 |   0.0 |   0.0 |  0.0 | 0.134277 | 0.017026 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042029 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           3.159415 |          8.9 |                 0.903886 |                      0.833518 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.363326 |                   1.10789 |                 0.012225 |    -0.046162 |        -0.012225 |         -0.744564 |             10.561798 |                 100.0 |               345.164099 |               69.444444 |      0.0 |      0.0 |  0.290175 |  0.290175 |      0.0 | -0.013522 |       0.0 |          0.0 |     0.728421 | 0.489627 | 0.489627 |      0.0 |      0.0 |  -0.03607 |  -0.13969 | -0.163536 | 0.971471 | 0.283127 |  0.70267 | 0.235415 |  0.0 |  0.0 | 0.0 |  0.0 | 0.235415 | 0.153011 | 0.042503 | 0.954154 |  2.162958 |   3.09303 |       0.0 | 1.196334 |  0.762295 |  86.03059 | 0.235415 |      0.0 |   0.0 |   0.0 |  0.0 | 0.137677 | 0.015334 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042503 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           2.804424 |          7.9 |                 0.879017 |                      0.843583 |                      0.61078 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.363962 |                  1.107494 |                 0.010983 |    -0.046146 |        -0.010983 |         -0.743532 |             11.898734 |                 100.0 |               346.409011 |               69.444444 |      0.0 |      0.0 |  0.197149 |  0.197149 |      0.0 | -0.016545 |       0.0 |          0.0 |     0.727945 | 0.455852 | 0.455852 |      0.0 |      0.0 | -0.044134 | -0.119103 | -0.152254 | 0.970315 | 0.282453 | 0.655355 | 0.219354 |  0.0 |  0.0 | 0.0 |  0.0 | 0.219354 | 0.153803 | 0.042723 | 0.757005 |  1.716238 |   2.45422 |       0.0 | 1.203642 |  0.793195 |  86.46659 | 0.219354 |      0.0 |   0.0 |   0.0 |  0.0 | 0.139993 |  0.01381 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042723 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           2.520432 |          7.1 |                 0.860763 |                       0.85201 |                      0.61078 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.364594 |                  1.107885 |                 0.010915 |    -0.046162 |        -0.010915 |         -0.743291 |             13.239437 |                 100.0 |               351.307816 |               69.444444 |      0.0 |      0.0 |   0.15874 |   0.15874 |      0.0 | -0.015346 |       0.0 |          0.0 |     0.727473 | 0.401993 | 0.401993 |      0.0 |      0.0 | -0.040935 | -0.104245 | -0.134266 | 0.969158 |  0.28178 | 0.578444 |  0.19341 |  0.0 |  0.0 | 0.0 |  0.0 |  0.19341 | 0.154144 | 0.042818 | 0.598265 |  1.322463 |  1.891123 |       0.0 |  1.21095 |  0.824095 | 86.851624 |  0.19341 |      0.0 |   0.0 |   0.0 |  0.0 | 0.141706 | 0.012438 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042818 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           2.271939 |          6.4 |                  0.86681 |                       0.86026 |                      0.61078 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.365222 |                  1.108737 |                 0.010948 |    -0.046197 |        -0.010948 |         -0.743514 |               14.6875 |                 100.0 |               346.953939 |               69.444444 |      0.0 |      0.0 |  0.148995 |  0.148995 |      0.0 | -0.014314 |       0.0 |          0.0 |     0.727003 | 0.379152 | 0.379152 |      0.0 |      0.0 | -0.038181 | -0.099403 | -0.126637 | 0.968002 | 0.281108 | 0.546291 | 0.182459 |  0.0 |  0.0 | 0.0 |  0.0 | 0.182459 | 0.154101 | 0.042806 |  0.44927 |  0.950451 |  1.359145 |       0.0 | 1.218258 |  0.854995 | 87.215456 | 0.182459 |      0.0 |   0.0 |   0.0 |  0.0 |   0.1429 | 0.011201 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042806 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           2.555931 |          7.2 |                  0.86681 |                      0.867569 |                      0.61078 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.365846 |                  1.109465 |                 0.009883 |    -0.046228 |        -0.009883 |         -0.743619 |             13.055556 |                 100.0 |               340.312444 |               69.444444 |      0.0 |      0.0 |  0.137771 |  0.137771 |      0.0 | -0.019229 |       0.0 |          0.0 |     0.726537 | 0.445595 | 0.445595 |      0.0 |      0.0 | -0.051293 | -0.107419 | -0.148829 | 0.966845 | 0.280437 | 0.646216 | 0.215616 |  0.0 |  0.0 | 0.0 |  0.0 | 0.215616 |  0.15444 |   0.0429 | 0.311499 |  0.511999 |  0.732158 |       0.0 | 1.225566 |  0.885895 | 87.646056 | 0.215616 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144352 | 0.010088 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0429 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           2.484933 |          7.0 |                 0.854752 |                      0.873404 |                      0.61078 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.366466 |                   1.11033 |                 0.009631 |    -0.046264 |        -0.009631 |         -0.743864 |             13.428571 |                 100.0 |                342.00064 |               69.444444 |      0.0 |      0.0 |  0.115706 |  0.115706 |      0.0 | -0.019586 |       0.0 |          0.0 |     0.726074 | 0.430317 | 0.430317 |      0.0 |      0.0 | -0.052245 | -0.101112 | -0.143726 | 0.965689 | 0.279766 | 0.631814 | 0.210604 |  0.0 |  0.0 | 0.0 |  0.0 | 0.210604 | 0.155213 | 0.043115 | 0.289793 |  0.083867 |  0.119931 |       0.0 | 1.232874 |  0.916795 | 88.067266 | 0.210604 |      0.0 |   0.0 |   0.0 |  0.0 | 0.146128 | 0.009085 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043115 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |           2.413935 |          6.8 |                 0.802302 |                      0.876717 |                      0.61078 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |      0.0 |                   0.0 |                       0.0 |                   0.367082 |                  1.112997 |                  0.01195 |    -0.046375 |         -0.01195 |         -0.745915 |             13.823529 |                 100.0 |               376.513363 |               69.444444 |      0.0 |      0.0 |   0.12638 |   0.12638 |      0.0 | -0.010581 |       0.0 |          0.0 |     0.725614 | 0.288837 | 0.083867 |      0.0 |      0.0 | -0.028225 | -0.080196 | -0.096472 | 0.964532 | 0.279097 | 0.130512 |  0.04344 |  0.0 |  0.0 | 0.0 |  0.0 |  0.04344 | 0.153807 | 0.042724 | 0.257413 |       0.0 |       0.0 |       0.0 | 1.240182 |  0.947695 | 88.154338 |  0.04344 |      0.0 |   0.0 |   0.0 |  0.0 | 0.145625 | 0.008182 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042724 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           2.271939 |          6.4 |                 0.807986 |                      0.879785 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.08424 |                   0.0 |                       0.0 |                   1.170072 |                  1.116023 |                      0.0 |    -0.046501 |              0.0 |          0.054049 |               14.6875 |                 100.0 |                372.43897 |               69.444444 |      0.0 |      0.0 |  0.121939 |  0.121939 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.960223 | 0.276609 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.149791 | 0.041609 | 0.135475 |       0.0 |       0.0 |       0.0 |      nan |  1.062835 | 88.154338 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.142422 | 0.007369 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041609 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |            2.23644 |          6.3 |                 0.796653 |                      0.881066 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.07992 |                   0.0 |                       0.0 |                   1.170072 |                  1.119656 |                      0.0 |    -0.046652 |              0.0 |          0.050416 |             14.920635 |                 100.0 |                379.21964 |               69.444444 |      0.0 |      0.0 |  0.111599 |  0.111599 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.956076 | 0.274224 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.14538 | 0.040383 | 0.023875 |       0.0 |       0.0 |       0.0 |      nan |  1.173655 | 88.154338 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.138743 | 0.006637 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040383 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |           2.023445 |          5.7 |                  0.78546 |                      0.880041 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.0756 |                   0.0 |                       0.0 |                   1.170072 |                  1.123475 |                      0.0 |    -0.046811 |              0.0 |          0.046597 |             16.491228 |                 100.0 |               382.896762 |               69.444444 |      0.0 |      0.0 |  0.076984 |  0.076984 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |  0.95209 | 0.271943 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.141136 | 0.039205 | 0.040892 |       0.0 |       0.0 |       0.0 |      nan |  1.280155 | 88.154338 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.135159 | 0.005977 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039205 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |            1.84595 |          5.2 |                 0.774405 |                      0.877228 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.07128 |                   0.0 |                       0.0 |                   1.170072 |                  1.127474 |                      0.0 |    -0.046978 |              0.0 |          0.042598 |             18.076923 |                 100.0 |               388.585014 |               69.444444 |      0.0 |      0.0 |  0.055406 |  0.040892 | 0.004609 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.948266 | 0.269762 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.137051 |  0.03807 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.382335 | 88.149729 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.131668 | 0.005383 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03807 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |           1.632956 |          4.6 |                 0.774405 |                      0.873404 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.07128 |                   0.0 |                       0.0 |                   1.170072 |                  1.131856 |                      0.0 |    -0.047161 |              0.0 |          0.038216 |             20.434783 |                 100.0 |               389.391663 |               69.444444 |      0.0 |      0.0 |  0.051903 |       0.0 | 0.017849 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944442 | 0.267591 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.133115 | 0.036976 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.484515 | 88.131879 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.128267 | 0.004848 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036976 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           1.739453 |          4.9 |                 0.758081 |                      0.868834 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.0648 |                   0.0 |                       0.0 |                   1.170072 |                  1.136687 |                      0.0 |    -0.047362 |              0.0 |          0.033385 |             19.183673 |                 100.0 |               406.503898 |               69.444444 |      0.0 |      0.0 |  0.065481 |       0.0 | 0.021505 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |  0.94086 | 0.265565 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.12932 | 0.035922 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.580215 | 88.110374 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124953 | 0.004366 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035922 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |            2.23644 |          6.3 |                  0.70079 |                      0.860763 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.04104 |                   0.0 |                       0.0 |                   1.170072 |                  1.144252 |                      0.0 |    -0.047677 |              0.0 |           0.02582 |             14.920635 |                 100.0 |               476.187631 |               69.444444 |      0.0 |      0.0 |  0.108633 |       0.0 | 0.029433 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.938168 | 0.264048 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.125658 | 0.034905 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.652155 | 88.080941 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121726 | 0.003932 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034905 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           2.058944 |          5.8 |                 0.666353 |                      0.851015 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 | -0.02592 |                   0.0 |                       0.0 |                   1.170072 |                  1.152031 |                      0.0 |    -0.048001 |              0.0 |          0.018042 |             16.206897 |                 100.0 |               527.511054 |               69.444444 |      0.0 |      0.0 |  0.092335 |       0.0 | 0.026177 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.936041 | 0.262852 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.122123 | 0.033923 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.708975 | 88.054764 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.118582 | 0.003541 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033923 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           2.165441 |          6.1 |                 0.647352 |                      0.839643 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 | -0.01728 |              0.002833 |                  0.000526 |                   1.170072 |                  1.158654 |                      0.0 |    -0.045445 |              0.0 |          0.011418 |             15.409836 |                 100.0 |               554.147004 |               98.271341 |      0.0 |      0.0 |  0.060806 |       0.0 | 0.012678 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.934238 |  0.26184 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.118708 | 0.032974 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.757155 | 88.042086 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115518 | 0.003189 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032974 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           2.094443 |          5.9 |                 0.647352 |                      0.828647 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 | -0.01728 |              0.056113 |                  0.010421 |                   1.170072 |                  1.166643 |                      0.0 |     0.007503 |              0.0 |          0.003429 |             15.932203 |                 100.0 |                561.16968 |               514.27391 |      0.0 |      0.0 |  0.093283 |       0.0 | 0.004612 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.932435 | 0.260831 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.115407 | 0.032058 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.805335 | 88.037474 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.112535 | 0.002872 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032058 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           2.271939 |          6.4 |                 0.633417 |                      0.816579 |                          0.0 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |  -0.0108 |              0.172031 |                  0.031949 |                   1.170072 |                  1.175846 |                      0.0 |     0.123037 |              0.0 |         -0.005774 |               14.6875 |                 100.0 |               597.191011 |              543.735846 |      0.0 |      0.0 |  0.140468 |     0.094 | 0.002004 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.930874 | 0.259958 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.112215 | 0.031171 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.847035 |  88.03547 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109628 | 0.002587 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031171 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |           2.413935 |          6.8 |                 0.624275 |                      0.802774 |                          0.0 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 | -0.00648 |              0.258205 |                  0.047952 |                   1.170072 |                   1.18712 |                      0.0 |     0.208742 |              0.0 |         -0.017048 |             13.823529 |                 100.0 |               635.543756 |               574.68751 |      0.0 |      0.0 |  0.221518 |       0.0 | 0.008516 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.929475 | 0.259177 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.109126 | 0.030313 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.884415 | 88.026953 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106796 |  0.00233 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030313 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           2.484933 |          7.0 |                 0.606339 |                      0.787548 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |  0.00216 |              0.258716 |                  0.048047 |                   1.170072 |                  1.199167 |                      0.0 |      0.20875 |              0.0 |         -0.029095 |             13.428571 |                 100.0 |               690.244818 |              618.105321 |      0.0 |      0.0 |  0.238232 |       0.0 | 0.008227 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |   0.9284 | 0.258578 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.106136 | 0.029482 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.913155 | 88.018726 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.104037 | 0.002098 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029482 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |            2.23644 |          6.3 |                 0.593188 |                      0.771892 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |  0.00864 |              0.244186 |                  0.045349 |                   1.170072 |                   1.21168 |                      0.0 |     0.193699 |              0.0 |         -0.041608 |             14.920635 |                 100.0 |               746.233781 |              661.682922 |      0.0 |      0.0 |  0.249511 |       0.0 | 0.008852 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.927567 | 0.258114 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.10324 | 0.028678 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.935415 | 88.009874 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.10135 |  0.00189 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028678 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |            2.59143 |          7.3 |                  0.58456 |                      0.755838 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |  0.01296 |              0.190885 |                   0.03545 |                   1.231767 |                  1.349977 |                      0.0 |     0.134636 |              0.0 |          -0.11821 |             12.876712 |                 100.0 |               786.197114 |              692.267741 |      0.0 |      0.0 |  0.294252 |       0.0 | 0.008645 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926895 | 0.257741 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.100434 | 0.027898 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.953355 | 88.001229 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098732 | 0.001702 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027898 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           2.662428 |          7.5 |                 0.576044 |                       0.74008 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |  0.01728 |              0.047005 |                  0.008729 |                   1.262615 |                  1.428152 |                      0.0 |    -0.012502 |              0.0 |         -0.165538 |             12.533333 |                 100.0 |               829.056899 |              724.599776 |      0.0 |      0.0 |  0.296611 |       0.0 | 0.008084 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926386 | 0.257457 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.097714 | 0.027143 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.966975 | 87.993145 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.096182 | 0.001533 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027143 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           2.484933 |          7.0 |                 0.555231 |                      0.724396 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |  0.02808 |              0.000535 |                  0.000099 |                    1.32431 |                  1.571379 |                      0.0 |    -0.064939 |              0.0 |         -0.247069 |             13.428571 |                 100.0 |               934.457331 |               72.518302 |      0.0 |      0.0 |  0.256087 |       0.0 | 0.058493 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |  0.92628 | 0.257398 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095078 |  0.02641 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.969795 | 87.934652 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093697 |  0.00138 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02641 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           2.129942 |          6.0 |                 0.543065 |                      0.709217 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.03456 |                   0.0 |                       0.0 |                    1.32431 |                  1.583999 |                      0.0 |       -0.066 |              0.0 |         -0.259688 |             15.666667 |                 100.0 |               1008.78296 |               69.444444 |      0.0 |      0.0 |  0.202105 |       0.0 | 0.053184 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926417 | 0.257475 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.09252 |   0.0257 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.966135 | 87.881468 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091277 | 0.001243 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0257 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           1.703954 |          4.8 |                  0.52721 |                       0.69432 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.0432 |                   0.0 |                       0.0 |                    1.32431 |                    1.5975 |                      0.0 |    -0.066562 |              0.0 |          -0.27319 |             19.583333 |                 100.0 |              1132.495443 |               69.444444 |      0.0 |      0.0 |   0.14534 |       0.0 | 0.044429 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926877 | 0.257731 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.090039 | 0.025011 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.953835 | 87.837039 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088919 |  0.00112 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025011 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |            1.49096 |          4.2 |                 0.523311 |                      0.680107 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.04536 |                   0.0 |                       0.0 |                    1.32431 |                  1.611195 |                      0.0 |    -0.067133 |              0.0 |         -0.286884 |             22.380952 |                 100.0 |              1170.920561 |               69.444444 |      0.0 |      0.0 |  0.125512 |       0.0 | 0.041865 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.927419 | 0.258032 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.087631 | 0.024342 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.939375 | 87.795174 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086622 | 0.001008 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024342 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           1.739453 |          4.9 |                 0.519437 |                      0.665752 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.04752 |                   0.0 |                       0.0 |                    1.32431 |                  1.625096 |                      0.0 |    -0.067712 |              0.0 |         -0.300786 |             19.183673 |                 100.0 |              1215.821448 |               69.444444 |      0.0 |      0.0 |  0.150562 |       0.0 | 0.045237 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928041 | 0.258378 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.085293 | 0.023692 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.922755 | 87.749937 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084385 | 0.000908 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023692 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           1.419962 |          4.0 |                 0.504195 |                      0.650878 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.05616 |                   0.0 |                       0.0 |                    1.32431 |                  1.639665 |                      0.0 |    -0.068319 |              0.0 |         -0.315355 |                  23.5 |                 100.0 |              1407.767234 |               69.444444 |      0.0 |      0.0 |  0.111614 |       0.0 | 0.038133 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928986 | 0.258904 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.083023 | 0.023062 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.897495 | 87.711804 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.082205 | 0.000818 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023062 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |           1.206967 |          3.4 |                 0.485698 |                      0.635721 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.06696 |                   0.0 |                       0.0 |                    1.32431 |                  1.653965 |                      0.0 |    -0.068915 |              0.0 |         -0.329655 |             27.647059 |                 100.0 |               1760.90339 |               69.444444 |      0.0 |      0.0 |  0.082224 |       0.0 | 0.030873 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.930336 | 0.259657 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.080818 | 0.022449 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.861435 | 87.680931 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.080082 | 0.000737 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022449 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |           1.206967 |          3.4 |                 0.457351 |                      0.621065 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  0.08424 |                   0.0 |                       0.0 |                    1.32431 |                  1.664915 |                      0.0 |    -0.069371 |              0.0 |         -0.340605 |             27.647059 |                 100.0 |              2989.894775 |               69.444444 |      0.0 |      0.0 |  0.060145 |       0.0 | 0.022339 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.932332 | 0.260773 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.078676 | 0.021855 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.808095 | 87.658592 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.078013 | 0.000663 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021855 |

.. _lland_v3_acker_winter_ktschnee_hourly:

acre (snow surface temperature)
-------------------------------

Due to the reasons discussed in the daily integration test
:ref:`lland_v3_acker_winter_ktschnee_daily`, enabling the iterative
adjustment of the snow surface temperature results both in a more delayed
and smoothed snow-melt, which becomes even more apparent in the following
hourly simulation results:

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> test("lland_v3_acker_winter_ktschnee_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter_ktschnee)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                     0.335784 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.473858 | -0.003625 |                   0.0 |                       0.0 |                   0.704039 |                  4.014396 |                 0.226273 |    -0.167266 |        -0.226273 |         -3.310357 |             62.648725 |                 100.0 |               112.102145 |               98.814229 | 0.234493 | 0.234493 |  0.001835 |  0.001835 |      0.0 | -0.010648 | -0.001813 |    -7.924586 |     0.754819 |      0.0 |      0.0 | 0.319079 | 0.319079 | -0.028403 | -0.096736 |  0.101134 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.058165 | 19.132901 | 20.466014 |       0.0 | 0.854914 | -0.707632 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                     0.310511 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.358774 |  0.002699 |                   0.0 |                       0.0 |                   0.705952 |                  4.024777 |                 0.228638 |    -0.167699 |        -0.228638 |         -3.318825 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 | -0.004543 | -0.004543 |      0.0 | -0.010457 |       0.0 |    -8.921564 |     0.754153 |      0.0 |      0.0 | 0.338868 | 0.338868 | -0.027895 | -0.084862 |  0.115881 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062708 | 19.481718 | 20.476472 |       0.0 | 0.862222 | -0.679431 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                     0.289025 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.264771 |  0.008821 |                   0.0 |                       0.0 |                    0.70779 |                  4.040774 |                 0.231601 |    -0.168366 |        -0.231601 |         -3.332984 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.002708 | 0.002708 | -0.004605 | -0.004605 |      0.0 | -0.008842 | -0.000032 |    -9.827517 |     0.753513 |      0.0 |      0.0 | 0.359251 | 0.359251 | -0.023586 | -0.079236 |  0.128779 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.064605 | 19.852245 | 20.488022 |       0.0 | 0.869294 | -0.657352 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                     0.283863 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.191174 |  0.011559 |                   0.0 |                       0.0 |                   0.709571 |                  4.052814 |                   0.2377 |    -0.168867 |          -0.2377 |         -3.343243 |             71.001889 |                 100.0 |               112.102145 |               98.814229 | 0.004605 | 0.004605 | -0.004915 | -0.004915 |      0.0 | -0.009247 |  -0.00006 |   -10.054055 |     0.752893 |      0.0 |      0.0 | 0.360481 | 0.360481 | -0.024667 | -0.081134 |    0.1319 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.064915 | 20.226454 | 20.501874 |       0.0 | 0.876199 | -0.638011 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                     0.261711 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.126703 |  0.017864 |                   0.0 |                       0.0 |                    0.71133 |                  4.064535 |                  0.23996 |    -0.169356 |         -0.23996 |         -3.353205 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.004915 | 0.004915 | -0.008645 | -0.008645 |      0.0 | -0.008703 | -0.000075 |   -11.069779 |      0.75228 |      0.0 |      0.0 | 0.384132 |  0.27542 | -0.023214 | -0.070657 |  0.146089 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.068645 | 20.515492 | 20.515492 |  -0.03631 | 0.883071 | -0.624975 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                     0.239222 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.083248 |  0.033745 |                   0.0 |                       0.0 |                   0.712546 |                  4.080005 |                 0.243122 |        -0.17 |        -0.243122 |          -3.36746 |             92.611159 |                 100.0 |               112.102145 |               98.814229 | 0.008645 | 0.008645 | -0.007726 | -0.007726 |      0.0 | -0.007603 |  -0.00015 |   -12.182266 |     0.751857 |      0.0 |      0.0 | 0.468192 |      0.0 | -0.020281 |  -0.06918 |  0.153661 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.067726 |  20.53174 |  20.53174 | -0.156376 | 0.887849 |  -0.62782 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                     0.225573 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.092732 |  0.068385 |                   0.0 |                       0.0 |                   0.713188 |                  4.093695 |                 0.234639 |    -0.170571 |        -0.234639 |         -3.380507 |             101.43127 |                 100.0 |               112.102145 |               98.814229 | 0.007726 | 0.007726 | -0.010331 | -0.010331 |      0.0 | -0.009114 | -0.000139 |   -12.903761 |     0.751633 |      0.0 |      0.0 | 0.675874 |      0.0 | -0.024311 | -0.072717 |  0.137611 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.070331 |  20.54858 |  20.54858 | -0.225742 | 0.890384 | -0.665304 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                     0.217835 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -2.21768 |  0.086502 |                   0.0 |                       0.0 |                   0.656507 |                  3.682358 |                  0.21558 |    -0.153432 |         -0.21558 |         -3.025851 |            152.146905 |                 100.0 |               112.102145 |               98.814229 | 0.010331 | 0.010331 | -0.008377 | -0.008377 |      0.0 | -0.008113 | -0.000184 |   -13.330161 |     0.751474 |      0.0 |      0.0 | 0.800491 |      0.0 | -0.021641 | -0.065999 |   0.12794 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.068377 | 20.567024 | 20.567024 | -0.267364 | 0.892185 | -0.720906 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                     0.206114 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 | -2.403019 |  0.097533 |              0.000992 |                  0.000992 |                   0.568958 |                  3.039527 |                 0.201303 |    -0.125655 |        -0.200311 |         -2.470569 |            193.641515 |                 100.0 |               112.102145 |               99.616909 | 0.008377 | 0.008377 | -0.005975 | -0.005975 |      0.0 | -0.006565 | -0.000166 |   -14.002557 |     0.751319 |      0.0 |      0.0 | 0.890752 |      0.0 | -0.017511 | -0.055286 |  0.127514 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.065975 | 20.581966 | 20.581966 | -0.297511 | 0.893946 | -0.787539 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                     0.221441 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 |  -2.62513 |  0.097703 |              0.016719 |                  0.016719 |                   0.483954 |                  2.412046 |                 0.184344 |    -0.083783 |        -0.167625 |         -1.928092 |            213.005666 |                 100.0 |               112.102145 |              101.107639 | 0.005975 | 0.005975 | -0.001951 | -0.001951 |      0.0 | -0.005069 | -0.000115 |   -13.129818 |     0.751158 |      0.0 |      0.0 | 0.920922 |      0.0 |  -0.01352 | -0.046439 |  0.107665 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061951 | 20.593009 | 20.593009 | -0.307588 | 0.895788 | -0.854342 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                      0.29413 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -2.847807 |  0.080161 |              0.050445 |                  0.050445 |                   0.404168 |                  1.815487 |                 0.156128 |      -0.0252 |        -0.105683 |          -1.41132 |            152.146905 |                 100.0 |               112.102145 |              101.107639 | 0.001951 | 0.001951 |  0.002352 |  0.002352 |      0.0 | -0.005185 | -0.000028 |    -9.606973 |     0.750969 |      0.0 |      0.0 | 0.845269 |      0.0 | -0.013831 | -0.036988 |  0.054864 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.057648 | 20.600146 | 20.600146 |  -0.28232 | 0.897949 | -0.903603 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                      0.39994 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -3.012009 |  0.052367 |               0.09616 |                   0.09616 |                   0.361317 |                  1.493002 |                 0.132043 |     0.033952 |        -0.035883 |         -1.131685 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.008336 |  0.008336 |      0.0 | -0.003673 |       0.0 |     -5.66461 |     0.750755 |      0.0 |      0.0 |  0.70078 |      0.0 | -0.009799 | -0.021977 |  0.004108 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.049313 | 20.603819 | 20.603819 | -0.234061 | 0.900392 |  -0.92507 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                     0.432548 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -3.083565 |   0.03377 |              0.082154 |                  0.082154 |                   0.339595 |                  1.331032 |                 0.114092 |     0.026694 |        -0.031938 |         -0.991437 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.010305 |  0.010305 |      0.0 | -0.002951 |       0.0 |    -4.636808 |     0.750522 |      0.0 |      0.0 | 0.599126 |      0.0 | -0.007871 |  -0.02425 | -0.000183 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.039007 |  20.60677 |  20.60677 | -0.200108 |  0.90307 | -0.927939 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                     0.474822 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -3.093131 |   0.01893 |              0.072349 |                  0.072349 |                   0.336403 |                  1.306141 |                 0.099095 |     0.017927 |        -0.026746 |         -0.969738 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.010759 |  0.010759 |      0.0 | -0.003602 |       0.0 |    -3.401939 |     0.750272 |      0.0 |      0.0 | 0.511858 |      0.0 | -0.009608 | -0.027355 | -0.010217 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.028248 | 20.610372 | 20.610372 | -0.170961 | 0.905952 |  -0.91597 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                     0.498466 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -3.053232 |  0.008893 |              0.038808 |                  0.038808 |                   0.336764 |                  1.307301 |                 0.085168 |    -0.015663 |         -0.04636 |         -0.970537 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.010389 |  0.010389 |      0.0 | -0.006464 |       0.0 |    -2.753097 |     0.750003 |      0.0 |      0.0 |  0.44687 |      0.0 | -0.017242 | -0.041931 | -0.012813 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.017859 | 20.616836 | 20.616836 | -0.149255 | 0.909051 | -0.893963 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                     0.522365 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -2.979875 |  0.001425 |              0.009299 |                  0.009299 |                   0.337148 |                  1.307629 |                 0.071637 |    -0.045186 |        -0.062338 |         -0.970481 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.007083 |  0.007083 |      0.0 | -0.009785 |       0.0 |    -2.124364 |     0.749719 |      0.0 |      0.0 | 0.392943 |      0.0 | -0.026102 | -0.052823 | -0.016587 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.010775 | 20.626621 | 20.626621 | -0.131243 | 0.912347 | -0.864488 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                     0.542712 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 | -2.881625 | -0.004308 |              0.000025 |                  0.000025 |                   0.337435 |                   1.30686 |                 0.066594 |    -0.054427 |        -0.066568 |         -0.969425 |             68.711505 |                 100.0 |               112.102145 |               98.953162 | 0.044775 | 0.011194 |  0.005889 |  0.005889 |      0.0 | -0.011159 |  0.000164 |    -1.608792 |     0.749506 |      0.0 |      0.0 | 0.347505 | 0.033582 | -0.029765 | -0.056124 | -0.019321 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.054111 | 20.682555 | 20.682555 | -0.104851 | 0.914828 | -0.829279 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                     0.553889 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.764264 | -0.010537 |                   0.0 |                       0.0 |                   0.337881 |                  1.305996 |                  0.06888 |    -0.054417 |         -0.06888 |         -0.968115 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 |  0.007388 |  0.007388 |      0.0 | -0.010885 |       0.0 |    -1.332814 |     0.749175 |      0.0 |      0.0 | 0.294618 |      0.0 | -0.029035 | -0.056831 | -0.016985 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.046723 |  20.69344 |  20.69344 | -0.098402 |  0.91869 | -0.787842 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                     0.565716 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -2.62614 | -0.012158 |                   0.0 |                       0.0 |                   0.337633 |                  1.303739 |                 0.072235 |    -0.054322 |        -0.072235 |         -0.966106 |             48.410379 |                 100.0 |               112.102145 |               98.814229 | 0.268723 | 0.067181 |  0.003845 |  0.003845 |      0.0 | -0.012838 |  0.000984 |    -1.046088 |     0.749359 |      0.0 |      0.0 | 0.273253 | 0.201542 | -0.034246 | -0.056299 |  -0.01831 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.056155 | 20.975001 | 20.975001 | -0.023951 | 0.916542 | -0.744784 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                     0.589936 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.482615 | -0.031773 |                   0.0 |                       0.0 |                   0.338048 |                  1.301296 |                 0.069234 |    -0.054221 |        -0.069234 |         -0.963247 |             56.054123 |                 100.0 |               112.102145 |               98.814229 | 0.184155 | 0.027623 |  0.000004 |  0.000004 |      0.0 | -0.012181 |  0.000998 |    -0.475078 |      0.74905 |      0.0 |      0.0 | 0.134936 | 0.134936 | -0.032493 | -0.046399 | -0.009658 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.059996 | 21.149729 | 21.171338 |       0.0 | 0.920143 | -0.682111 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                      0.61078 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.273703 | -0.040969 |                   0.0 |                       0.0 |                   0.338492 |                  1.298642 |                 0.071334 |     -0.05411 |        -0.071334 |          -0.96015 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 0.375996 |   0.0376 |  0.000346 |  0.000346 |      0.0 | -0.014629 |  0.002393 |          0.0 |     0.748721 |      0.0 |      0.0 |  0.05168 |  0.05168 | -0.039021 | -0.053629 | -0.021316 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.059654 | 21.253428 | 21.561963 |       0.0 | 0.924004 | -0.610242 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                      0.61078 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.034139 | -0.043937 |                   0.0 |                       0.0 |                   0.339329 |                  1.295068 |                 0.059541 |    -0.053961 |        -0.059541 |         -0.955738 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 0.657654 |      0.0 | -0.002532 | -0.002532 |      0.0 | -0.020116 |  0.005507 |          0.0 |     0.748099 |  0.06527 |  0.06527 |      0.0 |      0.0 |  -0.05366 | -0.066112 | -0.060231 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062532 | 21.207341 | 22.239733 |       0.0 | 0.931339 | -0.535404 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.784681 | -0.038549 |                   0.0 |                       0.0 |                    0.34016 |                  1.290674 |                 0.049365 |    -0.053778 |        -0.049365 |         -0.950513 |             36.725115 |                 100.0 |               112.102145 |               98.814229 | 0.472532 |      0.0 | -0.001838 | -0.001838 |      0.0 | -0.024931 |  0.004748 |          0.0 |     0.747483 | 0.194277 | 0.194277 |      0.0 |      0.0 | -0.066503 | -0.081551 |  -0.09869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061838 |  21.03613 | 22.737196 |       0.0 |  0.93867 | -0.465955 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.553184 | -0.033549 |                   0.0 |                       0.0 |                   0.340987 |                   1.28612 |                 0.047044 |    -0.053588 |        -0.047044 |         -0.945133 |             34.355753 |                 100.0 |               112.102145 |               98.814229 | 0.941838 |      0.0 | -0.000987 | -0.000987 |      0.0 | -0.027164 |  0.009858 |          0.0 |     0.746869 | 0.273324 | 0.273324 |      0.0 |      0.0 | -0.072459 | -0.089566 | -0.114981 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.060987 | 20.786625 | 23.706198 |       0.0 | 0.946021 | -0.401506 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.338354 | -0.028908 |                   0.0 |                       0.0 |                   0.339465 |                  1.277745 |                 0.041776 |    -0.053239 |        -0.041776 |          -0.93828 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 1.880987 |      0.0 | -0.001175 | -0.001175 |      0.0 | -0.027338 |  0.021263 |          0.0 |     0.746256 | 0.337589 | 0.337589 |      0.0 |      0.0 | -0.072924 | -0.089252 |   -0.1204 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.061175 | 20.470885 | 25.614524 |       0.0 | 0.953411 | -0.341698 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.138993 | -0.024602 |                   0.0 |                       0.0 |                   0.340279 |                  1.268373 |                 0.027852 |    -0.052849 |        -0.027852 |         -0.928095 |             40.962628 |                 100.0 |               112.102145 |               98.814229 | 1.975175 |      0.0 | -0.002517 | -0.002517 |      0.0 | -0.031933 |  0.026463 |          0.0 |     0.745648 | 0.476222 | 0.476222 |      0.0 |      0.0 |  -0.08518 |  -0.09987 | -0.157197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.062517 | 20.017805 | 27.621633 |       0.0 | 0.960808 | -0.286196 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.953985 | -0.020606 |                   0.0 |                       0.0 |                   0.341082 |                  1.257174 |                 0.008709 |    -0.052382 |        -0.008709 |         -0.916092 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 1.694517 |      0.0 | -0.003094 | -0.003094 |      0.0 | -0.043255 |  0.016391 |          0.0 |     0.745047 | 0.698133 | 0.698133 |      0.0 |      0.0 | -0.115381 |  -0.13072 | -0.237391 |      1.0 |      0.3 | 1.690527 | 0.577692 |  0.0 |  0.0 | 0.0 |  0.0 | 0.577692 | 0.010414 | 0.002893 | 0.063094 | 19.348866 | 27.668878 |       0.0 | 0.968167 |  -0.23469 | 72.559096 |      0.5 | 0.077692 |   0.0 |   0.0 |  0.0 | 0.006486 | 0.003928 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002893 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.782298 | -0.016898 |                   0.0 |                       0.0 |                   0.341875 |                  1.245134 |                -0.004886 |    -0.051881 |         0.004886 |          -0.90326 |             32.273586 |                 100.0 |                110.51162 |               98.814229 | 1.225094 |      0.0 | -0.002985 | -0.002985 |      0.0 | -0.055801 |       0.0 |          0.0 |     0.744455 | 0.905983 | 0.905983 |      0.0 |      0.0 | -0.148847 | -0.165762 | -0.319496 |      1.0 |      0.3 | 2.523263 | 0.864696 |  0.0 |  0.0 | 0.0 |  0.0 | 0.864696 | 0.045004 | 0.012501 | 0.062985 | 18.480078 | 26.426511 |       0.0 | 0.975475 | -0.186892 | 74.217662 |      0.5 | 0.364696 |   0.0 |   0.0 |  0.0 | 0.019234 | 0.025771 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012501 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.622973 | -0.013456 |                   0.0 |                       0.0 |                   0.342661 |                  1.232336 |                -0.014528 |    -0.051347 |         0.014528 |         -0.889676 |             30.429381 |                 100.0 |               108.159556 |               98.814229 | 0.472985 |      0.0 |  0.000597 |  0.000597 |      0.0 |  -0.06294 |       0.0 |          0.0 |     0.743868 | 1.073828 | 1.073828 |      0.0 |      0.0 |  -0.16789 | -0.189697 | -0.372115 |      1.0 |      0.3 | 2.012216 | 0.691408 |  0.0 |  0.0 | 0.0 |  0.0 | 0.691408 | 0.082347 | 0.022874 | 0.059403 | 17.447706 |  24.95022 |       0.0 | 0.982783 | -0.142536 |  75.53847 |      0.5 | 0.191408 |   0.0 |   0.0 |  0.0 | 0.031652 | 0.050694 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022874 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.475119 | -0.010263 |                   0.0 |                       0.0 |                   0.343441 |                  1.218349 |                -0.012189 |    -0.050765 |         0.012189 |         -0.874907 |             29.178858 |                 100.0 |                106.30129 |               98.814229 | 0.187403 |      0.0 |  0.007215 |  0.007215 |      0.0 | -0.061894 |       0.0 |          0.0 |     0.743284 | 1.087936 | 1.087936 |      0.0 |      0.0 | -0.165098 | -0.196345 | -0.373633 |      1.0 |      0.3 |  1.74701 | 0.601638 |  0.0 |  0.0 | 0.0 |  0.0 | 0.601638 | 0.103891 | 0.028859 | 0.052785 | 16.400354 | 23.452506 |       0.0 | 0.990091 | -0.101373 | 76.683843 |      0.5 | 0.101638 |   0.0 |   0.0 |  0.0 |  0.04375 | 0.060141 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028859 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -0.33791 | -0.007299 |                   0.0 |                       0.0 |                   0.344216 |                  1.204312 |                -0.020102 |     -0.05018 |         0.020102 |         -0.860096 |              29.58412 |                 100.0 |                104.69982 |               98.814229 | 0.462785 |      0.0 |  0.008179 |  0.008179 |      0.0 |  -0.06548 |       0.0 |          0.0 |     0.742705 | 1.178833 | 1.178833 |      0.0 |      0.0 | -0.174665 | -0.206262 | -0.401029 |      1.0 |      0.3 | 2.153223 | 0.743402 |  0.0 |  0.0 | 0.0 |  0.0 | 0.743402 | 0.126968 | 0.035269 | 0.051821 |  15.26402 | 21.827548 |       0.0 | 0.997399 | -0.063174 | 78.093663 |      0.5 | 0.243402 |   0.0 |   0.0 |  0.0 | 0.055536 | 0.071433 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035269 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.210581 | -0.004549 |                   0.0 |                       0.0 |                   0.344985 |                  1.190153 |                -0.018227 |     -0.04959 |         0.018227 |         -0.845168 |             28.027061 |                 100.0 |               102.740595 |               98.814229 | 0.179821 |      0.0 |  0.014145 |  0.014145 |      0.0 | -0.065772 |       0.0 |          0.0 |      0.74213 | 1.211988 | 1.211988 |      0.0 |      0.0 | -0.175444 | -0.215681 | -0.409353 |      1.0 |      0.3 | 1.918187 | 0.663887 |  0.0 |  0.0 | 0.0 |  0.0 | 0.663887 | 0.151521 | 0.042089 | 0.045855 | 14.094374 | 20.154955 |       0.0 | 1.004707 | -0.027726 | 79.347964 |      0.5 | 0.163887 |   0.0 |   0.0 |  0.0 | 0.067016 | 0.084504 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042089 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 | -0.092419 | -0.001996 |              0.000838 |                  0.000838 |                   0.345749 |                  1.174779 |                -0.024601 |    -0.048111 |         0.025439 |          -0.82903 |             29.178858 |                 100.0 |               101.008015 |               99.257154 | 0.079855 |      0.0 |  0.019009 |  0.019009 |      0.0 | -0.065786 |       0.0 |          0.0 |     0.741559 | 1.256923 | 1.256923 |      0.0 |      0.0 | -0.175481 | -0.220889 | -0.421809 | 0.999806 | 0.299884 | 1.883121 | 0.653078 |  0.0 |  0.0 | 0.0 |  0.0 | 0.653078 | 0.170048 | 0.047236 | 0.040991 | 12.879352 | 18.417474 |       0.0 | 1.012015 |  0.005171 | 80.578007 |      0.5 | 0.153078 |   0.0 |   0.0 |  0.0 | 0.078201 | 0.091847 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047236 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 |       0.0 |       0.0 |              0.016101 |                  0.016101 |                   0.346507 |                   1.15845 |                -0.023742 |    -0.032168 |         0.039843 |         -0.811944 |             30.000798 |                 100.0 |                99.318029 |               99.930369 | 0.262991 |      0.0 |  0.016396 |  0.016396 |      0.0 | -0.065321 |       0.0 |          0.0 |     0.740993 | 1.275415 | 1.275415 |      0.0 |      0.0 | -0.174242 | -0.211904 | -0.425989 |  0.99865 | 0.299191 | 2.093303 | 0.726452 |  0.0 |  0.0 | 0.0 |  0.0 | 0.726452 | 0.190739 | 0.052983 | 0.043604 | 11.645093 | 16.652483 |       0.0 | 1.019323 |  0.036071 | 81.944858 |      0.5 | 0.226452 |   0.0 |   0.0 |  0.0 | 0.089096 | 0.101643 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052983 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 |       0.0 |       0.0 |              0.049169 |                  0.049169 |                   0.347259 |                  1.144953 |                 -0.02789 |     0.001462 |         0.077058 |         -0.797694 |             30.000798 |                 100.0 |                97.450072 |               99.735174 | 0.077604 |      0.0 |   0.02648 |   0.02648 |      0.0 | -0.066592 |       0.0 |          0.0 |      0.74043 | 1.420927 | 1.420927 |      0.0 |      0.0 | -0.177631 |   -0.2199 |  -0.47459 | 0.997494 | 0.298498 | 2.117655 | 0.735475 |  0.0 |  0.0 | 0.0 |  0.0 | 0.735475 | 0.214214 | 0.059504 |  0.03352 | 10.265052 | 14.679024 |       0.0 | 1.026631 |  0.066971 | 83.327039 |      0.5 | 0.235475 |   0.0 |   0.0 |  0.0 |  0.09971 | 0.114504 |   0.0 |   0.0 |  0.0 |   0.0 | 0.059504 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 |       0.0 |       0.0 |              0.075044 |                  0.075044 |                   0.338849 |                  1.076886 |                -0.043295 |     0.030174 |         0.118338 |         -0.738037 |             23.933221 |                 100.0 |                95.571281 |               99.531964 |  0.06752 |      0.0 |  0.060126 |      0.06 |      0.0 | -0.080429 |       0.0 |          0.0 |     0.739872 | 1.878635 | 1.878635 |      0.0 |      0.0 |  -0.21454 | -0.294586 | -0.627464 | 0.996337 | 0.297806 | 2.768687 | 0.962918 |  0.0 |  0.0 | 0.0 |  0.0 | 0.962918 | 0.248077 |  0.06891 |      0.0 |  8.432367 | 12.058285 |       0.0 | 1.033939 |  0.097871 | 85.132809 |      0.5 | 0.462918 |   0.0 |   0.0 |  0.0 |  0.11005 | 0.138027 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06891 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 |       0.0 |       0.0 |              0.075336 |                  0.075336 |                   0.339571 |                  1.067855 |                -0.048721 |     0.030842 |         0.124057 |         -0.728284 |             22.421649 |                 100.0 |                93.131102 |               99.257118 |      0.0 |      0.0 |  0.073997 |       0.0 |      0.0 | -0.085956 |       0.0 |          0.0 |     0.739317 | 2.036362 | 2.036362 |      0.0 |      0.0 | -0.229285 | -0.326804 | -0.680145 | 0.995181 | 0.297115 | 2.932756 |  1.02181 |  0.0 |  0.0 | 0.0 |  0.0 |  1.02181 | 0.293418 | 0.081505 |      0.0 |  6.441599 |  9.211486 |       0.0 | 1.041247 |  0.128771 | 87.043755 |      0.5 |  0.52181 |   0.0 |   0.0 |  0.0 | 0.120123 | 0.173295 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081505 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 |       0.0 |       0.0 |              0.071319 |                  0.071319 |                   0.340289 |                   1.06128 |                -0.040462 |     0.027099 |         0.111781 |         -0.720991 |             26.962743 |                 100.0 |                 90.56538 |               98.953929 |    0.034 |      0.0 |  0.085441 |      0.06 |      0.0 | -0.061599 |       0.0 |          0.0 |     0.738767 | 1.663522 | 1.663522 |      0.0 |      0.0 | -0.164312 | -0.279524 | -0.555616 | 0.994024 | 0.296425 | 2.428744 | 0.847411 |  0.0 |  0.0 | 0.0 |  0.0 | 0.847411 | 0.329051 | 0.091403 |      0.0 |  4.810029 |  6.878341 |       0.0 | 1.048555 |  0.159671 | 88.625087 |      0.5 | 0.347411 |   0.0 |   0.0 |  0.0 | 0.129935 | 0.199115 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091403 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 |       0.0 |       0.0 |              0.039796 |                  0.039796 |                   0.341001 |                  1.056546 |                -0.036448 |    -0.004227 |         0.076244 |         -0.715546 |              22.18809 |                 100.0 |                88.454237 |               98.692761 |      0.0 |      0.0 |  0.116315 |       0.0 |      0.0 | -0.063414 |       0.0 |          0.0 |      0.73822 | 1.737454 | 1.737454 |      0.0 |      0.0 | -0.169154 | -0.334911 |  -0.58031 | 0.992868 | 0.295736 | 2.507465 | 0.876046 |  0.0 |  0.0 | 0.0 |  0.0 | 0.876046 | 0.354797 | 0.098555 |      0.0 |  3.100902 |   4.43429 |       0.0 | 1.055863 |  0.190571 | 90.256507 |      0.5 | 0.376046 |   0.0 |   0.0 |  0.0 | 0.139494 | 0.215302 |   0.0 |   0.0 |  0.0 |   0.0 | 0.098555 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 |       0.0 |       0.0 |              0.011912 |                  0.011912 |                   0.350942 |                  1.110113 |                -0.025663 |    -0.034343 |         0.037575 |         -0.759171 |             23.407216 |                 100.0 |                86.286908 |               98.412886 |    0.034 |      0.0 |  0.102965 |      0.06 |      0.0 | -0.057639 |       0.0 |          0.0 |     0.737678 | 1.501159 | 1.501159 |      0.0 |      0.0 |  -0.15375 | -0.310062 | -0.501387 | 0.991711 | 0.295047 |  2.20856 | 0.772475 |  0.0 |  0.0 | 0.0 |  0.0 | 0.772475 | 0.374848 | 0.104124 |      0.0 |  1.620538 |  2.317369 |       0.0 | 1.063171 |  0.221471 | 91.692592 |      0.5 | 0.272475 |   0.0 |   0.0 |  0.0 | 0.148807 | 0.226042 |   0.0 |   0.0 |  0.0 |   0.0 | 0.104124 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 |       0.0 |       0.0 |              0.000021 |                  0.000021 |                   0.351663 |                  1.107879 |                -0.014096 |    -0.046141 |         0.014117 |         -0.756217 |              24.48341 |                 100.0 |                84.387546 |               98.778718 |      0.0 |      0.0 |  0.096543 |       0.0 |      0.0 | -0.047871 |       0.0 |          0.0 |     0.737139 | 1.261344 | 1.261344 |      0.0 |      0.0 | -0.127693 | -0.279479 | -0.421289 | 0.990555 |  0.29436 | 1.840982 | 0.644352 |  0.0 |  0.0 | 0.0 |  0.0 | 0.644352 | 0.382057 | 0.106127 |      0.0 |  0.366614 |  0.524258 |       0.0 | 1.070479 |  0.252371 | 92.889221 |      0.5 | 0.144352 |   0.0 |   0.0 |  0.0 | 0.157878 | 0.224178 |   0.0 |   0.0 |  0.0 |   0.0 | 0.106127 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.352379 |                   1.10789 |                 0.008426 |    -0.046162 |        -0.008426 |         -0.755511 |             23.933221 |                 100.0 |                 82.81063 |               98.814229 |      0.0 |      0.0 |  0.129358 |       0.0 | 0.040511 | -0.025712 |       0.0 |          0.0 |     0.736604 | 0.975359 | 0.366614 |      0.0 |      0.0 | -0.068585 |  -0.26561 |  -0.32577 | 0.989399 | 0.293673 | 0.549969 | 0.192373 |  0.0 |  0.0 | 0.0 |  0.0 | 0.192373 | 0.371671 | 0.103242 |      0.0 |       0.0 |       0.0 |       0.0 | 1.077787 |  0.283271 | 93.206307 | 0.192373 |      0.0 |   0.0 |   0.0 |  0.0 | 0.162725 | 0.208946 |   0.0 |   0.0 |  0.0 |   0.0 | 0.103242 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                          0.0 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.11016 |                   0.0 |                       0.0 |                   1.204044 |                  1.107494 |                      0.0 |    -0.046146 |              0.0 |           0.09655 |             26.962743 |                 100.0 |                82.393619 |               98.814229 |      0.0 |      0.0 |  0.087608 |       0.0 | 0.029418 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.984119 | 0.290547 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.349175 | 0.096993 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.424331 | 93.176888 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160995 | 0.188179 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096993 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                          0.0 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10368 |                   0.0 |                       0.0 |                   1.204044 |                  1.107885 |                      0.0 |    -0.046162 |              0.0 |          0.096159 |             30.000798 |                 100.0 |                82.432294 |               98.814229 |      0.0 |      0.0 |  0.070419 |       0.0 | 0.025143 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.979083 | 0.287581 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.326313 | 0.090643 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.558911 | 93.151745 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.156837 | 0.169476 |   0.0 |   0.0 |  0.0 |   0.0 | 0.090643 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.108737 |                      0.0 |    -0.046197 |              0.0 |          0.095308 |             33.282135 |                 100.0 |                 82.46535 |               98.814229 |      0.0 |      0.0 |  0.066191 |       0.0 | 0.025199 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.973965 | 0.284582 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.305418 | 0.084838 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.695651 | 93.126546 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.152786 | 0.152632 |   0.0 |   0.0 |  0.0 |   0.0 | 0.084838 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.109465 |                      0.0 |    -0.046228 |              0.0 |          0.094579 |              29.58412 |                 100.0 |                82.498483 |               98.814229 |      0.0 |      0.0 |  0.060894 |       0.0 | 0.021594 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.968848 |   0.2816 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.286301 | 0.079528 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.832391 | 93.104952 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.148839 | 0.137462 |   0.0 |   0.0 |  0.0 |   0.0 | 0.079528 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10152 |                   0.0 |                       0.0 |                   1.204044 |                   1.11033 |                      0.0 |    -0.046264 |              0.0 |          0.093714 |             30.429381 |                 100.0 |                82.526876 |               98.814229 |    0.034 |      0.0 |  0.050931 |  0.050931 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.963892 | 0.278726 |    0.034 | 0.011425 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011425 | 0.268943 | 0.074706 | 0.009069 |       0.0 |       0.0 |       0.0 |      nan |  0.964811 | 93.127528 | 0.011425 |      0.0 |   0.0 |   0.0 |  0.0 | 0.145143 |   0.1238 |   0.0 |   0.0 |  0.0 |   0.0 | 0.074706 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08208 |                   0.0 |                       0.0 |                   1.204044 |                  1.112997 |                      0.0 |    -0.046375 |              0.0 |          0.091047 |             31.324363 |                 100.0 |                82.497192 |               98.814229 | 0.043069 |      0.0 |  0.055782 |  0.055782 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.959664 | 0.276286 | 0.043069 | 0.014376 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014376 | 0.253223 |  0.07034 | 0.004218 |       0.0 |       0.0 |       0.0 |      nan |  1.077791 | 93.156221 | 0.014376 |      0.0 |   0.0 |   0.0 |  0.0 | 0.141727 | 0.111496 |   0.0 |   0.0 |  0.0 |   0.0 |  0.07034 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08424 |                   0.0 |                       0.0 |                   1.196665 |                  1.116023 |                      0.0 |    -0.046501 |              0.0 |          0.080642 |             33.282135 |                 100.0 |                82.459466 |               98.814229 |      0.0 |      0.0 |  0.053849 |  0.004218 | 0.018581 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.955354 | 0.273811 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.238665 | 0.066296 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.192931 | 93.137639 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.138251 | 0.100414 |   0.0 |   0.0 |  0.0 |   0.0 | 0.066296 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07992 |                   0.0 |                       0.0 |                   1.196665 |                  1.119656 |                      0.0 |    -0.046652 |              0.0 |          0.077009 |             33.810423 |                 100.0 |                82.483896 |               98.814229 |      0.0 |      0.0 |  0.049163 |       0.0 | 0.018518 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.951207 | 0.271438 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.225114 | 0.062532 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.303751 | 93.119121 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.13468 | 0.090434 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062532 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0756 |                   0.0 |                       0.0 |                   1.196665 |                  1.123475 |                      0.0 |    -0.046811 |              0.0 |           0.07319 |             37.369415 |                 100.0 |                82.508245 |               98.814229 |    0.034 |      0.0 |  0.033548 |  0.033548 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.947221 | 0.269168 |    0.034 | 0.011126 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011126 | 0.212791 | 0.059109 | 0.026452 |       0.0 |       0.0 |       0.0 |      nan |  1.410251 | 93.141996 | 0.011126 |      0.0 |   0.0 |   0.0 |  0.0 | 0.131345 | 0.081446 |   0.0 |   0.0 |  0.0 |   0.0 | 0.059109 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.127474 |                      0.0 |    -0.046978 |              0.0 |           0.06919 |             40.962628 |                 100.0 |                82.478169 |               98.814229 |      0.0 |      0.0 |  0.023802 |  0.023802 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.943397 | 0.266999 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.201447 | 0.055957 |  0.00265 |       0.0 |       0.0 |       0.0 |      nan |  1.512431 | 93.141996 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.128095 | 0.073351 |   0.0 |   0.0 |  0.0 |   0.0 | 0.055957 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.131856 |                      0.0 |    -0.047161 |              0.0 |          0.064808 |              46.30558 |                 100.0 |                82.478169 |               98.814229 |      0.0 |      0.0 |  0.022296 |   0.00265 | 0.008765 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.939573 | 0.264839 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.190848 | 0.053013 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.614611 |  93.13323 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124787 | 0.066061 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053013 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0648 |                   0.0 |                       0.0 |                   1.196665 |                  1.136687 |                      0.0 |    -0.047362 |              0.0 |          0.059978 |             43.470544 |                 100.0 |                82.489694 |               98.814229 |      0.0 |      0.0 |  0.028418 |       0.0 |  0.01221 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.935991 | 0.262824 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.181059 | 0.050294 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.710311 |  93.12102 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121563 | 0.059495 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050294 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.04104 |                   0.0 |                       0.0 |                   1.196665 |                  1.144252 |                      0.0 |    -0.047677 |              0.0 |          0.052412 |             33.810423 |                 100.0 |                82.505748 |               98.814229 |      0.0 |      0.0 |  0.047697 |       0.0 | 0.017452 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.933299 | 0.261314 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.172005 | 0.047779 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.782251 | 93.103568 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.118423 | 0.053582 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047779 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.02592 |                   0.0 |                       0.0 |                   1.196665 |                  1.152031 |                      0.0 |    -0.048001 |              0.0 |          0.044634 |             36.725115 |                 100.0 |                82.528697 |               98.814229 |      0.0 |      0.0 |  0.040381 |       0.0 | 0.015359 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.931173 | 0.260125 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.163621 |  0.04545 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.839071 | 93.088209 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115364 | 0.048257 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04545 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 |  -0.01728 |              0.002897 |                  0.002897 |                   1.196665 |                  1.158654 |                      0.0 |     -0.04538 |              0.0 |          0.038011 |             34.918962 |                 100.0 |                82.548893 |               98.501655 |      0.0 |      0.0 |   0.02629 |       0.0 | 0.009662 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.929369 | 0.259118 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.155845 |  0.04329 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.887251 | 93.078547 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.112384 | 0.043461 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04329 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 |  -0.01728 |              0.057389 |                  0.057389 |                   1.196665 |                  1.166643 |                      0.0 |     0.008779 |              0.0 |          0.030022 |             36.102655 |                 100.0 |                82.561599 |                97.90162 |      0.0 |      0.0 |  0.045635 |       0.0 | 0.017178 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.927566 | 0.258114 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.148622 | 0.041284 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.935431 | 93.061369 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109481 | 0.039141 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041284 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                     0.548474 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |  0.014361 |              0.020003 |                  0.020003 |                   0.136052 |                  1.175846 |                 0.098108 |     -0.02899 |        -0.078105 |         -1.039794 |             33.282135 |                 100.0 |                82.584188 |               97.904842 |    0.034 |  0.01275 |  0.052426 |  0.052426 |      0.0 |  0.003209 |  0.000025 |    -1.465907 |     0.897677 |      0.0 |      0.0 | 0.000099 | 0.000099 |  0.008561 | -0.072247 |  0.014419 | 0.926947 | 0.257769 | 0.015768 | 0.004993 |  0.0 |  0.0 | 0.0 |  0.0 | 0.004993 | 0.141969 | 0.039436 | 0.007574 |  0.010587 |  0.015023 |       0.0 |  0.00743 |   1.95197 | 93.072144 | 0.004993 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106718 | 0.035251 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039436 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                     0.526367 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 |  0.019828 |              0.030684 |                  0.030684 |                   0.139046 |                   1.18712 |                 0.107855 |    -0.018779 |        -0.077171 |         -1.048074 |             31.324363 |                 100.0 |                 82.57002 |               97.902821 |      0.0 |      0.0 |  0.084311 |  0.007574 |      0.0 |  0.012109 |       0.0 |    -2.021578 |     0.895425 |      0.0 |      0.0 | 0.000112 | 0.000112 |  0.032299 | -0.089605 |  0.019865 | 0.926533 | 0.257539 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.135773 | 0.037715 |      0.0 |  0.002076 |  0.002914 |       0.0 | 0.014738 |  1.963042 | 93.072144 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.104026 | 0.031747 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037715 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                     0.507806 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |  0.024583 |              0.031397 |                  0.031397 |                   0.141997 |                  1.199167 |                 0.113262 |    -0.018568 |        -0.081865 |          -1.05717 |             30.429381 |                 100.0 |                 82.57002 |               97.902821 |      0.0 |      0.0 |  0.092017 |       0.0 | 0.030724 |  0.002914 |       0.0 |     -2.50427 |     0.893205 |      0.0 |      0.0 | 0.000027 | 0.000027 |  0.037744 | -0.095017 |  0.024591 | 0.926296 | 0.257408 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.129931 | 0.036092 |      0.0 |       0.0 |       0.0 |       0.0 | 0.022046 |  1.969359 |  93.04142 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.101339 | 0.028592 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036092 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |   0.00864 |              0.249736 |                  0.249736 |                   1.196665 |                   1.21168 |                      0.0 |     0.199249 |              0.0 |         -0.015015 |             33.810423 |                 100.0 |                82.610424 |               97.908583 |      0.0 |      0.0 |  0.131541 |       0.0 | 0.046722 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.925463 | 0.256945 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.124471 | 0.034575 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.991619 | 92.994698 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098721 |  0.02575 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034575 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |   0.01296 |              0.195223 |                  0.195223 |                   1.259762 |                  1.349977 |                      0.0 |     0.138974 |              0.0 |         -0.090215 |             29.178858 |                 100.0 |                82.671875 |               97.917336 |      0.0 |      0.0 |  0.146498 |       0.0 | 0.047218 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.924792 | 0.256572 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.119362 | 0.033156 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  2.009559 |  92.94748 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.096171 | 0.023191 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033156 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |   0.01728 |              0.048073 |                  0.048073 |                   1.291311 |                  1.428152 |                      0.0 |    -0.011433 |              0.0 |         -0.136842 |             28.400756 |                 100.0 |                82.733986 |               97.926172 |      0.0 |      0.0 |  0.135014 |       0.0 | 0.042609 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.924282 | 0.256289 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.114573 | 0.031826 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  2.023179 | 92.904871 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093687 | 0.020886 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031826 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |   0.02808 |              0.000547 |                  0.000547 |                   1.354408 |                  1.571379 |                      0.0 |    -0.064927 |              0.0 |         -0.216971 |             30.429381 |                 100.0 |                 82.79004 |               98.773039 |      0.0 |      0.0 |   0.11254 |       0.0 | 0.036657 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.924177 | 0.256231 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.110077 | 0.030577 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  2.025999 | 92.868214 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091267 |  0.01881 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030577 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.03456 |                   0.0 |                       0.0 |                   1.354408 |                  1.583999 |                      0.0 |       -0.066 |              0.0 |          -0.22959 |             35.500944 |                 100.0 |                 82.83827 |               98.814229 |      0.0 |      0.0 |   0.08847 |       0.0 |  0.03163 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.924314 | 0.256307 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.10585 | 0.029403 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  2.022339 | 92.836584 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088909 | 0.016941 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029403 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |    0.0432 |                   0.0 |                       0.0 |                   1.354408 |                    1.5975 |                      0.0 |    -0.066562 |              0.0 |         -0.243092 |             44.376181 |                 100.0 |                82.879889 |               98.814229 |      0.0 |      0.0 |  0.063209 |       0.0 | 0.025658 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.924774 | 0.256562 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.10187 | 0.028297 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  2.010039 | 92.810926 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086613 | 0.015257 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028297 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04536 |                   0.0 |                       0.0 |                   1.354408 |                  1.611195 |                      0.0 |    -0.067133 |              0.0 |         -0.256786 |             50.715635 |                 100.0 |                82.913653 |               98.814229 |      0.0 |      0.0 |  0.054415 |       0.0 |  0.02375 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.925315 | 0.256862 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.098116 | 0.027254 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.995579 | 92.787176 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084375 | 0.013741 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027254 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04752 |                   0.0 |                       0.0 |                   1.354408 |                  1.625096 |                      0.0 |    -0.067712 |              0.0 |         -0.270688 |             43.470544 |                 100.0 |                82.944908 |               98.814229 |      0.0 |      0.0 |  0.065452 |       0.0 |  0.02619 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.925937 | 0.257208 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.094571 |  0.02627 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.978959 | 92.760986 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.082196 | 0.012375 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02627 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.05616 |                   0.0 |                       0.0 |                   1.354408 |                  1.639665 |                      0.0 |    -0.068319 |              0.0 |         -0.285257 |             53.251417 |                 100.0 |                82.979376 |               98.814229 |      0.0 |      0.0 |  0.048111 |       0.0 | 0.021424 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926883 | 0.257733 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.091218 | 0.025338 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.953699 | 92.739561 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.080073 | 0.011145 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025338 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.06696 |                   0.0 |                       0.0 |                   1.354408 |                  1.653965 |                      0.0 |    -0.068915 |              0.0 |         -0.299557 |             62.648725 |                 100.0 |                83.007575 |               98.814229 |      0.0 |      0.0 |  0.034966 |       0.0 | 0.016822 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928232 | 0.258484 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.088042 | 0.024456 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.917639 | 92.722739 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.078004 | 0.010037 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024456 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.08424 |                   0.0 |                       0.0 |                   1.354408 |                  1.664915 |                      0.0 |    -0.069371 |              0.0 |         -0.310507 |             62.648725 |                 100.0 |                83.029717 |               98.814229 |      0.0 |      0.0 |  0.024921 |       0.0 |  0.01189 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.930228 | 0.259597 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.085029 | 0.023619 |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.864299 | 92.710848 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.075989 |  0.00904 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023619 |
"""
# import...
# ...from HydPy
from hydpy.exe.modelimports import *
from hydpy.core import masktools
from hydpy.core import modeltools

# ...from lland
from hydpy.models.lland import lland_model
from hydpy.models.lland import lland_masks
from hydpy.models.lland.lland_constants import *


class Model(modeltools.AdHocModel):
    """Penman-Monteith and Knauf version of HydPy-L-Land (|lland_v3|)."""

    INLET_METHODS = (lland_model.Pick_QZ_V1,)
    RECEIVER_METHODS = ()
    ADD_METHODS = (
        lland_model.Return_AdjustedWindSpeed_V1,
        lland_model.Return_ActualVapourPressure_V1,
        lland_model.Return_NetLongwaveRadiationSnow_V1,
        lland_model.Return_DailyGlobalRadiation_V1,
        lland_model.Return_Penman_V1,
        lland_model.Return_PenmanMonteith_V1,
        lland_model.Return_EnergyGainSnowSurface_V1,
        lland_model.Return_SaturationVapourPressure_V1,
        lland_model.Return_SaturationVapourPressureSlope_V1,
        lland_model.Return_NetRadiation_V1,
        lland_model.Return_WSensSnow_V1,
        lland_model.Return_WLatSnow_V1,
        lland_model.Return_WSurf_V1,
        lland_model.Return_TempS_V1,
        lland_model.Return_WG_V1,
        lland_model.Return_BackwardEulerError_V1,
        lland_model.Return_ESnow_V1,
        lland_model.Return_TempSSurface_V1,
    )
    RUN_METHODS = (
        lland_model.Calc_QZH_V1,
        lland_model.Update_LoggedTemL_V1,
        lland_model.Calc_TemLTag_V1,
        lland_model.Update_LoggedRelativeHumidity_V1,
        lland_model.Calc_DailyRelativeHumidity_V1,
        lland_model.Update_LoggedSunshineDuration_V1,
        lland_model.Calc_NKor_V1,
        lland_model.Calc_TKor_V1,
        lland_model.Calc_TKorTag_V1,
        lland_model.Calc_WindSpeed2m_V1,
        lland_model.Calc_ReducedWindSpeed2m_V1,
        lland_model.Update_LoggedWindSpeed2m_V1,
        lland_model.Calc_DailyWindSpeed2m_V1,
        lland_model.Calc_WindSpeed10m_V1,
        lland_model.Calc_SolarDeclination_V1,
        lland_model.Calc_TSA_TSU_V1,
        lland_model.Calc_DailySunshineDuration_V1,
        lland_model.Calc_EarthSunDistance_V1,
        lland_model.Calc_ExtraterrestrialRadiation_V1,
        lland_model.Calc_PossibleSunshineDuration_V1,
        lland_model.Calc_DailyPossibleSunshineDuration_V1,
        lland_model.Calc_SP_V1,
        lland_model.Calc_GlobalRadiation_V1,
        lland_model.Update_LoggedGlobalRadiation_V1,
        lland_model.Calc_DailyGlobalRadiation_V1,
        lland_model.Calc_AdjustedGlobalRadiation_V1,
        lland_model.Calc_SaturationVapourPressure_V1,
        lland_model.Calc_DailySaturationVapourPressure_V1,
        lland_model.Calc_SaturationVapourPressureSlope_V1,
        lland_model.Calc_DailySaturationVapourPressureSlope_V1,
        lland_model.Calc_ActualVapourPressure_V1,
        lland_model.Calc_DailyActualVapourPressure_V1,
        lland_model.Calc_NBes_Inzp_V1,
        lland_model.Calc_SNRatio_V1,
        lland_model.Calc_SBes_V1,
        lland_model.Calc_WATS_V1,
        lland_model.Calc_WaDa_WAeS_V1,
        lland_model.Calc_WNied_ESnow_V1,
        lland_model.Calc_TempS_V1,
        lland_model.Update_TauS_V1,
        lland_model.Calc_ActualAlbedo_V1,
        lland_model.Calc_NetShortwaveRadiationSnow_V1,
        lland_model.Calc_TZ_V1,
        lland_model.Calc_WG_V1,
        lland_model.Update_ESnow_V1,
        lland_model.Calc_SchmPot_V2,
        lland_model.Calc_Schm_WATS_V1,
        lland_model.Calc_GefrPot_V1,
        lland_model.Calc_Gefr_WATS_V1,
        lland_model.Calc_EvS_WAeS_WATS_V1,
        lland_model.Update_WaDa_WAeS_V1,
        lland_model.Update_ESnow_V2,
        lland_model.Calc_NetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetLongwaveRadiation_V1,
        lland_model.Calc_NetRadiation_V1,
        lland_model.Calc_DailyNetRadiation_V1,
        lland_model.Calc_DryAirPressure_V1,
        lland_model.Calc_DensityAir_V1,
        lland_model.Calc_AerodynamicResistance_V1,
        lland_model.Calc_SoilSurfaceResistance_V1,
        lland_model.Calc_LanduseSurfaceResistance_V1,
        lland_model.Calc_ActualSurfaceResistance_V1,
        lland_model.Calc_G_V2,
        lland_model.Calc_EvPo_V2,
        lland_model.Calc_EvI_Inzp_V1,
        lland_model.Calc_EvB_V2,
        lland_model.Update_EBdn_V1,
        lland_model.Calc_SFF_V1,
        lland_model.Calc_FVG_V1,
        lland_model.Calc_QKap_V1,
        lland_model.Calc_QBB_V1,
        lland_model.Calc_QIB1_V1,
        lland_model.Calc_QIB2_V1,
        lland_model.Calc_QDB_V1,
        lland_model.Update_QDB_V1,
        lland_model.Calc_BoWa_V1,
        lland_model.Calc_QBGZ_V1,
        lland_model.Calc_QIGZ1_V1,
        lland_model.Calc_QIGZ2_V1,
        lland_model.Calc_QDGZ_V1,
        lland_model.Calc_QDGZ1_QDGZ2_V1,
        lland_model.Calc_QBGA_V1,
        lland_model.Calc_QIGA1_V1,
        lland_model.Calc_QIGA2_V1,
        lland_model.Calc_QDGA1_V1,
        lland_model.Calc_QDGA2_V1,
        lland_model.Calc_QAH_V1,
        lland_model.Calc_QA_V1,
    )
    OUTLET_METHODS = (lland_model.Pass_QA_V1,)
    SENDER_METHODS = ()
    SUBMODELS = (
        lland_model.PegasusESnow,
        lland_model.PegasusTempSSurface,
    )
    idx_hru = modeltools.Idx_HRU()


class Masks(masktools.Masks):
    """Masks applicable to |lland_v3|."""

    # pylint: disable=no-member
    # bug of pylint 2.4?
    CLASSES = lland_masks.Masks.CLASSES


tester = Tester()
cythonizer = Cythonizer()
cythonizer.finalise()
